/**************************************************************************************
    arg_traits.cxx
    functions for manipulating the argument structure for the
      portable version of the text management library


    (c) Copyright Apollo Computer Inc., 1987 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

#include <tml/tml_objects.hxx>

#define NULL 0

/***************************************************************************
    Class functions for the argument vector elements

**************************************************************************/

/*constructor for an argument vector element*/
argument::argument(long arg_size , model* to_model )
{
    register argument    *new_arg;

    arg_size = arg_size ? arg_size : sizeof(argument);

    if (!this){
        if (to_model){
            new_arg = to_model->get_current_oheap()->get_tf_arg(arg_size);
            if (new_arg)
                to_model->get_current_oheap()->tf_arg = new_arg->next;
            else {
                new_arg = (argument *)(to_model->get_current_oheap()->alloc(arg_size));
                if (!new_arg)
                    return;
                }
            this = new_arg;
            }
        else{
            new_arg = (argument *)(new char [(long) sizeof(argument)]);
            if (!new_arg)
                return;
            else
                this = new_arg;
            }
        }
    else
        this = this;
    in_model = to_model;
    ref_count = 1;
    next = 0;
    a_type = no_val_arg;
}


/*destructor for an argument vector element*/
argument::~argument()
{
    next = 0;
    ref_count = 0;
    a_type = no_val_arg;
    if (in_model){
        in_model->current_oheap->free(this);
        this = 0;
        }
}
                
/*get value calls that will soon be obsolete*/
long argument::get_ival()
{
   return ((a_type == int_val_arg) && arg_val) ? (*(long*)arg_val) : -1;
}

float argument::get_fval()
{
   return ((a_type == float_val_arg) && arg_val) ? (*(float*)arg_val) : -1;
}

char* argument::get_sval()
{
   return (a_type == string_val_arg) ? (char*)arg_val : 0;
}

long argument::get_scount()
{
   register string_arg*   sa = (string_arg*)this;

   return (a_type == string_val_arg) ? sa->get_scount() : 0;
}


/*copy a single argument vector element to another model*/
argument* argument::copy(model* to_model)
{
    register argument *new_arg;

    if (!to_model)
        to_model = in_model;
    new_arg = to_model ? to_model->make_arg((long)1) : new argument;
    new_arg->a_type = no_val_arg;
    return new_arg;
}

/*copy all of an argument vector to another model*/
argument* argument::copy_all(model* to_model )
{
    register argument *new_arg;

    new_arg = copy(to_model);
    if (next)
        new_arg->next = next->copy(to_model);
    return new_arg;
}

/* Call this function when a call to argument::detach() returns 0 from */
/*   a request to detach an argument from a property object's entry argument vector. */
/* Argument "on_prop" is the property whose entry argument vector contained only */
/*   the one argument object "removed". */
void first_entry_arg_hist(property* on_prop, argument* removed)
{
    history  *past;
    atom     *this_atom;
    atom_arg *arg1, *arg2;

    if (NULL != on_prop->get_model()) {
        past = on_prop->get_model()->get_past();
        if (past->has_hist() && past->is_active()) {

            // if history is active, prepare to record this call
            // NOTE: the prop_vec object's reference counter will 
            //   have been handled properly by 
            // 
            arg1 = new atom_this_arg(past, on_prop);
            arg2 = new atom_ptr_arg(past, removed);
            this_atom = new atom(past, (WHOLE_PTR) &property::attach_entry, (WHOLE_PTR) &property::detach_entry,
                                 arg1, arg2, NULL, arg1, NULL);
            past->add_atom(this_atom);
            }
        }
}

/* Call this function when a call to argument::detach() returns 0 from */
/*   a request to detach an argument from a property object's exit argument vector. */
/* Argument "on_prop" is the property whose exit argument vector contained only */
/*   the one argument object "removed". */
void first_exit_arg_hist(property* on_prop, argument* removed)
{
    history  *past;
    atom     *this_atom;
    atom_arg *arg1, *arg2;

    if (NULL != on_prop->get_model()) {
        past = on_prop->get_model()->get_past();
        if (past->has_hist() && past->is_active()) {

            // if history is active, prepare to record this call
            // NOTE: the prop_vec object's reference counter will 
            //   have been handled properly by 
            // 
            arg1 = new atom_this_arg(past, on_prop);
            arg2 = new atom_ptr_arg(past, removed);
            this_atom = new atom(past, (WHOLE_PTR) &property::attach_exit, (WHOLE_PTR) &property::detach_exit,
                                 arg1, arg2, NULL, arg1, NULL);
            past->add_atom(this_atom);
            }
        }
}

/*detach an argument from an argument vector*/
/* WARNING: If the argument being detached in the only argument vector element in from_vec, */
/*  history is not written, but the return value is 0 in this case and this case only. */
/*  THEREFORE, if you get a 0 return from this function, you are obliged to add the */
/*  appropriate atom to history. NOTE: A call to first_entry_arg_hist() or */
/*  first_exit_arg_hist (see above) will perform this work for you. Use first_entry_arg_hist() */
/*  when argument::detach() returns 0 from a call to detach an argument from an entry argument */
/*  vector; use first_exit_arg_hist() when argument::detach() returns 0 from a call to detach */
/*  an argument from an exit argument vector. EXCEPTION: 0 returns from this function on undo of */
/*  argument::append are not significant because history is not active during undo. */
argument* argument::detach(argument *from_vec)
{
    register argument *hold_arg;
    register argument *return_arg;
                                           
    history  *past;
    long      has_hist = 0;
    long      hist_active = 0;
    atom*     this_atom;
    atom_arg *arg1 = 0, *arg2 = 0, *arg3 = 0, *arg4;

    // to keep ref_count straight, must have 'this' as both atom_this_arg and atom_ptr_arg

    if (NULL != in_model) {
        past = in_model->get_past();
        if (hist_active = ((has_hist = past->has_hist()) && past->is_active())) {

            // If history is active, record this call

            arg1 = new atom_this_arg(past, this);       // 'this' as object in redo call
            arg2 = new atom_ptr_arg(past, this);        // 'this' as argument for undo and redo calls
            arg3 = new atom_val_arg(past, from_vec);
            }
        }

    if (from_vec == this) {
        return_arg = next;

        // no way to record history of argument::detach() if only one argument in vector
        //   return value of 0 signals this condition

        if (next && hist_active) {

            // If history is active, record this call

            arg4 = new atom_this_arg(past, next);
            next = 0;    // prevent repetitions on undo
            this_atom = new atom(past, (WHOLE_PTR) &argument::prepend,
                                 (WHOLE_PTR) &argument::detach,
                                 arg4, arg2, NULL, arg1, arg3, NULL);
            past->add_atom(this_atom);
            }
        else {
            delete arg1;
            delete arg2;
            delete arg3;
            }
        }
    else {
        return_arg = from_vec;
        for (hold_arg = from_vec; hold_arg->next != 0; hold_arg = hold_arg->next)
            if (hold_arg->next == this) break;
        if (!hold_arg->next) return(return_arg);
        hold_arg->next = next;

        if (hist_active) {

            // If history is active, record this call

            arg4 = new atom_this_arg(past, next);
            next = 0;    // prevent repetitions on undo
            this_atom = new atom(past, (WHOLE_PTR) &argument::prepend,
                                 (WHOLE_PTR) &argument::detach,
                                 arg3, arg2, NULL, arg1, arg4, NULL);
            past->add_atom(this_atom);
            }
        else {
            delete arg1;
            delete arg2;
            delete arg3;
            }
        }

    if (0 == has_hist)
        if (!--ref_count) delete this;
    return(return_arg);
}

/*append an argument vector element to the end of an argument vector*/
void argument::append(argument *add_arg)
{
    register argument *append_arg,
                      *end_element;
    register long      i;

    history  *past;
    long      has_hist;
    long      is_active = 0;
    atom     *this_atom;
    atom_arg *arg1, *arg2, *arg3;

    // If add_arg node in same model as "this", copy to this model
    // NOTE: This block must precede block setting history

    if (in_model != add_arg->in_model)
        append_arg = add_arg->copy_all(in_model);
    else
        append_arg = add_arg;

    /*find the end element of the new head of the list, and attach the old list to it*/
    for(end_element = this; end_element->next != 0; end_element = end_element->next);
    end_element->next = append_arg;

    // NOTE: This block must follow block copying add_arg to this model
    //   so history will be set properly

    if (NULL != in_model) {
        past = in_model->get_past();
        if ((has_hist = past->has_hist()) && (is_active = past->is_active())) {

            // If history is active, record this call

            arg1 = new atom_this_arg(past, this);
            arg2 = new atom_ptr_arg(past, 0);
            arg3 = new atom_ptr_arg(past, append_arg);
            this_atom = new atom(past, (WHOLE_PTR) &argument::set_m_ptr, (WHOLE_PTR) &argument::set_m_ptr,
                                 arg1, arg2, NULL, arg1, arg3, NULL);
            past->add_atom(this_atom);
            }
        }

    /*update the ref counts on this to reflect references to the new head*/
    if (is_active || (0 == has_hist)) {
        i = end_element->ref_count - 1;
        if (i){
            for(end_element = append_arg; end_element != 0; end_element = end_element->next)
                end_element->ref_count += i;
             }
        }
}

/*prepend and argument vector element to the beginning of an argument vector*/
argument *argument::prepend(argument *add_arg)
{
    register argument *return_vec,
                      *end_element;
    register long      i;

    history  *past;
    long      has_hist;
    long      is_active = 0;
    argument *cur_vec;
    argument *next_vec;
    atom     *this_atom;
    atom_arg *arg1, *arg2, *arg3;

    // If add_arg not in same model as "this", copy to this model
    // NOTE: This block must precede block setting history

    if (in_model != add_arg->in_model)
        return_vec = add_arg->copy_all(in_model);
    else
        return_vec = add_arg;

    /*find the end element of the new head of the list, and attach the old list to it*/
    for(end_element = return_vec; end_element->next != 0; end_element = end_element->next);
    end_element->next = this;

    if (NULL != in_model) {

        // NOTE: This block must follow block copying add_arg to this model
        //   so history will be set properly

        past = in_model->get_past();
        if ((has_hist = past->has_hist()) && (is_active = past->is_active())) {

            // If history is active, record this call

            for (cur_vec = return_vec, next_vec = return_vec->next;
              cur_vec != this;
              cur_vec = cur_vec->next, next_vec = next_vec->next) {
                arg1 = new atom_this_arg(past, cur_vec);
                arg2 = new atom_ptr_arg(past, return_vec);
                arg3 = new atom_this_arg(past, next_vec);

                /*this works because atoms are undone/redone in reverse order of creation*/

                this_atom = new atom(past, (WHOLE_PTR) &argument::detach, (WHOLE_PTR) &argument::prepend,
                                     arg1, arg2, NULL, arg3, arg2, NULL);
                past->add_atom(this_atom);
                }
            }
        }

    /*update the ref counts on this to reflect references to the new head*/
    if ((is_active) || (0 == has_hist)) {
        i = end_element->ref_count - 1;
        if (i){
            for(end_element = this; end_element != 0; end_element = end_element->next)
                end_element->ref_count += i;
            }
        }
    return(return_vec);
}


/*****************************************************************************************************

        Member functions for derived class int_arg

*****************************************************************************************************/
                                                  
int_arg::int_arg(model* to_model, long size ) : (size ? size : (long) sizeof(int_arg), to_model)
{
    register int_arg   *new_arg;

    size = size ? size : sizeof(int_arg);

    if (!this){
        if (to_model){
            new_arg = (int_arg*)to_model->get_current_oheap()->get_tf_arg(size);
            if (new_arg)
                to_model->get_current_oheap()->tf_arg = new_arg->next;
            else {
                new_arg = (int_arg*)(to_model->get_current_oheap()->alloc(size));
                if (!new_arg)
                    return;
                }
            this = new_arg;
            }
        else{
            new_arg = (int_arg*)(new char [size]);
            if (!new_arg)
                return;
            else
                this = new_arg;
            }
        }
    else
        this = this;

    set_model(to_model);

    set_vtype (int_val_arg);
    arg_val = NULL;
}


int_arg::~int_arg()
{
    arg_val = NULL;
}


/* function declared because it was impossible to get pointers */
/*   to the overloaded argument::set_val(...) functions. */

void fpi(int_arg* arg, long i)
{
    arg->set_val(i);
}


void int_arg::set_val(long value)
{
    history  *past;
    atom     *this_atom;
    model    *this_model;
    atom_arg *arg1, *arg2, *arg3, *arg4;

    if (this_model = get_model()) {
        past = this_model->get_past();
        if (past->has_hist() && past->is_active()) {

            // If history is active, record this call

            arg1 = new atom_this_arg(past, NULL);
            arg2 = new atom_this_arg(past, this);
            if (arg_val)
                arg3 = new atom_val_arg(past, *((long*) arg_val));
            else
                arg3 = new atom_val_arg(past, (long) 0);
            arg4 = new atom_val_arg(past, value);
            this_atom = new atom(past, (WHOLE_PTR) &fpi, (WHOLE_PTR) &fpi,
                                 arg1, arg2, arg3, NULL, arg1, arg2, arg4, NULL);
            past->add_atom(this_atom);
        }
    }
    set_vtype(int_val_arg);
    if (!arg_val) {
        if (get_model())
            arg_val = (long*) get_model()->get_current_oheap()->alloc((long) sizeof(WHOLE_PTR));
        else
            arg_val =  new long;
        }
    *((long*)arg_val) = value;
}


argument* int_arg::copy(model* to_model)
{
    register argument *new_arg;

    if (!to_model)
        to_model = get_model();

    new_arg = to_model ? to_model->make_arg(*((long*) arg_val)) : new int_arg(to_model);
    if (NULL == to_model)
        ((int_arg*) new_arg)->set_val(*((long*) arg_val));

    return new_arg;
}


/*****************************************************************************************************

        Member functions for derived class float_arg

*****************************************************************************************************/
                                                  
float_arg::float_arg(model* to_model, long size) : (size ? size : (long) sizeof(float_arg), to_model)
{
    register float_arg   *new_arg;

    size = size ? size : sizeof(float_arg);

    if (!this){
        if (to_model){
            new_arg = (float_arg*)to_model->get_current_oheap()->get_tf_arg(size);
            if (new_arg)
                to_model->get_current_oheap()->tf_arg = new_arg->next;
            else {
                new_arg = (float_arg*)(to_model->get_current_oheap()->alloc(size));
                if (!new_arg)
                    return;
                }
            this = new_arg;
            }
        else{
            new_arg = (float_arg*)(new char [size]);
            if (!new_arg)
                return;
            else
                this = new_arg;
            }
        }
    else
        this = this;

    set_model (to_model);

    set_vtype (float_val_arg);

    arg_val = NULL;
}


float_arg::~float_arg()
{
    arg_val = NULL;
}


/* function declared because it was impossible to get pointers */
/*   to the overloaded argument::set_val(...) functions. */

void fpf(float_arg* arg, float f)
{
    arg->set_val(f);
}


void float_arg::set_val(float value)
{
    history  *past;
    atom     *this_atom;
    model    *this_model;
    atom_arg *arg1, *arg2, *arg3, *arg4;

    if (this_model = get_model()) {
        past = this_model->get_past();
        if (past->has_hist() && past->is_active()) {

            // If history is active, record this call

            arg1 = new atom_this_arg(past, NULL);
            arg2 = new atom_this_arg(past, this);
            if (arg_val)
                arg3 = new atom_val_arg(past, *((float*) arg_val));
            else
                arg3 = new atom_val_arg(past, (float) 0);
            arg4 = new atom_val_arg(past, value);
            this_atom = new atom(past, (WHOLE_PTR) &fpf, (WHOLE_PTR) &fpf,
                                 arg1, arg2, arg3, NULL, arg1, arg2, arg4, NULL);
            past->add_atom(this_atom);
            }
        }
    set_vtype (float_val_arg);
    if (!arg_val) {
        if (get_model())
            arg_val = (void*) get_model()->get_current_oheap()->alloc((long) sizeof(float));
        else
            arg_val = (void*) new float;
        }
    *((float*) arg_val) = value;
}

  
argument* float_arg::copy(model* to_model )
{
    register argument *new_arg;

    if (!to_model)
        to_model = get_model();
    new_arg = to_model ? to_model->make_arg(*((float*) arg_val)) : new float_arg(to_model);
    if (NULL == to_model)
        ((float_arg*) new_arg)->set_val(*((float*) arg_val));
    return new_arg;
}


/*****************************************************************************************************

        Member functions for derived class string_arg

*****************************************************************************************************/
                                                  
string_arg::string_arg (model* to_model, long size ) : (size ? size : (long) sizeof(string_arg), to_model)
{
    register string_arg   *new_arg;

    size = size ? size : sizeof(string_arg);

    if (!this){
        if (to_model){
            new_arg = (string_arg*)to_model->get_current_oheap()->get_tf_arg(size);
            if (new_arg)
                to_model->get_current_oheap()->tf_arg = new_arg->next;
            else {
                new_arg = (string_arg*)(to_model->get_current_oheap()->alloc(size));
                if (!new_arg)
                    return;
                }
            this = new_arg;
            }
        else{
            new_arg = (string_arg*)(new char [size]);
            if (!new_arg)
                return;
            else
                this = new_arg;
            }
        }
    else
        this = this;

    set_model (to_model);
    set_vtype (string_val_arg);
    arg_val = NULL;
    count = 0;
}


string_arg::~string_arg()
{
    arg_val = NULL;
}


/* function declared because it was impossible to get pointers */
/*   to the overloaded argument::set_val(...) functions. */

void fps(string_arg* arg, char* s, long count)
{
    arg->set_val(s, count);
}


void string_arg::set_val(char *value, long val_count)
{
    history  *past;
    atom     *this_atom;
    model    *this_model;
    atom_arg *arg1, *arg2, *arg3, *arg4, *arg5, *arg6;

    if (this_model = get_model()) {
        past = this_model->get_past();
        if (past->has_hist() && past->is_active()) {

            // If history is active, record this call

            arg1 = new atom_this_arg(past, NULL);
            arg2 = new atom_this_arg(past, this);
            arg3 = new atom_val_arg(past, arg_val);
            arg4 = new atom_val_arg(past, count);
            arg5 = new atom_val_arg(past, value);
            arg6 = new atom_val_arg(past, val_count);
            this_atom = new atom(past, (WHOLE_PTR) &fps, (WHOLE_PTR) &fps,
                                arg1, arg2, arg3, arg4, NULL, arg1, arg2, arg5, arg6, NULL);
            past->add_atom(this_atom);
        }
    }
    set_vtype (string_val_arg);
    count = val_count;
    if (get_model())
        arg_val = get_model()->get_current_sheap()->alloc(value, val_count);
    else
        arg_val = value;
}


argument* string_arg::copy(model* to_model)
{
    register argument *new_arg;

    if (!to_model)
        to_model = get_model();
    new_arg = to_model ? to_model->make_arg((char*) arg_val, count) : new string_arg(to_model);
    if (NULL == to_model)
        ((string_arg*) new_arg)->set_val((char*) arg_val, count);
    return new_arg;
}

/*****************************************************************************************************

        Member functions for derived class subtree_arg

*****************************************************************************************************/
                                                  
subtree_arg::subtree_arg (model* to_model , long size ) : (size ? size : (long) sizeof(subtree_arg), to_model)
{
    register subtree_arg   *new_arg;

    size = size ? size : sizeof(subtree_arg);

    if (!this){
        if (to_model){
            new_arg = (subtree_arg*)to_model->get_current_oheap()->get_tf_arg(size);
            if (new_arg)
                to_model->get_current_oheap()->tf_arg = new_arg->next;
            else {
                new_arg = (subtree_arg*)(to_model->get_current_oheap()->alloc(size));
                if (!new_arg)
                    return;
                }
            this = new_arg;
            }
        else{
            new_arg = (subtree_arg*)(new char [size]);
            if (!new_arg)
                return;
            else
                this = new_arg;
            }
        }
    else
        this = this;

    set_model (to_model);
    set_vtype (subtree_val_arg);
    arg_val = NULL;
}


subtree_arg::~subtree_arg()
{
    arg_val = NULL;
}


/* function declared because it was impossible to get pointers */
/*   to the overloaded argument::set_val(...) functions. */

void fpn(subtree_arg* arg, node *subtree)
{
    arg->set_val(subtree);
}


void subtree_arg::set_val(node *subtree)
{
    history  *past;
    atom     *this_atom;
    model    *this_model;
    atom_arg *arg1, *arg2, *arg3, *arg4;

    if (this_model = get_model()) {
        past = this_model->get_past();
        if (past->has_hist() && past->is_active()) {

            // If history is active, record this call

            arg1 = new atom_this_arg(past, NULL);
            arg2 = new atom_this_arg(past, this);
            arg3 = new atom_val_arg(past, arg_val);
            arg4 = new atom_val_arg(past, subtree);
            this_atom = new atom(past, (WHOLE_PTR) &fpn, (WHOLE_PTR) &fpn,
                                 arg1, arg2, arg3, NULL, arg1, arg2, arg4, NULL);
            past->add_atom(this_atom);
        }
    }
    set_vtype(subtree_val_arg);
    if (subtree) {
        if (get_model() != subtree->get_model())
            arg_val = subtree->copy_all(get_model());
        else
            arg_val = subtree;
        }
    else
        arg_val = NULL;
}

argument* subtree_arg::copy(model* to_model )
{
    register argument *new_arg;

    if (!to_model)
        to_model = get_model();
    new_arg = to_model ? to_model->make_arg((node*) arg_val) : new subtree_arg(to_model);
    if (NULL == to_model)
        ((subtree_arg*) new_arg)->set_val((node*) arg_val);
    return new_arg;
}

/*****************************************************************************************************

        Member functions for derived class char_set_arg

*****************************************************************************************************/
                                                  
char_set_arg::char_set_arg(model* to_model , long size ) : (size ? size : (long) sizeof(char_set_arg), to_model)
{
    register char_set_arg *new_arg;

    size = size ? size : sizeof(char_set_arg);

    if (!this){
        if (to_model){
            new_arg = (char_set_arg*)to_model->get_current_oheap()->get_tf_arg(size);
            if (new_arg)
                to_model->get_current_oheap()->tf_arg = new_arg->next;
            else {
                new_arg = (char_set_arg*)(to_model->get_current_oheap()->alloc(size));
                if (!new_arg)
                    return;
                }
            this = new_arg;
            }
        else{
            new_arg = (char_set_arg*)(new char [size]);
            if (!new_arg)
                return;
            else
                this = new_arg;
            }
        }
    else
        this = this;

    set_model(to_model);
    set_vtype(char_set_val_arg);
    arg_val = NULL;
}


char_set_arg::~char_set_arg()
{
    delete ((char_set*) arg_val);
    arg_val = NULL;
}


/* function declared because it was impossible to get pointers */
/*   to the overloaded argument::set_val(...) functions. */

void fpcs(char_set_arg* arg, char_set* new_char_set)
{
    arg->set_val(new_char_set);
}


void char_set_arg::set_val(char_set* new_char_set)
{
    history  *past;
    atom     *this_atom;
    model    *this_model;
    atom_arg *arg1, *arg2, *arg3, *arg4;

    if (this_model = get_model()) {
        past = this_model->get_past();
        if (past->has_hist() && past->is_active()) {

            // If history is active, record this call

            arg1 = new atom_this_arg(past, NULL);
            arg2 = new atom_this_arg(past, this);
            arg3 = new atom_val_arg(past, arg_val);
            arg4 = new atom_val_arg(past, new_char_set);
            this_atom = new atom(past, (WHOLE_PTR) &fpcs, (WHOLE_PTR) &fpcs,
                                 arg1, arg2, arg3, NULL, arg1, arg2, arg4, NULL);
            past->add_atom(this_atom);
        }
    }
    set_vtype(char_set_val_arg);
    if (0 == past->has_hist())
        new_char_set->incr_ref_count();
    arg_val = new_char_set;
}

argument* char_set_arg::copy(model* to_model )
{
    register argument *new_arg;

    if (!to_model)
        to_model = get_model();
    new_arg = to_model ? to_model->make_arg((char_set*) arg_val) : new char_set_arg(to_model);
    if (NULL == to_model)
        ((char_set_arg*) new_arg)->set_val((char_set*) arg_val);
    return new_arg;
}


/*****************************************************************************************************

        Member functions for derived class union_char_set_arg

*****************************************************************************************************/
                                                  
union_char_set_arg::union_char_set_arg(model* to_model , long size ) : (size ? size : (long) sizeof(union_char_set_arg), to_model)
{
    register union_char_set_arg *new_arg;

    size = size ? size : sizeof(union_char_set_arg);

    if (!this){
        if (to_model){
            new_arg = (union_char_set_arg*)to_model->get_current_oheap()->get_tf_arg(size);
            if (new_arg)
                to_model->get_current_oheap()->tf_arg = new_arg->next;
            else {
                new_arg = (union_char_set_arg*)(to_model->get_current_oheap()->alloc(size));
                if (!new_arg)
                    return;
                }
            this = new_arg;
            }
        else{
            new_arg = (union_char_set_arg*)(new char [size]);
            if (!new_arg)
                return;
            else
                this = new_arg;
            }
        }
    else
        this = this;

    set_model(to_model);
    set_vtype(union_char_set_val_arg);
    arg_val = NULL;
}

union_char_set_arg::~union_char_set_arg()
{
    delete ((union_char_set*) arg_val);
    arg_val = NULL;
}

/* function declared because it was impossible to get pointers */
/*   to the overloaded argument::set_val(...) functions. */

void fpucs(union_char_set_arg* arg, union_char_set* new_union_char_set)
{
    arg->set_val(new_union_char_set);
}


void union_char_set_arg::set_val(union_char_set* new_union_char_set)
{
    history  *past;
    atom     *this_atom;
    model    *this_model;
    atom_arg *arg1, *arg2, *arg3, *arg4;

    if (this_model = get_model()) {
        past = this_model->get_past();
        if (past->has_hist() && past->is_active()) {

            // If history is active, record this call

            arg1 = new atom_this_arg(past, NULL);
            arg2 = new atom_this_arg(past, this);
            arg3 = new atom_val_arg(past, arg_val);
            arg4 = new atom_val_arg(past, new_union_char_set);
            this_atom = new atom(past, (WHOLE_PTR) &fpucs, (WHOLE_PTR) &fpucs,
                                 arg1, arg2, arg3, NULL, arg1, arg2, arg4, NULL);
            past->add_atom(this_atom);
        }
    }
    set_vtype (union_char_set_val_arg);
    if (0 == past->has_hist())
        new_union_char_set->incr_ref_count();
    arg_val = new_union_char_set;
}

argument* union_char_set_arg::copy(model* to_model )
{
    register argument *new_arg;

    if (!to_model)
        to_model = get_model();

    new_arg = to_model ? to_model->make_arg((union_char_set*) arg_val) : new union_char_set_arg(to_model);

    ((union_char_set_arg*) new_arg)->set_val((union_char_set*) arg_val);
    return (new_arg);
}

