/**************************************************************************************
    char_set_mgr.cxx
    Member functions for classes implementing character set
      management

    (c) Copyright Apollo Computer Inc., 1989 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/
               
#include <stdio.h>
#include <string.h>

#ifndef apollo
#ifdef bsd
#include <strings.h>
#endif
#endif

#include <tml/tml_os_depends.hxx>
#include <tml/char_set_mgr.hxx>

/***************************************************************************************
    definition of minimum size to trigger sorting of ranges in a char_range_set
      to increase efficiency of lookup
****************************************************************************************/

#define CSM_SORT_CRITERION  8

/**************************************************************************************
    global default character set
***************************************************************************************/

// To make this global per-process, must be declared with attribute specifier
//   for Apollo builds to allow use in global and inlib'd libraries
#ifdef apollo
union_char_set* csm_dflt_union_char_set __attribute((section(csm_dflt_union_char_set))) = NULL;
#else
union_char_set* csm_dflt_union_char_set = NULL;
#endif

union_char_set* get_dflt_union_char_set()
{
    alpha_char_set* new_char_set;

    if (NULL == csm_dflt_union_char_set) {
        new_char_set = new alpha_char_set(sizeof(char));
        ((alpha_char_set*) new_char_set)->become_iso_latin_1();
        csm_dflt_union_char_set = new union_char_set(new_char_set);
        }
    return csm_dflt_union_char_set;
};


/**************************************************************************************
    maximum index of default entities associated with character sets (zero-based)
***************************************************************************************/

const long LAST_DEFAULT_HASH = 1;

/**************************************************************************************
    member functions for class csm_character
***************************************************************************************/

csm_character::csm_character()
{
    value = NULL;
    num_chars = 0;
    collation = MAX_CHARS_IN_SET;
    in_char_set = NULL;
    if (NULL != csm_dflt_union_char_set)
        in_char_set = csm_dflt_union_char_set->get_sets();
}

csm_character::csm_character(char_set* for_char_set)
{
    value = NULL;
    num_chars = 0;
    collation = MAX_CHARS_IN_SET;
    if (for_char_set)
        in_char_set = for_char_set;
    else
        in_char_set = get_dflt_union_char_set()->get_sets();
}

csm_character::csm_character(unsigned char* val, char_set* for_char_set)
{
    if (NULL == for_char_set)
        for_char_set = get_dflt_union_char_set()->get_sets();
    init(val, for_char_set->cs_strlen(val), for_char_set, MAX_CHARS_IN_SET);
}

csm_character::csm_character(unsigned char* val, unsigned long len, char_set* for_char_set)
{
    if (NULL == for_char_set)
        for_char_set = get_dflt_union_char_set()->get_sets();
    init(val, len, for_char_set, MAX_CHARS_IN_SET);
}

csm_character::csm_character(unsigned char* val, unsigned long len, unsigned long collation_order, char_set* for_char_set)
{
    if (NULL == for_char_set)
        for_char_set = get_dflt_union_char_set()->get_sets();
    init(val, len, for_char_set, collation_order);
}

csm_character::~csm_character()
{
    delete value;
    value = NULL;
    in_char_set = NULL;
}


// comp_val must be null-terminated in operator overloadings

long csm_character::operator==(unsigned char* comp_val)
{
    return 0 == ((in_char_set) 
                ? in_char_set->cs_strcmp(value, comp_val)
                : strcmp((char*) value, (char*) comp_val));
}

long csm_character::operator!= (unsigned char* comp_val)
{
    return 0 != ((in_char_set) 
                ? in_char_set->cs_strcmp(value, comp_val)
                : strcmp((char*) value, (char*) comp_val));
}

long csm_character::operator<(unsigned char* comp_val)
{
    return 0 > ((in_char_set) 
                ? in_char_set->cs_strcmp(value, comp_val)
                : strcmp((char*) value, (char*) comp_val));
}

long csm_character::operator<=(unsigned char* comp_val)
{
    return 0 >= ((in_char_set) 
                ? in_char_set->cs_strcmp(value, comp_val)
                : strcmp((char*) value, (char*) comp_val));
}

long csm_character::operator>(unsigned char* comp_val)
{
    return 0 < ((in_char_set) 
                ? in_char_set->cs_strcmp(value, comp_val)
                : strcmp((char*) value, (char*) comp_val));
}

long csm_character::operator>=(unsigned char* comp_val)

{
    return 0 <= ((in_char_set) 
                ? in_char_set->cs_strcmp(value, comp_val)
                : strcmp((char*) value, (char*) comp_val));
}

// comp_val need not be null-terminated in these comparison functions

long csm_character::eq(unsigned char* comp_val, unsigned long len)
{
    if (num_chars == len)
        return 0 == ((in_char_set) 
                ? in_char_set->cs_strncmp(value, comp_val, len)
                : strncmp((char*) value, (char*) comp_val, (int) len));
    return 0;
}

long csm_character::ne(unsigned char* comp_val, unsigned long len)
{
    if (num_chars == len)
        return 0 != ((in_char_set) 
                ? in_char_set->cs_strncmp(value, comp_val, len)
                : strncmp((char*) value, (char*) comp_val, (int) len));
    return -1;
}

long csm_character::lt(unsigned char* comp_val, unsigned long len)
{
    if (num_chars >= len)
        return 0 > ((in_char_set) 
                ? in_char_set->cs_strncmp(value, comp_val, len)
                : strncmp((char*) value, (char*) comp_val, (int) len));
    return 0 >= ((in_char_set) 
                ? in_char_set->cs_strncmp(value, comp_val, num_chars)
                : strncmp((char*) value, (char*) comp_val, (int) num_chars));
}

long csm_character::le(unsigned char* comp_val, unsigned long len)
{
    if (num_chars <= len)
        return 0 >= ((in_char_set) 
                ? in_char_set->cs_strncmp(value, comp_val, num_chars)
                : strncmp((char*) value, (char*) comp_val, (int) num_chars));
    return 0 > ((in_char_set) 
                ? in_char_set->cs_strncmp(value, comp_val, len)
                : strncmp((char*) value, (char*) comp_val, (int) len));
}

long csm_character::gt(unsigned char* comp_val, unsigned long len)
{
    if (num_chars <= len)
        return 0 < ((in_char_set) 
                ? in_char_set->cs_strncmp(value, comp_val, num_chars)
                : strncmp((char*) value, (char*) comp_val, (int) num_chars));
    return 0 <= ((in_char_set) 
                ? in_char_set->cs_strncmp(value, comp_val, len)
                : strncmp((char*) value, (char*) comp_val, (int) len));
}

long csm_character::ge(unsigned char* comp_val, unsigned long len)
{
    if (num_chars >= len)
        return 0 <= ((in_char_set) 
                ? in_char_set->cs_strncmp(value, comp_val, len)
                : strncmp((char*) value, (char*) comp_val, (int) len));
    return 0 < ((in_char_set) 
                ? in_char_set->cs_strncmp(value, comp_val, num_chars)
                : strncmp((char*) value, (char*) comp_val, (int) num_chars));
}

void csm_character::init(unsigned char* val, unsigned long len, char_set* for_char_set , unsigned long index )
{
    register unsigned long char_size =
      (for_char_set) ? for_char_set->get_char_size() : 1;
    register unsigned long byte_len = len * char_size;

    value = (unsigned char*) (new char[byte_len + char_size]);
    in_char_set = for_char_set;
    if (for_char_set) {
        value = in_char_set->cs_strncpy(value, val, len);
        for_char_set->get_terminator(value + byte_len);
        }
    else {
        value = (unsigned char*) (strncpy((char*) value, (char*) val, (int) len));
        *(value + len) = '\0';
        }
    num_chars = len;
    collation = index;
}

void csm_character::save(cs_heap* heaps, gc_hash_tbl& hash_tbl)
{
    d_csm_character* savee;
    unsigned char*   val;
    d_ptr            d_ptr_tmp;
    unsigned long    num_bytes = num_chars * in_char_set->get_char_size();

    savee = (d_csm_character*) heaps->obj_alloc(sizeof(d_csm_character));
    val = (unsigned char*) heaps->obj_alloc(num_bytes);
    val = (unsigned char*) (tml_os_byte_copy((char*) value, (char*) val, num_bytes));
    savee->num_chars = num_chars;
    savee->collation = collation;
}

void csm_character::restore(cs_heap* heaps, gc_hash_tbl& hash_tbl)
{
    d_csm_character* restoree;
    unsigned char*   val;
    unsigned long    num_bytes;

    restoree = (d_csm_character*) heaps->obj_realloc(sizeof(d_csm_character));
    restoree->unscramble(heaps->their_endian);
    num_chars = restoree->num_chars;
    in_char_set = heaps->get_cur_char_set();
    num_bytes = num_chars * in_char_set->get_char_size();
    val = (unsigned char*) heaps->obj_realloc(num_bytes);
    value = (unsigned char*) (new char[num_bytes]);
    value = (unsigned char*) (tml_os_byte_copy((char*) val, (char*) value, num_bytes));
    // characters of 2 and 4 bytes must be unscrambled
    switch (in_char_set->get_char_size()) {
        case 2:
            value = (unsigned char*) heaps->their_endian->short_arr_shuf((unsigned short*) value, num_chars);
            break;
        case 4:
            value = (unsigned char*) heaps->their_endian->long_arr_shuf((unsigned long*) value, num_chars);
            break;
        }
    collation = restoree->collation;
}

/**************************************************************************************
    member functions for class char_range
***************************************************************************************/

char_range::char_range()
{
    low.value = NULL;
    low.num_chars = 0;
    low.collation = MAX_CHARS_IN_SET;
    high.value = NULL;
    high.num_chars = 0;
    high.collation = MAX_CHARS_IN_SET;
}

char_range::char_range(unsigned char* lo, unsigned char* hi, char_set* for_char_set,
                       unsigned long lo_map, unsigned long hi_map)
{
    low.init (lo, for_char_set->cs_strlen(lo), for_char_set, lo_map);
    high.init(hi, for_char_set->cs_strlen(hi), for_char_set, hi_map);
}

char_range::char_range(unsigned char* lo, unsigned long lo_len,
                       unsigned char* hi, unsigned long hi_len,
                       char_set* for_char_set,
                       unsigned long lo_map, unsigned long hi_map)
{
    low.init (lo, lo_len, for_char_set, lo_map);
    high.init(hi, hi_len, for_char_set, hi_map);
}

unsigned long char_range::get_size()
{
    char_set* cur_char_set = NULL;

    if (((NULL != low.in_char_set) || (NULL != high.in_char_set)) &&
      (low.in_char_set != high.in_char_set))
        printf("char_range::get_size: different character sets\n");
    if (NULL == (cur_char_set = low.in_char_set))
        cur_char_set = high.in_char_set;
    if (cur_char_set)
        return   cur_char_set->as_ulong(high.value)
               - cur_char_set->as_ulong(low.value)
               + 1;
    // assume standard one-byte character set otherwise
    else
        return   *(high.value) - *(low.value) + 1;
}

void char_range::save(cs_heap* heaps, gc_hash_tbl& hash_tbl)
{
    high.save(heaps, hash_tbl);
    low.save(heaps, hash_tbl);
}

void char_range::restore(cs_heap* heaps, gc_hash_tbl& hash_tbl)
{
    high.restore(heaps, hash_tbl);
    low.restore(heaps, hash_tbl);
}

/**************************************************************************************
    member functions for class char_range_set
***************************************************************************************/

char_range_set::char_range_set()
{
    ranges = NULL;
    num_ranges = 0;
}

char_range_set::~char_range_set()
{
    delete ranges;
}

long char_range_set::contain(unsigned char* candidate)
{
    register char_range* next_range = ranges;
    register long        i;
    
    for (i = 0; i < num_ranges; i++)
        if ((next_range++)->contains(candidate, 1))
            return(-1);
    return(0);
}

void char_range_set::init_range(unsigned long slot,
                               unsigned char* lo, unsigned char* hi,
                               char_set* for_char_set,
                               unsigned long spec_size,
                               unsigned long lo_map, unsigned long hi_map)
{
    unsigned char* lo_str;
    unsigned char* hi_str;
    unsigned long i, j;
    unsigned long char_size = for_char_set->get_char_size();

    if (spec_size > char_size) {
        printf("char_range_set::init_range: size specified greater than size of character in this char_set\n");
        return;
        }
    lo_str = (unsigned char*) (new char[char_size]);
    hi_str = (unsigned char*) (new char[char_size]);
    for (i = 0; i < char_size - spec_size; i++) {
        lo_str[i] = '\0';
        hi_str[i] = '\0';
        }
    for (j = 0; j < spec_size; i++, j++) {
        lo_str[i] = *(lo + j);
        hi_str[i] = *(hi + j);
        }
    if (MAX_CHARS_IN_SET == lo_map)
        lo_map = for_char_set->as_ulong(lo_str);
    if (MAX_CHARS_IN_SET == hi_map)
        hi_map = for_char_set->as_ulong(hi_str);
    set_range(lo_str, hi_str, for_char_set, slot, lo_map, hi_map);
    delete lo_str;
    delete hi_str;
}

inline void char_range_set::set_range(char_range* from_range, char_set* for_char_set, unsigned long slot)
{
    set_range(from_range->high.value, from_range->low.value, for_char_set, slot);
}

void char_range_set::set_range(unsigned char* lo, unsigned char* hi,
                               char_set* for_char_set,
                               unsigned long slot,
                               unsigned long lo_map, unsigned long hi_map)
{
    char_range* this_range = ranges + slot;

    this_range->low.init (lo, 1, for_char_set, lo_map);
    this_range->high.init(hi, 1, for_char_set, hi_map);
}

void char_range_set::set_range(unsigned char* lo, unsigned long lo_len,
                               unsigned char* hi, unsigned long hi_len,
                               char_set* for_char_set,
                               unsigned long slot,
                               unsigned long lo_map, unsigned long hi_map)
{
    char_range* this_range = ranges + slot;

    this_range->low.init (lo, lo_len, for_char_set, lo_map);
    this_range->high.init(hi, hi_len, for_char_set, hi_map);
}

unsigned long char_range_set::lookup(unsigned char* sought)
{
    unsigned long char_size = ranges->low.in_char_set->get_char_size();
    char_range*   mid_range;
    long          lo;
    long          hi;
    long          mid;

    char_range*    next_range;
    unsigned long  i;
    char_range*    found_range = NULL;
    unsigned long  found;
    unsigned long  low;
    unsigned char* val;

    // ranges have been sorted for logarithmic search
    if (num_ranges >= CSM_SORT_CRITERION) {
        lo = 0;
        hi = num_ranges - 1;
        // optimize integer division by 2
        mid = (lo + hi) >> 1;
        while (lo <= hi) {
            mid_range = ranges + mid;
            if (mid_range->low.gt(sought, char_size))
                hi = mid - 1;
            else if (mid_range->high.lt(sought, char_size))
                lo = mid + 1;
            else {
                found_range = mid_range;
                break;
                }
            // optimize integer division by 2
            mid = (lo + hi) >> 1;
            }
        }
    // just step through the ranges
    else {
        next_range = ranges;
        for (i = 0; i < num_ranges; next_range++, i++) {
            if (next_range->contains(sought, char_size)) {
                found_range = next_range;
                break;
                }
            }
        }

    if (found_range) {
        // range defines primary collation of an equivalence class
        //   or collation of a single character
        if (found_range->low.collation == found_range->high.collation)
            return found_range->low.collation;
        // range defines per-character collation for its members
        switch (char_size) {
            case 1:
                low = (unsigned long) *(found_range->low.value);
                found = (unsigned long) *sought;
                break;
            case 2:
                low = (unsigned long) *((unsigned short*) found_range->low.value);
                found = (unsigned long) *((unsigned short*) sought);
                break;
            case 3:
                val = found_range->low.value;
                low    =     (((unsigned long) *val) << 16)
                          |  (((unsigned long) *(val + 1)) << 8)
                          |  (((unsigned long) *(val + 2)));
                found  =     (((unsigned long) *sought) << 16)
                          |  (((unsigned long) *(sought + 1)) << 8)
                          |  (((unsigned long) *(sought + 2)));
                break;
            case 4:
                low = (unsigned long) *((unsigned long*) found_range->low.value);
                found = (unsigned long) *((unsigned long*) sought);
                break;
            default:
                printf("char_range_set::lookup(character): no range entry for character of more than 4 bytes\n");
                return MAX_CHARS_IN_SET;
            }
        return(found_range->low.collation + found - low);
        }
    return MAX_CHARS_IN_SET;
}

/* return in user-supplied storage at c the character that collates */
/*   at collation in this char_range_set */
unsigned char* char_range_set::lookup(unsigned long collation, unsigned char* c)
{
    char_range* next_range;
    unsigned long char_size = ranges->low.in_char_set->get_char_size();
    unsigned long i;

    unsigned long  low;
    unsigned char* val;

    next_range = ranges;
    for (i = 0; i < num_ranges; next_range++, i++) {
        if (next_range->contains(collation)) {
            // range defines equivalence class -- try again
            if ((next_range->low.collation == next_range->high.collation)
              && (next_range->low != next_range->high))
                break;
            switch (char_size) {
                case 1:
                    low = (unsigned long) *(next_range->low.value);
                    *c = (unsigned char) (low + (collation - next_range->low.collation));
                    return c;
                case 2:
                    low = (unsigned long) *((unsigned short*) next_range->low.value);
                    *((unsigned short*) c) = (unsigned short) (low + (collation - next_range->low.collation));
                    return c;
                case 3:
                    val = next_range->low.value;
                    low      =    (((unsigned long) *val) << 16)
                                | (((unsigned long) *(val + 1)) << 8)
                                | (((unsigned long) *(val + 2)));
                    low += (collation - next_range->low.collation);
                    *c       = (unsigned char) (low & 0x00FF0000 >> 16);
                    *(c + 1) = (unsigned char) (low & 0x0000FF00 >>  8);
                    *(c + 2) = (unsigned char) (low & 0x000000FF);
                    return c;
                case 4:
                    low = (unsigned long) *((unsigned unsigned long*) next_range->low.value);
                    *((unsigned long*) c) = low + (collation - next_range->low.collation);
                    return c;
                default:
                    printf("char_range_set::lookup(collation): no range entry for character of more than 4 bytes\n");
                    return NULL;
                }
            }
        }
    return NULL;
}

/* given two char_range_set objects and a character in the first range set */
/*   return (in the space occupied by the first character) the character */
/*   in the second char_range_set that has the same collation order as */
/*   the first_character has in its set. */
/* if the argument character is larger than 4 bytes or is not found in the */
/*   first char_range_set, the return value is NULL. */
unsigned char* char_range_set::corresponding(char_range_set& other_set, unsigned char* c)
{
    unsigned long collation = lookup(c);

    if (MAX_CHARS_IN_SET == collation) {
        printf("char_range_set::corresponding: character not in this range_set\n");
        return NULL;
        }
    else
        return(other_set.lookup(collation, c));
}

void char_range_set::save(cs_heap* heaps, gc_hash_tbl& hash_tbl)
{
    d_char_range_set* savee;
    unsigned long     i;

    savee = (d_char_range_set*) heaps->obj_alloc(sizeof(d_char_range_set));
    savee->num_ranges = num_ranges;
    for (i = 0; i < num_ranges; i++)
        (ranges + i)->save(heaps, hash_tbl);
}

void char_range_set::restore(cs_heap* heaps, gc_hash_tbl& hash_tbl)
{
    d_char_range_set* restoree;
    unsigned long     i;

    restoree = (d_char_range_set*) heaps->obj_realloc(sizeof(d_char_range_set));
    restoree->unscramble(heaps->their_endian);
    num_ranges = restoree->num_ranges;
    ranges = new char_range[num_ranges];
    for (i = 0; i < num_ranges; i++)
        (ranges + i)->restore(heaps, hash_tbl);
}

/**************************************************************************************
    member functions for class collation_map
***************************************************************************************/

collation_map::collation_map()
{
    many_to_one = NULL;
    substitute   = NULL;
    in_char_set = NULL;
}

int col_map_cmp_for_qsort(char_range* r1, char_range* r2)
{
    if (r1->low < r2->low)
        return -1;
    else if (r1->low > r2->low)
        return 1;
    return 0;
}

/* construct collation map from collecion of char_range objects */
/* The lookup functions assume that if the collation of a character in the */
/*   collation map is MAX_CHARS_IN_SET, or if there is no collation map for */
/*   the character set, then the character is self-collating according to its */
/*   (one-, two-, three- or four-byte) value. These values are determined */
/*   as (a) unextended value of single byte; (b) value of a short* at same */
/*   address as a two-byte character; (c) value of three bytes with high byte */
/*   at lowest address (note "lowest address" here means lowest with respect */
/*   to C language pointer arithmetic); (d) value of a long* at same address */
/*   as a four-byte character. Two- and four- byte characters should therefore */
/*   be unscrambled with the appropriate endian when read from disk. The */
/*   appropriate endian can be chosen through use of a magic number in a file */
/*   header; see files endian.hxx and char_set_magic.cxx for details. (Use */
/*   tml_magic.cxx instead of char_set_magic.cxx for TMLib applciations). It is */
/*   assumed that byte arrays written to and read from disk AS BYTE ARRAYS */
/*   retain their order a la UNIX (tm). All two- and four-byte characters used */
/*   in the character set manager itself are appropriately unscrambled assuming */
/*   user calls read_ucs_binary_file() appropriately to read character set */
/*   information from disk. */
/* Default argument "start_at" specifies the lowest collation for this character set. */
/* If the high and low csm_characters of a range are different, it is assumed */
/*   that a collating range is represented. If the collation of the low character */
/*   is not MAX_CHARS_IN_SET, it is assumed that the collation of the low character */
/*   is the starting collation for the range. */
/* A collating range for purposes of this function is a contiguous subset of a */
/*   character set. Contiguity here is contiguity of the numerical value of the */
/*   character determined as explained under (a), (b), (c), (d) above for characters */
/*   of one through four bytes respectively. */
/* It is an error for the number of bytes in the high and low csm_characters of a */
/*   range to differ. */
/* It is an error for the number of bytes in either the high or low csm_characters */
/*   of a range to differ from the number of bytes per character in the character set */
/*   for which the collation map is being constructed. This error is not flagged */
/*   by this function. */
/* If the high and low csm_characters of a range are the same and the number of characters */
/*   is greater than 1, it is assumed that a (multicharacter) special collating sequence */
/*   is represented. If the collation of the low character is not MAX_CHARS_IN_SET, it is */
/*   assumed that the collation of the low character is the collation of the collating */
/*   sequence. */
/* Should the low collation not be specified, the current value of the default */
/*   collation is used. */
/* Should the high collation not be specified, the current value of the default collation */
/*   plus the number of characters in the range, minus 1, is used. */
/* After each collation has been determined, a counter is advanced to the next */
/*   default collation (i.e. the one used for the start of the next collation range */
/*   or for the collation of the next special collating sequence. */
/* kmp_fsm() returns the root node of a data structure implementing the Knuth- */
/*   Morris-Pratt string search algorithm for the special collating sequences */
/*   given as its strings argument. See file kmp.cxx. */
/* This constructor initializes primary collation ranges and many_to_one members only. */
/*   To create secondary collations, call collation_map::make_secondaries. To create */
/*   one-to-many collations, call collation_map::make_one_to_many. */
collation_map::collation_map(char_range* sequences, unsigned long num_sequences,
                             char_set* for_char_set, unsigned long start_at)
{
    unsigned long   i;
    unsigned long   simple_ranges;
    unsigned long   char_size = for_char_set->char_size;
    unsigned long   num_collation_sequences;
    unsigned long   range_index;
    unsigned long   collation;
    unsigned long   coll_incr;
    unsigned long*  collations;
    char_range*     next_range;
    unsigned char** strings;

    // count simple ranges

    simple_ranges = 0;
    next_range = sequences;
    for (i = 0; i < num_sequences; i++) {
        if (next_range->low.num_chars != next_range->high.num_chars) {
            printf("collation_map::collation_map: range of differently-sized characters or strings\n");
            delete this;
            return;
            }
        if (next_range->low.num_chars == 1)
            ++simple_ranges;
        else if (next_range->low != next_range->high) {
            printf("collation_map::collation_map: strings differ in range for collation sequence\n");
            delete this;
            return;
            }
        ++next_range;
        }

    // allocate space for simple ranges

    primaries.make_ranges(simple_ranges);

    // fill in simple ranges
    // count collation sequences, initialize their collation,
    //   and move them to beginning of sequences array

    collation = start_at;
    range_index = 0;
    num_collation_sequences = 0;
    next_range = sequences;
    for (i = 0; i < num_sequences; i++) {
        // simple range
        if (next_range->low.num_chars == 1) {
            // range's start collation specified in input
            if (MAX_CHARS_IN_SET != next_range->low.collation)
                collation = next_range->low.collation;
            if (MAX_CHARS_IN_SET != next_range->high.collation)
                coll_incr = 1;
            else
                coll_incr = next_range->get_size();
            // range's end collation is calculated from size of range
            primaries.set_range(next_range->low.value, next_range->low.num_chars,
                                       next_range->high.value, next_range->high.num_chars,
                                       for_char_set,
                                       range_index++,
                                       collation,
                                       collation + coll_incr - 1);
            collation += coll_incr;
            }
        // collation sequence
        else {
            // bit copy
            *(sequences + num_collation_sequences) = *next_range;
            // sequence's collation specified in input
            if (MAX_CHARS_IN_SET != next_range->low.collation)
                collation = next_range->low.collation;
            (sequences + num_collation_sequences)->low.collation = collation;
            ++collation;
            ++num_collation_sequences;
            }
        ++next_range;
        }

    // if sufficiently many simple ranges, sort them for logarithmic search

    if (simple_ranges >= CSM_SORT_CRITERION)
        tml_os_qsort((char*) primaries.get_ranges(), simple_ranges, sizeof(char_range), (SORT_FUNC) col_map_cmp_for_qsort);

    // construct Knuth-Morris-Pratt finite state machine

    if (num_collation_sequences) {
        strings = new unsigned char*[num_collation_sequences];
        collations = new unsigned long[num_collation_sequences];
        next_range = sequences;
        for (i = 0; i < num_collation_sequences; i++) {
            *(strings + i) = next_range->low.value;
            *(collations + i) = next_range->low.collation;
            ++next_range;
            }
        many_to_one = kmp_fsm(strings, num_collation_sequences, collations, for_char_set);
        delete strings;
        delete collations;
        }
    else 
        many_to_one = NULL;
    substitute  = NULL;
    in_char_set = for_char_set;
}

collation_map::~collation_map()
{
    delete many_to_one;
    many_to_one = NULL;
    delete substitute;
    substitute = NULL;
    in_char_set = NULL;
}

void collation_map::make_secondaries(char_range* sequences, unsigned long num_sequences)
{
    unsigned long i;

    // if sufficienctly many secondaries, sort them for logarithmic search
    if (num_sequences >= CSM_SORT_CRITERION)
        tml_os_qsort((char*) sequences, num_sequences, sizeof(char_range), (SORT_FUNC) col_map_cmp_for_qsort);

    secondaries.make_ranges(num_sequences);
    for (i = 0; i < num_sequences; i++) {
        secondaries.set_range(sequences->low.value, sequences->low.num_chars,
                              sequences->high.value, sequences->high.num_chars,
                              in_char_set,
                              i,
                              sequences->low.collation,
                              sequences->high.collation);
        ++sequences;
        }
}

void collation_map::make_one_to_many(char_range* sequences, unsigned long num_sequences)
{
    unsigned long i;

    // if sufficiently many one-to-many specifications, sort them for logarithmic search
    if (num_sequences >= CSM_SORT_CRITERION)
        tml_os_qsort((char*) sequences, num_sequences, sizeof(char_range), (SORT_FUNC) col_map_cmp_for_qsort);

    one_to_many.make_ranges(num_sequences);
    for (i = 0; i < num_sequences; i++) {
        one_to_many.set_range(sequences->low.value, sequences->low.num_chars,
                              sequences->high.value, sequences->high.num_chars,
                              in_char_set,
                              i,
                              sequences->low.collation,
                              sequences->high.collation);
        ++sequences;
        }
}

void collation_map::make_substitutes(char** find, char** sub, unsigned long num_subs)
{
}

/* returns MAX_CHARS_IN_SET if character not found -- means */
/*   character is to be ignored in sorting */
unsigned long collation_map::lookup(unsigned char*& sought)
{
    unsigned char* advancee = sought;
    unsigned long  retval;

    if (many_to_one) {
        // many_to_one lookup if successful does its own advancing
        if (MAX_CHARS_IN_SET == (retval = many_to_one->lookup(advancee))) {
            in_char_set->advance(sought);
            return primaries.lookup(advancee);
            }
        else {
            sought = advancee;
            return retval;
            }
        }
    else {
        in_char_set->advance(sought);
        return primaries.lookup(advancee);
        }
}

long collation_map::compare(unsigned char* s1, unsigned char* s2)
{
    unsigned char* substr1;
    unsigned char* substr2;
    unsigned char* last_s1;
    unsigned char* last_s2;
    unsigned long  sublen1;
    unsigned long  sublen2;
    unsigned long  cmplen;
    char_range*    otm_ranges;
    char_range*    next_range;
    long           cmp;
    unsigned long  num_otm_ranges;
    unsigned long  char_size = (in_char_set) ? in_char_set->get_char_size() : 1;
    unsigned long  i;
    unsigned long  to_advance;
    register unsigned long  colpri_1;
    register unsigned long  colpri_2;
    register unsigned long  colsec_1;
    register unsigned long  colsec_2;
    unsigned long  wassec_1 = MAX_CHARS_IN_SET;
    unsigned long  wassec_2 = MAX_CHARS_IN_SET;
    unsigned long  one_is_null = in_char_set->is_null(s1);
    unsigned long  two_is_null = in_char_set->is_null(s2);

    if (num_otm_ranges = one_to_many.get_num_ranges())
        otm_ranges = one_to_many.get_ranges();
    while (! (one_is_null && two_is_null)) {
        if (two_is_null)
            return 1;
        else if (one_is_null)
            return -1;
        if (num_otm_ranges) {
            substr1 = NULL;
            substr2 = NULL;
            cmplen = 0;
            for (i = 0, next_range = otm_ranges; i < num_otm_ranges; i++, next_range++) {
                if (next_range->low.eq(s1, char_size)) {
                    substr1 = next_range->high.value;
                    sublen1 = next_range->high.num_chars;
                    }
                if (next_range->low.eq(s2, char_size)) {
                    substr2 = next_range->high.value;
                    sublen2 = next_range->high.num_chars;
                    }
                }
            if ((NULL == substr1) && (NULL == substr2)) {
                // collation_map::lookup(ptr) advances ptr to next
                //   collating character/sequence
                last_s1 = s1;
                last_s2 = s2;
                colpri_1 = lookup(s1);
                colpri_2 = lookup(s2);
                // characters at both last_s1 and last_s2 are to be ignored in sorting
                if ((MAX_CHARS_IN_SET == colpri_1) && (MAX_CHARS_IN_SET == colpri_2))
                    continue;
                // character at last_s1 is to be ignored in sorting
                else if (MAX_CHARS_IN_SET == colpri_1) {
                    s2 = last_s2;
                    continue;
                    }
                // character at last_s2 is to be ignored in sorting
                else if (MAX_CHARS_IN_SET == colpri_2) {
                    s1 = last_s1;
                    continue;
                    }
                }
            else {
                if (NULL == substr1)
                    sublen1 = char_size;
                if (NULL == substr2)
                    sublen2 = char_size;
                cmplen = (sublen1 > sublen2) ? sublen1 : sublen2;
                if (NULL == substr1) {
                    substr1 = s1;
                    to_advance = cmplen;
                    while (to_advance-- && (! in_char_set->is_null(s1)))
                        in_char_set->advance(s1);
                    }
                else
                    in_char_set->advance(s1);
                if (NULL == substr2) {
                    substr2 = s2;
                    to_advance = cmplen;
                    while (to_advance-- && (! in_char_set->is_null(s2)))
                        in_char_set->advance(s2);
                    }
                else
                    in_char_set->advance(s2);
                cmp = ncompare(substr1, substr2, cmplen);
                if (cmp > 0)
                    return 1;
                else if (cmp < 0)
                    return -1;
                else
                    colpri_1 = colpri_2 = MAX_CHARS_IN_SET;
                substr1 = NULL;
                substr2 = NULL;
                cmplen = 0;
                }
            }
        else {
            // collation_map::lookup(ptr) advances ptr to next
            //   collating character/sequence
            last_s1 = s1;
            last_s2 = s2;
            colpri_1 = lookup(s1);
            colpri_2 = lookup(s2);
            // characters at both last_s1 and last_s2 are to be ignored in sorting
            if ((MAX_CHARS_IN_SET == colpri_1) && (MAX_CHARS_IN_SET == colpri_2))
                continue;
            // character at last_s1 is to be ignored in sorting
            else if (MAX_CHARS_IN_SET == colpri_1) {
                s2 = last_s2;
                continue;
                }
            // character at last_s2 is to be ignored in sorting
            else if (MAX_CHARS_IN_SET == colpri_2) {
                s1 = last_s1;
                continue;
                }
            }
        if (colpri_1 > colpri_2)
            return 1;
        else if (colpri_2 > colpri_1)
            return -1;
        else if (secondaries.get_num_ranges()) {
            // -- either both or neither is MAX_CHARS_IN_SET
            if (MAX_CHARS_IN_SET == (colsec_1 = secondaries.lookup(last_s1)))
                continue;
            colsec_2 = secondaries.lookup(last_s2);
            // if secondaries differ, record this comparison as tie-breaker
            if (colsec_1 != colsec_2) {
                wassec_1 = colsec_1;
                wassec_2 = colsec_2;
                }
            }
        one_is_null = in_char_set->is_null(s1);
        two_is_null = in_char_set->is_null(s2);
        }
    if (wassec_1 > wassec_2)
        return 1;
    else if (wassec_2 > wassec_1)
        return -1;
    return 0;
}

long collation_map::ncompare(unsigned char* s1, unsigned char* s2, unsigned long n)
{
    unsigned char* substr1;
    unsigned char* substr2;
    unsigned char* last_s1;
    unsigned char* last_s2;
    unsigned long  sublen1;
    unsigned long  sublen2;
    unsigned long  cmplen;
    char_range*    otm_ranges;
    char_range*    next_range;
    long           cmp;
    unsigned long  num_otm_ranges;
    unsigned long  char_size = (in_char_set) ? in_char_set->get_char_size() : 1;
    unsigned long  i;
    unsigned long  to_advance;
    unsigned long  colpri_1;
    unsigned long  colpri_2;
    unsigned long  colsec_1;
    unsigned long  colsec_2;
    unsigned long  wassec_1 = MAX_CHARS_IN_SET;
    unsigned long  wassec_2 = MAX_CHARS_IN_SET;
    unsigned long  one_is_null = in_char_set->is_null(s1);
    unsigned long  two_is_null = in_char_set->is_null(s2);

    if (num_otm_ranges = one_to_many.get_num_ranges())
        otm_ranges = one_to_many.get_ranges();
    while (n--) {
        if (one_is_null && two_is_null)
            break;
        else if (two_is_null)
            return 1;
        else if (one_is_null)
            return -1;
        if (num_otm_ranges) {
            substr1 = NULL;
            substr2 = NULL;
            cmplen = 0;
            for (i = 0, next_range = otm_ranges; i < num_otm_ranges; i++, next_range++) {
                if (next_range->low.eq(s1, char_size)) {
                    substr1 = next_range->high.value;
                    sublen1 = next_range->high.num_chars;
                    }
                if (next_range->low.eq(s2, char_size)) {
                    substr2 = next_range->high.value;
                    sublen2 = next_range->high.num_chars;
                    }
                }
            if ((NULL == substr1) && (NULL == substr2)) {
                // collation_map::lookup(ptr) advances ptr to next
                //   collating character/sequence
                last_s1 = s1;
                last_s2 = s2;
                colpri_1 = lookup(s1);
                colpri_2 = lookup(s2);
                // characters at both last_s1 and last_s2 are to be ignored in sorting
                if ((MAX_CHARS_IN_SET == colpri_1) && (MAX_CHARS_IN_SET == colpri_2))
                    continue;
                // character at last_s1 is to be ignored in sorting
                else if (MAX_CHARS_IN_SET == colpri_1) {
                    s2 = last_s2;
                    continue;
                    }
                // character at last_s2 is to be ignored in sorting
                else if (MAX_CHARS_IN_SET == colpri_2) {
                    s1 = last_s1;
                    continue;
                    }
                }
            else {
                if (NULL == substr1)
                    sublen1 = char_size;
                if (NULL == substr2)
                    sublen2 = char_size;
                cmplen = (sublen1 > sublen2) ? sublen1 : sublen2;
                if (NULL == substr1) {
                    substr1 = s1;
                    to_advance = cmplen;
                    while (to_advance-- && (! in_char_set->is_null(s1)))
                        in_char_set->advance(s1);
                    }
                else
                    in_char_set->advance(s1);
                if (NULL == substr2) {
                    substr2 = s2;
                    to_advance = cmplen;
                    while (to_advance-- && (! in_char_set->is_null(s2)))
                        in_char_set->advance(s2);
                    }
                else
                    in_char_set->advance(s2);
                cmp = ncompare(substr1, substr2, cmplen);
                if (cmp > 0)
                    return 1;
                else if (cmp < 0)
                    return -1;
                else
                    colpri_1 = colpri_2 = MAX_CHARS_IN_SET;
                substr1 = NULL;
                substr2 = NULL;
                cmplen = 0;
                }
            }
        else {
            // collation_map::lookup(ptr) advances ptr to next
            //   collating character/sequence
            last_s1 = s1;
            last_s2 = s2;
            colpri_1 = lookup(s1);
            colpri_2 = lookup(s2);
            // characters at both last_s1 and last_s2 are to be ignored in sorting
            if ((MAX_CHARS_IN_SET == colpri_1) && (MAX_CHARS_IN_SET == colpri_2))
                continue;
            // character at last_s1 is to be ignored in sorting
            else if (MAX_CHARS_IN_SET == colpri_1) {
                s2 = last_s2;
                continue;
                }
            // character at last_s2 is to be ignored in sorting
            else if (MAX_CHARS_IN_SET == colpri_2) {
                s1 = last_s1;
                continue;
                }
            }
        if (colpri_1 > colpri_2)
            return 1;
        else if (colpri_2 > colpri_1)
            return -1;
        else if (secondaries.get_num_ranges()) {
            // -- either both or neither is MAX_CHARS_IN_SET
            if (MAX_CHARS_IN_SET == (colsec_1 = secondaries.lookup(s1)))
                continue;
            colsec_2 = secondaries.lookup(s2);
            // if secondaries differ, record this comparison as tie-breaker
            if (colsec_1 != colsec_2) {
                wassec_1 = colsec_1;
                wassec_2 = colsec_2;
                }
            }
        one_is_null = in_char_set->is_null(s1);
        two_is_null = in_char_set->is_null(s2);
        }
    if (wassec_1 > wassec_2)
        return 1;
    else if (wassec_2 > wassec_1)
        return -1;
    return 0;
}

void collation_map::save(cs_heap* heaps, gc_hash_tbl& hash_tbl)
{
    d_collation_map* savee;
    d_ptr            d_ptr_tmp;
    gc_hash_elem*    hash_elem;

    void kmp_tree_save(kmp_node* root, cs_heap* heaps, gc_hash_tbl& hash_tbl);

    if (NULL != hash_tbl.find(this))  //already saved
        return;
    hash_elem = hash_tbl.insert(this);
    d_ptr_tmp = (void*) (  LAST_DEFAULT_HASH +
                           heaps->header.num_unions +
                         ++heaps->header.num_maps +
                           heaps->header.num_kmp_nodes +
                           heaps->num_char_sets);
    hash_elem->set_new((void*) d_ptr_tmp.as_whole_ptr());
    savee = (d_collation_map*) heaps->obj_alloc(sizeof(d_collation_map));
    if (NULL == many_to_one) {
        d_ptr_tmp.get_d_ptr(NULL, NULL);
        savee->many_to_one = (void*) d_ptr_tmp.as_whole_ptr();
        }
    else {
        kmp_tree_save(many_to_one, heaps, hash_tbl);
        savee->many_to_one = hash_tbl.find(many_to_one);
        }
    primaries.save(heaps, hash_tbl);
    secondaries.save(heaps, hash_tbl);
    one_to_many.save(heaps, hash_tbl);
}

void collation_map::restore(cs_heap* heaps, gc_hash_tbl& hash_tbl)
{
    d_collation_map* restoree;
    gc_hash_elem*    hash_elem;
    d_ptr            d_ptr_tmp;
    kmp_node*        root;

    void kmp_tree_restore(kmp_node*& root, cs_heap* heaps, gc_hash_tbl& hash_tbl);

    d_ptr_tmp = (void*) (  LAST_DEFAULT_HASH +
                           heaps->header.num_unions +
                         ++heaps->header.num_maps +
                           heaps->header.num_kmp_nodes +
                           heaps->num_char_sets);
    hash_elem = hash_tbl.insert((void*) d_ptr_tmp.as_whole_ptr());
    hash_elem->set_new((void*) this);
    restoree = (d_collation_map*) heaps->obj_realloc(sizeof(d_collation_map));

    restoree->unscramble(heaps->their_endian);
    in_char_set = heaps->get_cur_char_set();
    d_ptr_tmp = restoree->many_to_one;
    if (d_ptr_tmp.is_null())
        many_to_one = NULL;
    else if (NULL ==
      (many_to_one = (kmp_node*) hash_tbl.find((void*) d_ptr_tmp.as_whole_ptr()))) {
        kmp_tree_restore(root, heaps, hash_tbl);
        many_to_one = root;
        }
    primaries.restore(heaps, hash_tbl);
    secondaries.restore(heaps, hash_tbl);
    one_to_many.restore(heaps, hash_tbl);
}

/**************************************************************************************
    global set of default unix escape characters
**************************************************************************************/

// This global is never written, so there is no need to add an attribute specifier 
//   for Apollo builds.
unsigned char* tml_unix_esc_chars[num_esc_chars] = {(unsigned char*)"\x07",
                                                    (unsigned char*)"\x08",
                                                    (unsigned char*)"\x0C",
                                                    (unsigned char*)"\x0A",
                                                    (unsigned char*)"\x0D",
                                                    (unsigned char*)"\x09",
                                                    (unsigned char*)"\x0B",
                                                    (unsigned char*) "\x5C",
                                                    (unsigned char*) "\x3F",
                                                    (unsigned char*) "\x27",
                                                    (unsigned char*) "\x22"};

/**************************************************************************************
    global sets of string functions
    one_byte characters may be aligned at odd or even byte addresses
    two-byte characters are assumed to be aligned at 16-bit boundaries
    three-byte characters may be aligned at odd or even byte addresses
    four-byte characters are assumed to be aligned at 32-bit boundaries
**************************************************************************************/

class three_byte {
    unsigned char* cp; // use pointer rather than data members to avoid structure padding hassles

    unsigned char dollar1(){return(*cp);}
    unsigned char dollar2(){return(*(cp + 1));}
    unsigned char dollar3(){return(*(cp + 2));}

public:
    three_byte::three_byte(unsigned char* p = 0){cp = p;}
    unsigned char* get_cp(){return cp;}
    long           operator==(three_byte with){return cp == with.cp;}
    void           operator++(){cp += 3;}
    void           operator--(){cp -= 3;}
    long           operator&(three_byte with){return ((dollar1() & with.dollar1()) & (dollar2() & with.dollar2()) &  (dollar3() & with.dollar3()));}
    long           eq(three_byte to){return ((dollar1() == to.dollar1()) && (dollar2() == to.dollar2()) && (dollar3() == to.dollar3())) ? -1 : 0;}
    long           ne(three_byte to){return ((dollar1() != to.dollar1()) || (dollar2() != to.dollar2()) || (dollar3() != to.dollar3())) ? -1 : 0;}
    long           lt(three_byte to){return (((dollar1() <  to.dollar1()) ? -1 : (dollar2() <  to.dollar2())) ? -1 : (dollar3() <  to.dollar3())) ? -1 :  0;}
    long           le(three_byte to){return (((dollar1() >  to.dollar1()) ?  0 : (dollar2() >  to.dollar2())) ?  0 : (dollar3() >  to.dollar3())) ?  0 : -1;}
    long           gt(three_byte to){return (((dollar1() >  to.dollar1()) ? -1 : (dollar2() >  to.dollar2())) ? -1 : (dollar3() >  to.dollar3())) ? -1 :  0;}
    long           ge(three_byte to){return (((dollar1() <  to.dollar1()) ?  0 : (dollar2() <  to.dollar2())) ?  0 : (dollar3() <  to.dollar3())) ?  0 : -1;}
    void           set_equal(three_byte to){*cp = to.dollar1();  *(cp + 1) = to.dollar2(); *(cp + 2) = to.dollar3();}
    void           set_zero(){*cp = '\0'; *(cp + 1) = '\0'; *(cp + 2) = '\0';}
    void           set_null(char_set* for_char_set = 0);
    long           is_zero(){return (('\0' == dollar1()) || ('\0' == dollar2()) || ('\0' == dollar3()));}
    long           is_null (char_set* for_char_set = 0);
    long           non_zero(){return (('\0' != dollar1()) || ('\0' != dollar2()) || ('\0' != dollar3()));}
    long           non_null(char_set* for_char_set = 0);
};

inline void three_byte::set_null(char_set* for_char_set)
{
    if (for_char_set)
        for_char_set->get_terminator(cp);
    else
        set_zero();
}

inline long three_byte::is_null(char_set* for_char_set)
{
    if (for_char_set)
        return(for_char_set->is_null(cp));
    else
        return is_zero();
}

inline long three_byte::non_null(char_set* for_char_set)
{
    if (for_char_set)
        return(! for_char_set->is_null(cp));
    else
        return non_zero();
}

unsigned char* strcpy8  (unsigned char*, unsigned char*, char_set* for_char_set = 0);
unsigned char* strncpy8 (unsigned char*, unsigned char*, long, char_set* for_char_set = 0);
long  strlen8  (unsigned char*, char_set* for_char_set = 0);

unsigned char* strcpy16 (unsigned char*, unsigned char*, char_set* for_char_set = 0);
unsigned char* strncpy16(unsigned char*, unsigned char*, long, char_set* for_char_set = 0);
long  strlen16 (unsigned char*, char_set* for_char_set = 0);

unsigned char* strcpy24 (unsigned char*, unsigned char*, char_set* for_char_set = 0);
unsigned char* strncpy24(unsigned char*, unsigned char*, long, char_set* for_char_set = 0);
long  strlen24 (unsigned char*, char_set* for_char_set = 0);

unsigned char* strcpy32 (unsigned char*, unsigned char*, char_set* for_char_set = 0);
unsigned char* strncpy32(unsigned char*, unsigned char*, long, char_set* for_char_set = 0);
long  strlen32 (unsigned char*, char_set* for_char_set = 0);

unsigned char* strcat8(unsigned char* s1, unsigned char* s2, char_set* for_char_set = 0)
{
    if ((NULL == for_char_set) || (for_char_set->zero_term()))
        return (unsigned char*) (::strcat((char*) s1, (char*) s2));
    return strcpy8(s1 + strlen8(s1, for_char_set), s2, for_char_set);
}

unsigned char* strcat16(unsigned char* s1, unsigned char* s2, char_set* for_char_set = 0)
{
    return strcpy16(s1 + (strlen16(s1, for_char_set) << 1), s2, for_char_set);
}

unsigned char* strcat24(unsigned char* s1, unsigned char* s2, char_set* for_char_set = 0)
{
    unsigned long len1 = strlen24(s1, for_char_set);

    return strcpy24(s1 + (len1 << 1) + len1, s2, for_char_set);
}

unsigned char* strcat32(unsigned char* s1, unsigned char* s2, char_set* for_char_set = 0)
{
    return strcpy32(s1 + (strlen32(s1, for_char_set) << 2), s2, for_char_set);
}

unsigned char* strncat8(unsigned char* s1, unsigned char* s2, long n, char_set* for_char_set = 0)
{
    long len1;
    long len2;
    long len3;

    if ((NULL == for_char_set) || (for_char_set->zero_term()))
        return (unsigned char*) (::strncat((char*)s1, (char*) s2, (int) n));
    len1 = strlen8(s1, for_char_set);
    len2 = strlen8(s2, for_char_set);
    len3 = (len2 < n) ? len2 : n;
    strncpy8(s1 + len1, s2, len3, for_char_set);
    for_char_set->get_terminator(s1 + len1 + len3);
    return s1;
}

unsigned char* strncat16(unsigned char* s1, unsigned char* s2, long n, char_set* for_char_set = 0)
{
    long len1 = strlen16(s1, for_char_set);
    long len2 = strlen16(s2, for_char_set);
    long len3 = (len2 < n) ? len2 : n;

    if ((NULL == for_char_set) || (for_char_set->zero_term())) {
        strncpy16(s1 + (len1 << 1), s2, len3, for_char_set);
        *(((unsigned short*) s1) + len1 + len3) = 0;
        return s1;
        }
    strncpy16(s1 + (len1 << 1), s2, len3, for_char_set);
    for_char_set->get_terminator(s1 + (len1 << 1) + (len3 << 1));
    return s1;
}

unsigned char* strncat24(unsigned char* s1, unsigned char* s2, long n, char_set* for_char_set = 0)
{
    long len1 = strlen24(s1, for_char_set);
    long len2 = strlen24(s2, for_char_set);
    long len3 = (len2 < n) ? len2 : n;
    long len4 = len1 + len3;
    three_byte tb;

    strncpy24(s1 + ((len1 << 1) + len1), s2, len3, for_char_set);
    tb = s1 + len4 + (len4 << 1);
    tb.set_null(for_char_set);
    return s1;
}

unsigned char* strncat32(unsigned char* s1, unsigned char* s2, long n, char_set* for_char_set = 0)
{
    long len1 = strlen32(s1, for_char_set);
    long len2 = strlen32(s2, for_char_set);
    long len3 = (len2 < n) ? len2 : n;

    if ((NULL == for_char_set) || (for_char_set->zero_term())) {
        strncpy32(s1 + (len1 << 2), s2, len3, for_char_set);
        *(((unsigned long*) s1) + len1 + len3) = 0;
        return s1;
        }
    strncpy32(s1 + (len1 << 2), s2, len3, for_char_set);
    for_char_set->get_terminator(s1 + (len1 << 2) + (len3 << 2));
    return s1;
}

long strcmp8(unsigned char* s1, unsigned char* s2, char_set* for_char_set = 0)
{
    if ((NULL == for_char_set) || ((for_char_set->zero_term() && for_char_set->hibit_is_clear())))
        return ::strcmp((char*) s1,(char*)  s2);
    if (for_char_set->zero_term()) {
        do {
            if (*s1 == '\0') {
                if (*s2 == '\0')
                    return 0;
                return -1;
                }
            else if (*s2 == '\0')
                return 1;
            else if (*s1 > *s2)
                return 1;
            else if (*s1 < *s2)
                return -1;
            ++s1; ++s2;
            } while (1);
        }
    else {
        do {
            if (for_char_set->is_null(s1)) {
                if (for_char_set->is_null(s2))
                    return 0;
                return -1;
                }
            else if (for_char_set->is_null(s2))
                return 1;
            else if (*s1 > *s2)
                return 1;
            else if (*s1 < *s2)
                return -1;
            ++s1; ++s2;
            } while (1);
        }
}

long strcmp16(unsigned char* s1, unsigned char* s2, char_set* for_char_set = 0)
{
    register unsigned short* sp1 = (unsigned short*) s1;
    register unsigned short* sp2 = (unsigned short*) s2;

    if ((NULL == for_char_set) || (for_char_set->zero_term())) {
        do {
            if ((0 == *sp1) && (0 == *sp2))
                return 0;
            else if (*sp1 > *sp2)
                return 1;
            else if (*sp1 < *sp2)
                return -1;
            ++sp1; ++sp2;
            } while (1);
        }
    do {
        if (for_char_set->is_null((unsigned char*) sp1)) {
            if (for_char_set->is_null((unsigned char*) sp2))
                return 0;
            return -1;
            }
        else if (for_char_set->is_null((unsigned char*) sp2))
            return 1;
        else if (*sp1 > *sp2)
            return 1;
        else if (*sp1 < *sp2)
            return -1;
        ++sp1; ++sp2;
        } while (1);
}

long strcmp24(unsigned char* s1, unsigned char* s2, char_set* for_char_set = 0)
{
    three_byte tb1 = s1;
    three_byte tb2 = s2;

    do {
        if ((tb1.is_null(for_char_set)) && (tb2.is_null(for_char_set)))
            return 0;
        else if (tb1.gt(tb2))
            return 1;
        else if (tb1.lt(tb2))
            return -1;
        ++tb1; ++tb2;
        } while (1);
}

long strcmp32(unsigned char* s1, unsigned char* s2, char_set* for_char_set = 0)
{
    register unsigned long * lp1 = (unsigned long*) s1;
    register unsigned long * lp2 = (unsigned long*) s2;

    if ((NULL == for_char_set) || (for_char_set->zero_term())) {
        do {
            if ((0 == *lp1) && (0 == *lp2))
                return 0;
            else if (*lp1 > *lp2)
                return 1;
            else if (*lp1 < *lp2)
                return -1;
            ++lp1; ++lp2;
            } while (1);
        }
    do {
        if (for_char_set->is_null((unsigned char*) lp1)) {
            if (for_char_set->is_null((unsigned char*) lp2))
                return 0;
            return -1;
            }
        else if (for_char_set->is_null((unsigned char*) lp2))
            return 1;
        else if (*lp1 > *lp2)
            return 1;
        else if (*lp1 < *lp2)
            return -1;
        ++lp1; ++lp2;
        } while (1);
}

long strncmp8(unsigned char* s1, unsigned char* s2, long n, char_set* for_char_set = 0)
{
    if ((NULL == for_char_set) || ((for_char_set->zero_term() && for_char_set->hibit_is_clear())))
        return ::strncmp((char*) s1, (char*) s2, (int) n);
    if (for_char_set->zero_term()) {
        while (n--) {
            if (*s1 == '\0') {
                if (*s2 == '\0')
                    return 0;
                return -1;
                }
            else if (*s2 == '\0')
                return 1;
            else if (*s1 > *s2)
                return 1;
            else if (*s1 < *s2)
                return -1;
            ++s1; ++s2;
            }
        return 0;
        }
    while (n--) {
        if (for_char_set->is_null(s1)) {
            if (for_char_set->is_null(s2))
                return 0;
            return -1;
            }
        else if (for_char_set->is_null(s2))
            return 1;
        else if (*s1 > *s2)
            return 1;
        else if (*s1 < *s2)
            return -1;
        ++s1; ++s2;
        }
    return 0;
}

long strncmp16(unsigned char* s1, unsigned char* s2, long n, char_set* for_char_set = 0)
{
    register unsigned short* sp1 = (unsigned short*) s1;
    register unsigned short* sp2 = (unsigned short*) s2;

    if ((NULL == for_char_set) || (for_char_set->zero_term())) {
        while (n--) {
            if ((0 == *sp1) && (0 == *sp2))
                return 0;
            else if (*sp1 > *sp2)
                return 1;
            else if (*sp1 < *sp2)
                return -1;
            ++sp1; ++sp2;
            }
        return 0;
        }
    while (n--) {
        if (for_char_set->is_null((unsigned char*) sp1)) {
            if (for_char_set->is_null((unsigned char*) sp2))
                return 0;
            return -1;
            }
        else if (for_char_set->is_null((unsigned char*) sp2))
            return 1;
        else if (*sp1 > *sp2)
            return 1;
        else if (*sp1 < *sp2)
            return -1;
        ++sp1; ++sp2;
        }
    return 0;
}

long strncmp24(unsigned char* s1, unsigned char* s2, long n, char_set* for_char_set = 0)
{
    three_byte tb1 = s1;
    three_byte tb2 = s2;

    while (n--) {
        if ((tb1.is_null(for_char_set)) && (tb2.is_null(for_char_set)))
            return 0;
        else if (tb1.gt(tb2))
            return 1;
        else if (tb1.lt(tb2))
            return -1;
        ++tb1; ++tb2;
        }
    return 0;
}

long strncmp32(unsigned char* s1, unsigned char* s2, long n, char_set* for_char_set = 0)
{
    register unsigned long * lp1 = (unsigned long*) s1;
    register unsigned long * lp2 = (unsigned long*) s2;

    if ((NULL == for_char_set) || (for_char_set->zero_term())) {
        while (n--) {
            if ((0 == *lp1) && (0 == *lp2))
                return 0;
            else if (*lp1 > *lp2)
                return 1;
            else if (*lp1 < *lp2)
                return -1;
            ++lp1; ++lp2;
            }
        return 0;
        }
    while (n--) {
        if (for_char_set->is_null((unsigned char*) lp1)) {
            if (for_char_set->is_null((unsigned char*) lp2))
                return 0;
            return -1;
            }
        else if (for_char_set->is_null((unsigned char*) lp2))
            return 1;
        else if (*lp1 > *lp2)
            return 1;
        else if (*lp1 < *lp2)
            return -1;
        ++lp1; ++lp2;
        }
    return 0;
}

unsigned char* strcpy8(unsigned char* s1, unsigned char* s2, char_set* for_char_set)
{
    register unsigned char* s3;

    if ((NULL == for_char_set) || (for_char_set->zero_term()))
        return (unsigned char*) (::strcpy((char*) s1, (char*) s2));
    s3 = s1;
    while (!for_char_set->is_null(s2))
        *s3++ = *s2++;
    for_char_set->get_terminator(s3);
    return s1;
}

unsigned char* strcpy16(unsigned char* s1, unsigned char* s2, char_set* for_char_set)
{
    register unsigned short* sp1 = (unsigned short*) s1;
    register unsigned short* sp2 = (unsigned short*) s2;

    if ((NULL == for_char_set) || (for_char_set->zero_term())) {
        while (*sp2)
            *sp1++ = *sp2++;
        *sp1 = (unsigned short) 0;
        return s1;
        }
    while (!for_char_set->is_null((unsigned char*) sp2))
        *sp1++ = *sp2++;
    for_char_set->get_terminator((unsigned char*) sp1);
    return s1;
}


unsigned char* strcpy24(unsigned char* s1, unsigned char* s2, char_set* for_char_set)
{
    three_byte tb1 = s1;
    three_byte tb2 = s2;

    while (tb2.non_null(for_char_set)) {
        tb1.set_equal(tb2);
        ++tb1; ++tb2;
        }
    tb1.set_null(for_char_set);
    return s1;
}

unsigned char* strcpy32(unsigned char* s1, unsigned char* s2, char_set* for_char_set)
{
    register unsigned long * lp1 = (unsigned long*) s1;
    register unsigned long * lp2 = (unsigned long*) s2;

    if ((NULL == for_char_set) || (for_char_set->zero_term())) {
        while (*lp2)
            *lp1++ = *lp2++;
        *lp1 = (unsigned long) 0;
        return s1;
        }
    while (!for_char_set->is_null((unsigned char*) lp2))
        *lp1++ = *lp2++;
    for_char_set->get_terminator((unsigned char*) lp1);
    return s1;
}

unsigned char* strncpy8(unsigned char* s1, unsigned char* s2, long n, char_set* for_char_set)
{
    register unsigned char* s3 = s1;

    if ((NULL == for_char_set) || (for_char_set->zero_term()))
        return (unsigned char*) (::strncpy((char*)s1, (char*)s2, (int) n));
    while (n && (!for_char_set->is_null(s2))) {
        *s3++ = *s2++;
        --n;
        }
    while (n > 0) {
        for_char_set->get_terminator(s3++);
        --n;
        }
    return s1;
}

unsigned char* strncpy16(unsigned char* s1, unsigned char* s2, long n, char_set* for_char_set)
{
    register unsigned short* sp1 = (unsigned short*) s1;
    register unsigned short* sp2 = (unsigned short*) s2;

    if ((NULL == for_char_set) || (for_char_set->zero_term())) {
        while (n && *sp2) {
            *sp1++ = *sp2++;
            --n;
            }
        while (n > 0) {
            *sp1++ = (unsigned short) 0;
            --n;
            }
        return s1;
        }
    while (n && (!for_char_set->is_null((unsigned char*) sp2))) {
        *sp1++ = *sp2++;
        --n;
        }
    while (n > 0) {
        for_char_set->get_terminator((unsigned char*) sp1++);
        --n;
        }
    return s1;
}

unsigned char* strncpy24(unsigned char* s1, unsigned char* s2, long n, char_set* for_char_set)
{
    three_byte tb1 = s1;
    three_byte tb2 = s2;

    while (n && (tb2.non_null(for_char_set))) {
        tb1.set_equal(tb2);
        ++tb1; ++tb2;
        --n;
        }
    while (n > 0) {
        tb1.set_null(for_char_set);
        ++tb1;
        --n;
        }
    return s1;
}

unsigned char* strncpy32(unsigned char* s1, unsigned char* s2, long n, char_set* for_char_set)
{
    register unsigned long * lp1 = (unsigned long*) s1;
    register unsigned long * lp2 = (unsigned long*) s2;

    if ((NULL == for_char_set) || (for_char_set->zero_term())) {
        while (n && *lp2) {
            *lp1++ = *lp2++;
            --n;
            }
        while (n > 0) {
            *lp1++ = (unsigned long) 0;
            --n;
            }
        return s1;
        }
    while (n && !(for_char_set->is_null((unsigned char*) lp2))) {
        *lp1++ = *lp2++;
        --n;
        }
    while (n > 0) {
        for_char_set->get_terminator((unsigned char*) lp1++);
        --n;
        }
    return s1;
}

long strlen8(unsigned char* s, char_set* for_char_set)
{
    register long retval = 0;

    if ((NULL == for_char_set) || (for_char_set->zero_term()))
        return ::strlen((char*)s);
    while (! for_char_set->is_null(s++))
        ++retval;
    return retval;
}

long strlen16(unsigned char* s, char_set* for_char_set)
{
    register unsigned short* sp = (unsigned short*) s;
    register long retval = 0;

    if ((NULL == for_char_set) || (for_char_set->zero_term())) {
        while (*sp++)
            ++retval;
        return retval;
        }
    while (! for_char_set->is_null((unsigned char*) sp++))
        ++retval;
    return retval;
}

long strlen24(unsigned char* s, char_set* for_char_set)
{
    three_byte tb = s;
    register long retval = 0;

    while(tb.non_null(for_char_set)) {
        ++retval;
        ++tb;
        }
    return retval;
}

long strlen32(unsigned char* s, char_set* for_char_set)
{
    register unsigned long* lp = (unsigned long*) s;
    register long retval = 0;

    if ((NULL == for_char_set) || (for_char_set->zero_term())) {
        while (*lp++)
            ++retval;
        return retval;
        }
    while (! for_char_set->is_null((unsigned char*) lp++))
        ++retval;
    return retval;
}

unsigned char* index8(unsigned char* s, unsigned char* c, char_set* for_char_set = 0)
{
// if sys5 is defined, compile call to ::strchr instead of ::index 
//   (which is not available in pure sys5).
#ifndef sys5
    if ((NULL == for_char_set) || (for_char_set->zero_term())) 
        return (unsigned char*)(::index((char*)s, *c));
#else
    if ((NULL == for_char_set) || (for_char_set->zero_term())) 
#ifdef __cplusplus
        return (unsigned char*) ::strchr((const char*) s, *c);
#else
        return (unsigned char*) ::strchr(s, *c);
#endif
#endif
    while ((! for_char_set->is_null(s)) && (*s != *c))
        ++s;
    if (*s == *c)
        return s;
    return 0;
}

unsigned char* index16(unsigned char* s, unsigned char* c, char_set* for_char_set = 0)
{
    register unsigned short* sp = (unsigned short*) s;
    register unsigned short sc = *((unsigned short*) c);

    if ((NULL == for_char_set) || (for_char_set->zero_term())) {
        while (*sp && (*sp != sc))
            ++sp;
        if (*sp)
            return (unsigned char*) sp;
        return 0;
        }
    while ((! for_char_set->is_null((unsigned char*) sp)) && (*sp != sc))
        ++sp;
    if (! for_char_set->is_null((unsigned char*) sp))
        return (unsigned char*) sp;
    return 0;
}

unsigned char* index24(unsigned char* s, unsigned char* c, char_set* for_char_set = 0)
{
    three_byte tb1 = s;
    three_byte tb2 = c;

    while (tb1.non_null(for_char_set) && (tb1.ne(c)))
        ++tb1;
    if (tb1.non_null(for_char_set))
        return tb1.get_cp();
    return 0;
}

unsigned char* index32(unsigned char* s, unsigned char* c, char_set* for_char_set = 0)
{
    register unsigned long* lp = (unsigned long*) s;
    register unsigned long lc = *((unsigned long*) c);

    if ((NULL == for_char_set) || (for_char_set->zero_term())) {
        while (*lp && (*lp != lc))
            ++lp;
        if (*lp)
            return (unsigned char*) lp;
        return 0;
        }
    while ((! for_char_set->is_null((unsigned char*) lp)) && (*lp != lc))
        ++lp;
    if (! for_char_set->is_null((unsigned char*) lp))
        return (unsigned char*) lp;
    return 0;
}

unsigned char* rindex8(unsigned char *s, unsigned char* c, char_set* for_char_set)
{
    unsigned char* s1 = s;
    unsigned char* s2 = s;

// if sys5 is defined, compile call to ::strrchr instead of 
//  ::index (which is not available in pure sys5).
#ifndef sys5
    if ((NULL == for_char_set) || (for_char_set->zero_term()))
        return (unsigned char*)(::rindex((char*)s, *c));
#else
    if ((NULL == for_char_set) || (for_char_set->zero_term()))
#ifdef __cplusplus
        return (unsigned char*) ::strrchr((const char*) s, *c);
#else
        return (unsigned char*) ::strrchr(s, *c);
#endif
#endif
    while (! for_char_set->is_null(s1))
        ++s1;
    while (s1 != s2) {
        if (*s1 == *c)
            return s1;
        --s1;
        }
    if (*s1 == *c)
        return s1;
    return 0;
}

unsigned char* rindex16(unsigned char* s, unsigned char* c, char_set* for_char_set)
{
    register unsigned short* sp1 = (unsigned short*) s;
    register unsigned short* sp2 = (unsigned short*) s;
    register unsigned short sc = *((unsigned short*) c);

    if ((NULL == for_char_set) || (for_char_set->zero_term())) {
        while (*sp1)
            ++sp1;
        }
    else {
        while (! for_char_set->is_null((unsigned char*) sp1))
            ++sp1;
        }
    while (sp1 != sp2) {
        if (*sp1 == sc)
            return (unsigned char*) sp1;
        --sp1;
        }
    if (*sp1 == sc)
        return (unsigned char*) sp1;
    return 0;
}

unsigned char* rindex24(unsigned char* s, unsigned char* c, char_set* for_char_set = 0)
{
    three_byte tb1 = s;
    three_byte tb2 = s;
    three_byte tbc = c;

    while (tb1.non_null(for_char_set))
        ++tb1;
    while (tb1.get_cp() != tb2.get_cp()) { // !=  i.e. if pointer members not equal
        if (tb1.eq(tbc))                   // .eq i.e. if contents equal
            return tb1.get_cp();
        --tb1;
        }
    if (tb1.eq(tbc))
        return tb1.get_cp();
    return 0;
}

unsigned char* rindex32(unsigned char* s, unsigned char* c, char_set* for_char_set = 0)
{
    register unsigned long* lp1 = (unsigned long*) s;
    register unsigned long* lp2 = (unsigned long*) s;
    register unsigned long lc = *((unsigned long*) c);

    if ((NULL == for_char_set) || (for_char_set->zero_term())) {
        while (*lp1)
            ++lp1;
        }
    else {
        while (! for_char_set->is_null((unsigned char*) lp1))
            ++lp1;
        }
    while (lp1 != lp2) {
        if (*lp1 == lc)
            return (unsigned char*) lp1;
        --lp1;
        }
    if (*lp1 == lc)
        return (unsigned char*) lp1;
    return 0;
}

// These globals are never written so there is no need to add an attribute specifier for Apollo builds
unsigned char* (*strcat_tab[]  )(unsigned char*, unsigned char*, char_set*)       = { strcat8,  strcat16 ,  strcat24,   strcat32 , NULL};
unsigned char* (*strncat_tab[])(unsigned char*, unsigned char*, long, char_set*)  = { strncat8, strncat16,  strncat24,  strncat32, NULL};
long           (*strcmp_tab[])(unsigned char*, unsigned char*, char_set*)         = { strcmp8,  strcmp16 ,  strcmp24 ,  strcmp32 , NULL};
long           (*strncmp_tab[])(unsigned char*, unsigned char*, long, char_set*)  = { strncmp8, strncmp16,  strncmp24,  strncmp32, NULL};
unsigned char* (*strcpy_tab[])(unsigned char*, unsigned char*, char_set*)         = { strcpy8,  strcpy16 ,  strcpy24 ,  strcpy32 , NULL};
unsigned char* (*strncpy_tab[])(unsigned char*, unsigned char*, long, char_set*)  = { strncpy8, strncpy16,  strncpy24,  strncpy32, NULL};
long           (*strlen_tab[])(unsigned char*, char_set*)                         = { strlen8,  strlen16 ,  strlen24 ,  strlen32 , NULL};
unsigned char* (*index_tab[])(unsigned char*, unsigned char*, char_set*)          = { index8 ,  index16  ,  index24  ,  index32  , NULL};
unsigned char* (*rindex_tab[])(unsigned char*, unsigned char*, char_set*)         = { rindex8,  rindex16 ,  rindex24 ,  rindex32 , NULL};

/**************************************************************************************
    member functions for class char_set
***************************************************************************************/

char_set::char_set(unsigned long per_char_size )
{
    unsigned long str_tab_index = ((per_char_size - 1) < 4) ? (per_char_size - 1) : 4;
    unsigned long i;

    char_set_type = tml_char_set;
    char_set_name = (unsigned char*)("char_set");
    xreg          = NULL;
    char_size     = per_char_size;
    num_chars     = 0;
    ref_count     = 1;
    predom_direc  = csd_undefined;
    char_direc    = lrud_undefined;
    next          = NULL;
    collate_map   = NULL;

    strcat_p      = strcat_tab      [str_tab_index];
    strncat_p     = strncat_tab     [str_tab_index];
    strcmp_p      = strcmp_tab      [str_tab_index];
    strncmp_p     = strncmp_tab     [str_tab_index];
    strcpy_p      = strcpy_tab      [str_tab_index];
    strncpy_p     = strncpy_tab     [str_tab_index];
    strlen_p      = strlen_tab      [str_tab_index];
    index_p       = index_tab       [str_tab_index];
    rindex_p      = rindex_tab      [str_tab_index];

    null_term     = 1;
    standard = (char_size == 1) ? 1 : 0;
    hibit_clear = 0;

    null_char.value = (unsigned char*) (new char[char_size]);
    for (i = 0; i < char_size; i++)
        *(null_char.value + i) = '\0';
    null_char.in_char_set = this;

    init_esc_chars(tml_unix_esc_chars);
    regexp_chars = NULL;
}

char_set::~char_set()
{
    // already decremented if not base class
    if (char_set_type == tml_char_set) {
        if (--ref_count)
            return;
        }
    delete next;
    next = NULL;
    if (char_set_name && (0 != strcmp((char*)char_set_name, "char_set")))
        delete char_set_name;
    char_set_name = NULL;
    delete collate_map;
    collate_map = NULL;
    delete regexp_chars;
    regexp_chars = NULL;
}

/* should never be called by user */
lrud_dir char_set::get_dir_to_char(unsigned char* c)
{
    if ((0 == c) || (char_direc != lrud_undefined))
        return char_direc;
    else
        return (lrud_dir) direction_ranges.lookup(c);
}

lrud_dir char_set::get_dir_to_line()
{
    switch (predom_direc) {
        case csd_undefined:
            return lrud_undefined;
        default:
            switch (predom_direc % 4) {
                case left:
                    return right;
                case right:
                    return left;
                case down:
                    return up;
                case up:
                    return down;
            }
        }
}

lrud_dir char_set::get_dir_of_line(lrud_dir dir )
{
    if ((lrud_undefined == dir) || (4 > (predom_direc % 8)))
        return (lrud_dir) (predom_direc % 4);
    else {
        switch (dir) {
            case left:
                return right;
            case right:
                return left;
            case up:
                return down;
            case down:
                return up;
            case lrud_undefined:
                return lrud_undefined;
            }
        }
}

void char_set::set_null(unsigned char* new_null)
{
    unsigned long i;

    for (i = 0; i < char_size; i++) {
        *(null_char.value + i) = *(new_null + i);
        if ('\0' != *(new_null + i))
            null_term = 0;
        }
}

void char_set::init_esc_chars(unsigned char** esc_chars, long spec_size )
{
    unsigned char* next_str;
    unsigned long  i, j, k;

    if (spec_size > char_size) {
        printf("char_set::init_esc_chars: size specified greater than size of character in this char_set\n");
        return;
        }
    next_str = (unsigned char*) (new char[char_size]);
    for (i = 0; i < char_size - spec_size; i++)
        next_str[i] = '\0';
    for (j = 0; j < num_esc_chars; j++) {
        for (k = 0, i = char_size - spec_size; k < spec_size; k++, i++)
            next_str[i] = esc_chars[j][k];
        escape_chars[j].init(next_str, 1, this);
        }
    delete next_str;
}

unsigned long char_set::as_ulong(unsigned char* s)
{
    unsigned long retval = 0;

    switch (char_size) {
        case 1:
            retval = *s;
            break;
        case 2:
            retval = *((unsigned short*) s);
            break;
        case 3:
            retval =   (((unsigned long) *s) << 16)
                     | (((unsigned long) *(s + 1)) << 8)
                     | (((unsigned long) *(s + 2)));
            break;
        case 4:
            retval = *((unsigned long*) s);
            break;
        default:
            retval = MAX_CHARS_IN_SET;
            break;
        }
    return retval;
}

long char_set::is_null(unsigned char* c)
{
    switch (char_size) {
        case 1:
            return *c == *null_char.value;
        case 2:
            return (*((unsigned short*) c)) == (*((unsigned short*) null_char.value));
        case 3:
            return ((three_byte) c).eq((three_byte) null_char.value);
        case 4:
            return (*((unsigned long*) c)) == (*((unsigned long*) null_char.value));
        }
}

long char_set::char_arr_cmp(unsigned char* s1, unsigned char* s2)
{
    if (NULL == collate_map)                 // this character set self-collates
        return (*strcmp_p)(s1, s2, this);
    else                                     // collation info exists for this char_set
        return collate_map->compare(s1, s2);
}

long char_set::char_arr_ncmp(unsigned char* s1, unsigned char* s2, unsigned long n)
{
    if (NULL == collate_map)                       // this character set self-collates
        return (*strncmp_p)(s1, s2, n, this);
    else                                           // collation info exists for this char_set
        return collate_map->ncompare(s1, s2, n);
}

char_set* char_set::prepend(char_set* new_sets)
{
    char_set*     cur_set;
    unsigned long i;

    if (NULL == new_sets)
        return this;
    for (cur_set = new_sets; cur_set->next; cur_set = cur_set->next)
        ;
    cur_set->set_next(this);
    i = cur_set->ref_count - 1;
    if (i) {
        for (cur_set = this; cur_set; cur_set = cur_set->next)
            cur_set->ref_count += i;
        }
    return new_sets;
}

char_set* char_set::append(char_set* new_sets)
{
    char_set*     cur_set;
    unsigned long i;

    if (NULL == new_sets)
        return this;
    for (cur_set = this; cur_set->next; cur_set = cur_set->next)
        ;
    cur_set->next = new_sets;
    i = cur_set->ref_count - 1;
    for (cur_set = new_sets; cur_set; cur_set = cur_set->next)
        cur_set->ref_count += i;
    return this;
}

char_set* char_set::detach(char_set* from_vec)
{
    char_set* cur_set;
    char_set* return_vec;

    if (this == from_vec)
        return_vec = next;
    else {
        for (cur_set = from_vec; cur_set->next; cur_set = cur_set->next) {
            if (cur_set->next == this)
                break;
            }
        return_vec = from_vec;
        if (cur_set->next)
            cur_set->next = next;
        }
    next = NULL;
    delete this;
    return return_vec;
}

struct string_2_string {
    char* prefixed_string;
    char* stripped_string;
};

int str2str_cmp(string_2_string* s2s1, string_2_string* s2s2)
{
    if (s2s1->stripped_string > s2s2->stripped_string)
        return 1;
    else if (s2s2->stripped_string > s2s1->stripped_string)
        return -1;
    return 0;
}

void char_set::ignored_prefix_sort(char** strings, unsigned long num_strings, 
                                  char** prefixes, unsigned long num_prefixes)
{
    void set_csm_char_set_for_qsort(char_set*);
    int csm_cmp_for_qsort(char**, char**);

    register long lo;
    register long mid;
    register long hi;

    long*            pre_lens  = new long[num_prefixes];
    string_2_string* index_arr = new string_2_string[num_strings];
    unsigned long    i;
    long             cmp;
    char*            sorted_string;
    char*            mid_string;

    set_csm_char_set_for_qsort(this);

    // sort prefixes for logarithmic treatment
    tml_os_qsort((char*) prefixes, num_prefixes, sizeof(char*), (SORT_FUNC) csm_cmp_for_qsort);

    // strip prefixes and establish links between prefixed strings
    //   and stripped strings
    for (i = 0; i < num_prefixes; i++)
        *(pre_lens + i) = cs_strlen((unsigned char*) (*(prefixes + i)));
    for (i = 0; i < num_strings; i++) {
        (index_arr + i)->prefixed_string = *(strings + i);
        lo = 0;
        hi = num_prefixes - 1;
        // optimize integer division by 2
        mid = (lo + hi) >> 1;
        while (lo <= hi) {
            cmp = char_arr_ncmp((unsigned char*) *(prefixes + mid), (unsigned char*) *(strings + i), *(pre_lens + mid));
            if (1 == cmp)
                hi = mid - 1;
            else if (-1 == cmp)
                lo = mid + 1;
            // this string has this prefix
            else {
                (index_arr + i)->stripped_string = (*(strings + i) + *(pre_lens + mid));
                break;
                }
            // optimize integer division by 2
            mid = (lo + hi) >> 1;
            }
        // unprefixed string
        if (0 != cmp)
            (index_arr + i)->stripped_string = *(strings + i);
        // prepare to sort stripped strings
        *(strings + i) = (index_arr + i)->stripped_string;
        }
    // sort stripped strings
    tml_os_qsort((char*) strings, num_strings, sizeof(char*), (SORT_FUNC) csm_cmp_for_qsort);
    // sort index array by address of stripped string data
    tml_os_qsort((char*) index_arr, num_strings, sizeof(string_2_string), (SORT_FUNC) str2str_cmp);
    // replace stripped string origins with prefixed string origins
    for (i = 0; i < num_strings; i++) {
        sorted_string = *(strings + i);
        lo = 0;
        hi = num_strings - 1;
        // optimize integer division by 2
        mid = (lo + hi) >> 1;
        while (lo <= hi) {
            mid_string = (index_arr + mid)->stripped_string;
            if (mid_string > sorted_string)
                hi = mid - 1;
            else if (sorted_string > mid_string)
                lo = mid + 1;
            else {
                *(strings + i) = (index_arr + mid)->prefixed_string;
                break;
                }
            // optimize integer division by 2
            mid = (lo + hi) >> 1;
            }
        }
}

void d_char_set::set_type_index(char_set* save_set)
{
    long i;
    tml_type *this_type;

    i = 0;
    for (this_type = csm_type_q_get_head(); this_type; this_type = this_type->next) {
        if (this_type == save_set->get_type()) {
            char_set_type = (cs_type) i;
            break;
            }
        ++i;
        }
}

void char_set::save(cs_heap* heaps, gc_hash_tbl& hash_tbl, d_char_set* savee )
{
    unsigned char* name;
    unsigned long  i;
    gc_hash_elem*  hash_elem;
    d_ptr          d_ptr_tmp;
    long*          re_chars_set;

    if (0 == savee) {                                  // not called from derived class
        if (NULL != hash_tbl.find(this))   // already saved
            return;
        hash_elem = hash_tbl.insert(this);
        d_ptr_tmp = (void*) (  LAST_DEFAULT_HASH +
                               heaps->header.num_unions +
                               heaps->header.num_maps +
                               heaps->header.num_kmp_nodes +
                             ++heaps->num_char_sets);
        hash_elem->set_new((void*) d_ptr_tmp.as_whole_ptr());
        savee  = (d_char_set*) heaps->obj_alloc(sizeof(d_char_set));
        }

    savee->flags = 0;
    if (null_term)
        savee->flags |= cs_null_term;
    if (standard)
        savee->flags |= cs_standard;
    if (hibit_clear)
        savee->flags |= cs_hibit_clear;

    savee->set_type_index(this);
    // name is in operating system's character set
    savee->name_chars = strlen((char*) char_set_name);
    name = (unsigned char*) heaps->obj_alloc(savee->name_chars);
    name = (unsigned char*) (tml_os_byte_copy((char*)char_set_name, (char*) name, savee->name_chars));
    savee->char_size = char_size;
    savee->num_chars = num_chars;
    if (NULL == collate_map) {
        d_ptr_tmp.get_d_ptr(NULL, NULL);
        savee->collate_map = (void*) d_ptr_tmp.as_whole_ptr();
        }
    else {
        if (NULL == (savee->collate_map = hash_tbl.find(collate_map)))
            collate_map->save(heaps, hash_tbl);
        savee->collate_map = hash_tbl.find(collate_map);
        }
    savee->predom_direc = predom_direc;
    savee->char_direc = char_direc;
    direction_ranges.save(heaps, hash_tbl);
    legal_ranges.save(heaps, hash_tbl);
    alpha_ranges.save(heaps, hash_tbl);
    digit_ranges.save(heaps, hash_tbl);
    xdigit_ranges.save(heaps, hash_tbl);
    space_ranges.save(heaps, hash_tbl);
    punct_ranges.save(heaps, hash_tbl);
    print_ranges.save(heaps, hash_tbl);
    graph_ranges.save(heaps, hash_tbl);
    cntrl_ranges.save(heaps, hash_tbl);
    null_char.save(heaps, hash_tbl);
    for (i = 0; i < num_esc_chars; i++)
        escape_chars[i].save(heaps, hash_tbl);

    if (NULL == next) {
        d_ptr_tmp.get_d_ptr(NULL, NULL);
        savee->next = (void*) d_ptr_tmp.as_whole_ptr();
        }

    // next not previously saved - will be saved by calling funciton

    else if (NULL == (savee->next = hash_tbl.find(next)))
        savee->next = (void*) (    LAST_DEFAULT_HASH +
                                   heaps->header.num_unions +
                                   heaps->header.num_maps +
                                   heaps->header.num_kmp_nodes +
                               1 + heaps->num_char_sets);
}

void char_set_set_type_from_index(char_set* restore_set, tml_ext_type* types)
{
    tml_ext_type *cur_ext_type;
    tml_type     *cur_type;

    cur_ext_type = types + ((long) restore_set->get_type());
    for (cur_type = csm_type_q_get_head(); cur_type; cur_type = cur_type->next) {
        if (*cur_ext_type == *(cur_type->type)) {
            restore_set->set_type(cur_type);
            break;
            }
        }
}

void char_set::restore(cs_heap* heaps, gc_hash_tbl& hash_tbl, d_char_set* restoree)
{
    unsigned char* name;
    unsigned long  i;
    gc_hash_elem*  hash_elem;
    d_ptr          d_ptr_tmp;
    long*          re_chars_set;

    if (0 == restoree) {                             // not called from derived class
        d_ptr_tmp = (void*) (  LAST_DEFAULT_HASH +
                               heaps->header.num_unions +
                               heaps->header.num_maps +
                               heaps->header.num_kmp_nodes +
                             ++heaps->num_char_sets);
        hash_elem = hash_tbl.insert((void*) d_ptr_tmp.as_whole_ptr());
        hash_elem->set_new((void*) this);
        restoree = (d_char_set*) heaps->obj_realloc(sizeof(d_char_set));
        }
    restoree->unscramble(heaps->their_endian);

    null_term   = restoree->flags & cs_null_term;
    standard    = restoree->flags & cs_standard;
    hibit_clear = restoree->flags & cs_hibit_clear;

    char_set_type = (cs_type) restoree->char_set_type;
    char_set_set_type_from_index(this, heaps->types);
    char_size = restoree->char_size;
    name = (unsigned char*) heaps->obj_realloc(restoree->name_chars);
    // name is in operating system's character set
    char_set_name = (unsigned char*)(new char[restoree->name_chars + 1]);
    char_set_name = (unsigned char*)(tml_os_byte_copy((char*)name, (char*)char_set_name, restoree->name_chars));
    *(char_set_name + restoree->name_chars) = '\0';
    num_chars = restoree->num_chars;
    d_ptr_tmp = restoree->collate_map;
    if (d_ptr_tmp.is_null())
        collate_map = NULL;
    else if (NULL == (collate_map =
      (collation_map*) hash_tbl.find(restoree->collate_map))) {
        collate_map = new collation_map;
        collate_map->restore(heaps, hash_tbl);
        }
    predom_direc = restoree->predom_direc;
    char_direc = restoree->char_direc;
    direction_ranges.restore(heaps, hash_tbl);
    legal_ranges.restore(heaps, hash_tbl);
    alpha_ranges.restore(heaps, hash_tbl);
    digit_ranges.restore(heaps, hash_tbl);
    xdigit_ranges.restore(heaps, hash_tbl);
    space_ranges.restore(heaps, hash_tbl);
    punct_ranges.restore(heaps, hash_tbl);
    print_ranges.restore(heaps, hash_tbl);
    graph_ranges.restore(heaps, hash_tbl);
    cntrl_ranges.restore(heaps, hash_tbl);
    null_char.restore(heaps, hash_tbl);
    for (i = 0; i < num_esc_chars; i++)
        escape_chars[i].restore(heaps, hash_tbl);
    d_ptr_tmp = restoree->next;
    if (d_ptr_tmp.is_null())
        next = NULL;
    else                    // will be restored later by union_char_set
        next = (char_set*) d_ptr_tmp.as_whole_ptr();
}

/**************************************************************************************
    member functions for class alpha_char_set
***************************************************************************************/

alpha_char_set::alpha_char_set(unsigned long per_char_size ) : (per_char_size)
{
    char_set_type = tml_alpha_char_set;
    char_set_name = (unsigned char*)("alpha_char_set");
}

alpha_char_set::~alpha_char_set()
{
    // already decremented if not base class
    if (char_set_type == tml_alpha_char_set) {
        if (--ref_count)
            return;
        }
    if (char_set_name && (0 != strcmp((char*)char_set_name, "alpha_char_set")))
        delete char_set_name;
    char_set_name = NULL;
}

/* should never be called by user -- only internally by */
/*   get_dflt_union_char_set() */
void alpha_char_set::become_iso_latin_1()
{
    char_range*   next_range;
    unsigned long i;

    char_set_name = (unsigned char*)("iso_char_set");
    num_chars = 256;
    predom_direc = cr_ld_i;
    char_direc = right;

    if (next_range = upper_ranges.get_ranges()) {
        for (i = 0; i < upper_ranges.get_num_ranges(); i++)
            delete next_range++;
        }
    if (next_range = legal_ranges.get_ranges()) {
        for (i = 0; i < legal_ranges.get_num_ranges(); i++)
            delete next_range++;
        }
    if (next_range = alpha_ranges.get_ranges()) {
        for (i = 0; i < alpha_ranges.get_num_ranges(); i++)
            delete next_range++;
        }
    if (next_range = digit_ranges.get_ranges()) {
        for (i = 0; i < digit_ranges.get_num_ranges(); i++)
            delete next_range++;
        }
    if (next_range = xdigit_ranges.get_ranges()) {
        for (i = 0; i < xdigit_ranges.get_num_ranges(); i++)
            delete next_range++;
        }
    if (next_range = space_ranges.get_ranges()) {
        for (i = 0; i < space_ranges.get_num_ranges(); i++)
            delete next_range++;
        }
    if (next_range = punct_ranges.get_ranges()) {
        for (i = 0; i < punct_ranges.get_num_ranges(); i++)
            delete next_range++;
        }
    if (next_range = print_ranges.get_ranges()) {
        for (i = 0; i < print_ranges.get_num_ranges(); i++)
            delete next_range++;
        }
    if (next_range = graph_ranges.get_ranges()) {
        for (i = 0; i < graph_ranges.get_num_ranges(); i++)
            delete next_range++;
        }
    if (next_range = cntrl_ranges.get_ranges()) {
        for (i = 0; i < cntrl_ranges.get_num_ranges(); i++)
            delete next_range++;
        }

    upper_ranges.make_ranges(3);
    /* A to Z */
    upper_ranges.init_range(0, (unsigned char*)("\x41"), (unsigned char*)("\x5A"), this);
    /* Grave A to Diaresis O */
    upper_ranges.init_range(1, (unsigned char*)("\xC0"), (unsigned char*)("\xD6"), this);
    /* Ore to Thorn */
    upper_ranges.init_range(2, (unsigned char*)("\xD8"), (unsigned char*)("\xDE"), this);

    lower_ranges.make_ranges(3);
    /* a to z */
    lower_ranges.init_range(0, (unsigned char*)("\x61"), (unsigned char*)("\x7A"), this);
    /* grave a to diaresis o */
    lower_ranges.init_range(1, (unsigned char*)("\xE0"), (unsigned char*)("\xF6"), this);
    /* ore to diaresis y */
    lower_ranges.init_range(2, (unsigned char*)("\xF8"), (unsigned char*)("\xFF"), this);

    legal_ranges.make_ranges(1);
    /* NULL to diaresis y */
    legal_ranges.init_range(0, (unsigned char*)("\x00"), (unsigned char*)("\xFF"), this);

    alpha_ranges.make_ranges(6);
    alpha_ranges.set_range(upper_ranges.get_range(0), this, 0);
    alpha_ranges.set_range(upper_ranges.get_range(1), this, 1);
    alpha_ranges.set_range(upper_ranges.get_range(2), this, 2);
    alpha_ranges.set_range(lower_ranges.get_range(0), this, 3);
    alpha_ranges.set_range(lower_ranges.get_range(1), this, 4);
    alpha_ranges.set_range(lower_ranges.get_range(2), this, 5);

    digit_ranges.make_ranges(1);
    /* 0 to 9 */
    digit_ranges.init_range(0, (unsigned char*)"\x30", (unsigned char*)"\x39", this);

    xdigit_ranges.make_ranges(3);
    /* 0 to 9 */
    xdigit_ranges.set_range(digit_ranges.get_range(0), this, 0);
    /* A to F */
    xdigit_ranges.init_range(1, (unsigned char*)"\x41", (unsigned char*)"\x46", this);
    /* a to f */
    xdigit_ranges.init_range(2, (unsigned char*)"\x61", (unsigned char*)"\x66", this);

    space_ranges.make_ranges(2);
    /* horizontal tab HT */
    /* newline NL */
    /* vertical tab VT */
    /* formfeed FF */
    /* carriage return CR */
    space_ranges.init_range(0, (unsigned char*)"\x09", (unsigned char*)"\x0D", this);
    /* space SP */
    space_ranges.init_range(1, (unsigned char*)"\x20", (unsigned char*)"\x20", this);
    /* no-break space NBSP not included */

    punct_ranges.make_ranges(7);
    /* space SP to slash */
    punct_ranges.init_range(0, (unsigned char*)"\x20", (unsigned char*)"\x2F", this);
    /* colon to at-sign */
    punct_ranges.init_range(1, (unsigned char*)"\x3A", (unsigned char*)"\x40", this);
    /* left bracket to backquote */
    punct_ranges.init_range(2, (unsigned char*)"\x5B", (unsigned char*)"\x60", this);
    /* left curly to tilde */
    punct_ranges.init_range(3, (unsigned char*)"\x7B", (unsigned char*)"\x7E", this);
    /* no-break space NBSP to inverted question mark */
    punct_ranges.init_range(4, (unsigned char*)"\xA0", (unsigned char*)"\xBF", this);
    /* multiplication sign */
    punct_ranges.init_range(5, (unsigned char*)"\xD7", (unsigned char*)"\xD7", this);
    /* division sign */
    punct_ranges.init_range(6, (unsigned char*)"\xF7", (unsigned char*)"\xF7", this);

    print_ranges.make_ranges(2);
    /* space SP to tilde */
    print_ranges.init_range(0, (unsigned char*)"\x20", (unsigned char*)"\x7E", this);
    /* no-break space NBSP to diaresis y */
    print_ranges.init_range(1, (unsigned char*)"\xA0", (unsigned char*)"\xFF", this);

    graph_ranges.make_ranges(2);
    /* exclamation mark to tilde */
    graph_ranges.init_range(0, (unsigned char*)"\x21", (unsigned char*)"\x7E", this);
    /* inverted exclamation mark to diaresis y */
    graph_ranges.init_range(1, (unsigned char*)"\xA1", (unsigned char*)"\xFF", this);

    cntrl_ranges.make_ranges(2);
    /* NUL to US */
    cntrl_ranges.init_range(0, (unsigned char*)"\x00", (unsigned char*)"\x1F", this);
    /* 0x80 to APC */
    cntrl_ranges.init_range(1, (unsigned char*)"\x80", (unsigned char*)"\x9F", this);
}

void alpha_char_set::become_ascii()
{
    char_range*   next_range;
    unsigned long i;

    char_set_name = (unsigned char*)("ascii_char_set");
    num_chars = 128;
    predom_direc = cr_ld_i;
    char_direc = right;

    if (next_range = upper_ranges.get_ranges()) {
        for (i = 0; i < upper_ranges.get_num_ranges(); i++)
            delete next_range++;
        }
    if (next_range = legal_ranges.get_ranges()) {
        for (i = 0; i < legal_ranges.get_num_ranges(); i++)
            delete next_range++;
        }
    if (next_range = alpha_ranges.get_ranges()) {
        for (i = 0; i < alpha_ranges.get_num_ranges(); i++)
            delete next_range++;
        }
    if (next_range = digit_ranges.get_ranges()) {
        for (i = 0; i < digit_ranges.get_num_ranges(); i++)
            delete next_range++;
        }
    if (next_range = xdigit_ranges.get_ranges()) {
        for (i = 0; i < xdigit_ranges.get_num_ranges(); i++)
            delete next_range++;
        }
    if (next_range = space_ranges.get_ranges()) {
        for (i = 0; i < space_ranges.get_num_ranges(); i++)
            delete next_range++;
        }
    if (next_range = punct_ranges.get_ranges()) {
        for (i = 0; i < punct_ranges.get_num_ranges(); i++)
            delete next_range++;
        }
    if (next_range = print_ranges.get_ranges()) {
        for (i = 0; i < print_ranges.get_num_ranges(); i++)
            delete next_range++;
        }
    if (next_range = graph_ranges.get_ranges()) {
        for (i = 0; i < graph_ranges.get_num_ranges(); i++)
            delete next_range++;
        }
    if (next_range = cntrl_ranges.get_ranges()) {
        for (i = 0; i < cntrl_ranges.get_num_ranges(); i++)
            delete next_range++;
        }

    upper_ranges.make_ranges(1);
    /* A to Z */
    upper_ranges.init_range(0, (unsigned char*)"\x41", (unsigned char*)"\x5A", this);

    lower_ranges.make_ranges(1);
    /* a to z */
    lower_ranges.init_range(0, (unsigned char*)"\x61", (unsigned char*)"\x7A", this);

    legal_ranges.make_ranges(1);
    /* NULL to DEL */
    legal_ranges.init_range(0, (unsigned char*)"\x00", (unsigned char*)"\x7F", this);

    alpha_ranges.make_ranges(2);
    alpha_ranges.set_range(upper_ranges.get_range(0), this, 0);
    alpha_ranges.set_range(lower_ranges.get_range(0), this, 1);

    digit_ranges.make_ranges(1);
    /* 0 to 9 */
    digit_ranges.init_range(0, (unsigned char*)"\x30", (unsigned char*)"\x39", this);

    xdigit_ranges.make_ranges(3);
    /* 0 to 9 */
    xdigit_ranges.set_range(digit_ranges.get_range(0), this, 0);
    /* A to F */
    xdigit_ranges.init_range(1, (unsigned char*)"\x41", (unsigned char*)"\x46", this);
    /* a to f */
    xdigit_ranges.init_range(2, (unsigned char*)"\x61", (unsigned char*)"\x66", this);

    space_ranges.make_ranges(2);
    /* horizontal tab HT */
    /* newline NL */
    /* vertical tab VT */
    /* formfeed FF */
    /* carriage return CR */
    space_ranges.init_range(0, (unsigned char*)"\x09", (unsigned char*)"\x0D", this);
    /* space SP */
    space_ranges.init_range(1, (unsigned char*)"\x20", (unsigned char*)"\x20", this);

    punct_ranges.make_ranges(4);
    /* space SP to slash */
    punct_ranges.init_range(0, (unsigned char*)"\x20", (unsigned char*)"\x2F", this);
    /* colon to at-sign */
    punct_ranges.init_range(1, (unsigned char*)"\x3A", (unsigned char*)"\x40", this);
    /* left bracket to backquote */
    punct_ranges.init_range(2, (unsigned char*)"\x5B", (unsigned char*)"\x60", this);
    /* left curly to tilde */
    punct_ranges.init_range(3, (unsigned char*)"\x7B", (unsigned char*)"\x7E", this);

    print_ranges.make_ranges(1);
    /* space SP to tilde */
    print_ranges.init_range(0, (unsigned char*)"\x20",(unsigned char*) "\x7E", this);

    graph_ranges.make_ranges(1);
    /* exclamation mark to tilde */
    graph_ranges.init_range(0, (unsigned char*)"\x21", (unsigned char*)"\x7E", this);

    cntrl_ranges.make_ranges(2);
    /* NUL to US */
    cntrl_ranges.init_range(0, (unsigned char*)"\x00", (unsigned char*)"\x1F", this);
    /* DEL */
    cntrl_ranges.init_range(1, (unsigned char*)"\x7F", (unsigned char*)"\x7F", this);

    hibit_clear = 1;
}

void alpha_char_set::save(cs_heap* heaps, gc_hash_tbl& hash_tbl, d_char_set* savee )
{
    gc_hash_elem* hash_elem;
    d_ptr         d_ptr_tmp;

    if (0 == savee) {                      // not called from derived class
        if (NULL != hash_tbl.find(this))   //already saved
            return;
        hash_elem = hash_tbl.insert(this);
        d_ptr_tmp = (void*) (  LAST_DEFAULT_HASH +
                               heaps->header.num_unions +
                               heaps->header.num_maps +
                               heaps->header.num_kmp_nodes +
                             ++heaps->num_char_sets);
        hash_elem->set_new((void*) d_ptr_tmp.as_whole_ptr());
        savee  = (d_char_set*) heaps->obj_alloc(sizeof(d_char_set));
        }

    upper_ranges.save(heaps, hash_tbl);
    lower_ranges.save(heaps, hash_tbl);

    char_set::save(heaps, hash_tbl, savee);
}

void alpha_char_set::restore(cs_heap* heaps, gc_hash_tbl& hash_tbl, d_char_set* restoree )
{
    gc_hash_elem*         hash_elem;
    d_ptr                 d_ptr_tmp;

    // not called from derived class
    if (0 == restoree) {                   
        d_ptr_tmp = (void*) (  LAST_DEFAULT_HASH +
                               heaps->header.num_unions +
                               heaps->header.num_maps +
                               heaps->header.num_kmp_nodes +
                             ++heaps->num_char_sets);
        hash_elem = hash_tbl.insert((void*) d_ptr_tmp.as_whole_ptr());
        hash_elem->set_new((void*) this);
        restoree = (d_char_set*) heaps->obj_realloc(sizeof(d_char_set));
        }
    upper_ranges.restore(heaps, hash_tbl);
    lower_ranges.restore(heaps, hash_tbl);

    char_set::restore(heaps, hash_tbl, restoree);
}

/**************************************************************************************
    member functions for class syll_frag_char_set();
***************************************************************************************/

syll_frag_char_set::syll_frag_char_set(unsigned long per_char_size ) : (per_char_size)
{
    char_set_type = tml_syll_frag_char_set;
    char_set_name = (unsigned char*)("syll_frag_char_set");
}
                 
syll_frag_char_set::~syll_frag_char_set()
{
    // already decremented if not base class
    if (char_set_type == tml_syll_frag_char_set) {
        if (--ref_count)
            return;
        }
    if (char_set_name && (0 != strcmp((char*)char_set_name, "alpha_char_set")))
        delete char_set_name;
    char_set_name = NULL;
    delete syllable_map;
    syllable_map = NULL;
}

unsigned long syll_frag_char_set::is_syll(unsigned char* start)
{
    unsigned char* advancee = start;

    return (MAX_CHARS_IN_SET != syllable_map->lookup(start));
}

unsigned long syll_frag_char_set::get_syll(unsigned char*& start)
{
    return syllable_map->lookup(start);
}

unsigned long syll_frag_char_set::get_syll_len(unsigned char* start)
{
    unsigned char* advancee = start;

    return syllable_map->lookup(advancee);
}

void syll_frag_char_set::save(cs_heap* heaps, gc_hash_tbl& hash_tbl, d_char_set* savee)
{
    d_syll_frag_char_set* d_syll_frag_savee;
    gc_hash_elem*         hash_elem;
    d_ptr                 d_ptr_tmp;

    if (0 == savee) {                    // not called from derived class
        if (NULL != hash_tbl.find(this)) // already saved
            return;
        hash_elem = hash_tbl.insert(this);
        d_ptr_tmp = (void*) (  LAST_DEFAULT_HASH +
                               heaps->header.num_unions +
                               heaps->header.num_maps +
                               heaps->header.num_kmp_nodes +
                             ++heaps->num_char_sets);
        hash_elem->set_new((void*) d_ptr_tmp.as_whole_ptr());
        savee  = (d_char_set*) heaps->obj_alloc(sizeof(d_syll_frag_char_set));
        }

    d_syll_frag_savee = (d_syll_frag_char_set*) savee;

    if (NULL == syllable_map) {
        d_ptr_tmp.get_d_ptr(NULL, NULL);
        d_syll_frag_savee->syllable_map = (void*) d_ptr_tmp.as_whole_ptr();
        }
    else if (NULL ==
      (d_syll_frag_savee->syllable_map = (collation_map*) hash_tbl.find(syllable_map))) {
        syllable_map->save(heaps, hash_tbl);
        d_ptr_tmp = hash_tbl.find(syllable_map);
        d_syll_frag_savee->syllable_map = (void*) d_ptr_tmp.as_whole_ptr();
        }
    if (NULL == output_char_set) {
        d_ptr_tmp.get_d_ptr(NULL, NULL);
        d_syll_frag_savee->output_char_set = (void*) d_ptr_tmp.as_whole_ptr();
        }
    else if (NULL ==
      (d_syll_frag_savee->output_char_set = (char_set*) hash_tbl.find(output_char_set))) {
        output_char_set->save(heaps, hash_tbl);
        d_ptr_tmp = hash_tbl.find(output_char_set);
        d_syll_frag_savee->output_char_set = (void*) d_ptr_tmp.as_whole_ptr();
        }
    char_set::save(heaps, hash_tbl, savee);
}

tml_type* char_set_get_type_from_index(unsigned long type_index, tml_ext_type* types)
{
    tml_ext_type *cur_ext_type;
    tml_type     *cur_type;

    cur_ext_type = types + type_index;
    for (cur_type = csm_type_q_get_head(); cur_type; cur_type = cur_type->next) {
        if (*cur_ext_type == *(cur_type->type))
            break;
        }
    return cur_type;
}

void syll_frag_char_set::restore(cs_heap* heaps, gc_hash_tbl& hash_tbl, d_char_set* restoree )
{
    d_syll_frag_char_set* d_syll_frag_restoree;
    d_char_set*           output_restoree;
    tml_type*             output_char_set_type;
    gc_hash_elem*         hash_elem;
    d_ptr                 d_ptr_tmp;

    if (0 == restoree) {       // not called from derived class
        d_ptr_tmp = (void*) (  LAST_DEFAULT_HASH +
                               heaps->header.num_unions +
                               heaps->header.num_maps +
                               heaps->header.num_kmp_nodes +
                             ++heaps->num_char_sets);
        hash_elem = hash_tbl.insert((void*) d_ptr_tmp.as_whole_ptr());
        hash_elem->set_new((void*) this);
        restoree = (d_char_set*) heaps->obj_realloc(sizeof(d_syll_frag_char_set));
        }
    d_syll_frag_restoree = (d_syll_frag_char_set*) restoree;
    d_syll_frag_restoree->unscramble(heaps->their_endian);

    d_ptr_tmp = d_syll_frag_restoree->syllable_map;
    if (d_ptr_tmp.is_null())
        syllable_map = NULL;
    else if (NULL == (syllable_map =
      (collation_map*) hash_tbl.find(d_syll_frag_restoree->syllable_map))) {
        syllable_map = new collation_map;
        syllable_map->restore(heaps, hash_tbl);
        }

    d_ptr_tmp = d_syll_frag_restoree->output_char_set;
    if (d_ptr_tmp.is_null())
        output_char_set = NULL;
    else if (NULL == (d_syll_frag_restoree->output_char_set =
      (char_set*) hash_tbl.find((void*) d_ptr_tmp.as_whole_ptr()))) {
        output_restoree = (d_char_set*) heaps->obj_noalloc();
        output_char_set_type = char_set_get_type_from_index((unsigned long) output_restoree->get_char_set_type(), heaps->types);
        output_char_set = ((char_set*) output_char_set_type->instance)->make_another(output_restoree->get_char_size());
        output_char_set->restore(heaps, hash_tbl);
        }
    char_set::restore(heaps, hash_tbl, restoree);
}

/**************************************************************************************
    member functions for class union_char_set
***************************************************************************************/

union_char_set::union_char_set()
{
    char_sets = NULL;
    num_sets = 0;
    next = NULL;
    ref_count = 1;
}

union_char_set::union_char_set(char_set* new_char_sets)
{
    char_set* cur_char_set;

    char_sets = new_char_sets;
    for (cur_char_set = new_char_sets; cur_char_set;
      cur_char_set = cur_char_set->next) {
        ++cur_char_set->ref_count;
        ++num_sets;
        }
    next = NULL;
    ref_count = 1;
}

union_char_set::~union_char_set()
{
    if (--ref_count)
        return;
    delete next;
    delete char_sets;
    char_sets = NULL;
    next = NULL;
}

char_set* union_char_set::prepend(char_set* new_sets)
{
    register char_set* cur_set;

    if (NULL == new_sets)
        return char_sets;
    for (cur_set = new_sets; cur_set; cur_set = cur_set->next)
        ++num_sets;
    if (char_sets)
        char_sets->prepend(new_sets);
    else {
        char_sets = new_sets;
        for (cur_set = new_sets; cur_set; cur_set = cur_set->next)
            cur_set->incr_ref_count();
        }
    return char_sets;
}

char_set* union_char_set::append(char_set* new_sets)
{
    register char_set* cur_set;

    if (NULL == new_sets)
        return char_sets;
    for (cur_set = new_sets; cur_set; cur_set = cur_set->next)
        ++num_sets;
    if (char_sets)
        char_sets->append(new_sets);
    else {
        char_sets = new_sets;
        for (cur_set = new_sets; cur_set; cur_set = cur_set->next)
            cur_set->incr_ref_count();
        }
    return char_sets;
}

char_set* union_char_set::detach(char_set* old_set)
{
    if (char_sets && old_set)
        return old_set->detach(char_sets);
    return NULL;
}

/* prepend the union_char_set vector new_unions to this */
union_char_set* union_char_set::prepend(union_char_set* new_unions)
{
    union_char_set* cur_union;
    char_set*       cur_set;
    unsigned long   old_sets = 0;
    unsigned long   i;

    if (NULL == new_unions)
        return this;
    for (cur_union = new_unions; cur_union->next;
      cur_union = cur_union->next)
        ;
    cur_union->next->next = this;
    i = cur_union->next->ref_count - 1;
    if (i) {
        for (cur_union = this; cur_union; cur_union = cur_union->next)
            cur_union->ref_count += i;
        }
    for (cur_union = this; cur_union; cur_union = cur_union->next) {
        for (cur_set = cur_union->get_sets(); cur_set; cur_set = cur_set->get_next())
            cur_set->ref_count += i;
        }
    return new_unions;
}

/* append the union_char_set vector new_unions to the end of the */
/*   vector beginning with this */
union_char_set* union_char_set::append(union_char_set* new_unions)
{
    union_char_set* cur_union;
    char_set*       cur_set;
    unsigned long   i;

    if (NULL == new_unions)
        return this;
    for (cur_union = this; cur_union->next;
      cur_union = cur_union->next)
            ;
    cur_union->next->next = new_unions;
    i = cur_union->next->ref_count - 1;
    for (cur_union = new_unions; cur_union; cur_union = cur_union->next) {
        cur_union->ref_count += i;
        for (cur_set = cur_union->get_sets(); cur_set; cur_set = cur_set->get_next())
            cur_set->ref_count += i;
        }
    return this;
}

/* detach a single union_char_set from the vector beginning with from_vec */
union_char_set* union_char_set::detach(union_char_set* from_vec)
{
    register union_char_set *hold_vec,
                            *return_vec;

    /* first, check for trash */
    if (!this || !from_vec)
        return 0;

    if (this == from_vec)
        return_vec = next;
    else {
        return_vec = from_vec;
        for (hold_vec = from_vec; hold_vec; hold_vec = hold_vec->next)
            if (hold_vec->next == this) break;
        if (!hold_vec) return(from_vec);
        hold_vec->next = next;
        }
    delete this;
    return return_vec;
}

/* if character of size char_size at c is contained in a char_set */
/*   in this union, return a pointer to that char_set */
/* else return 0 */
char_set* union_char_set::contains(unsigned char* c, unsigned long char_size)
{
    char_set* cur_char_set;

    for (cur_char_set = char_sets;
      cur_char_set; cur_char_set = cur_char_set->get_next()) {
        if (char_size == cur_char_set->get_char_size()) {
            if (cur_char_set->is_char(c))
                return(cur_char_set);
            }
        }
    return NULL;
}

long union_char_set::contains(unsigned long char_set_index, unsigned char* c)
{
    unsigned long i;
    char_set*     next_char_set = char_sets;
     
    for (i = 0; i < char_set_index; i++)
        next_char_set = next_char_set->get_next();
    if (next_char_set->is_char(c))
        return -1;
    return 0;
}

void union_char_set::save(cs_heap* heaps, gc_hash_tbl& hash_tbl)
{
    d_union_char_set* savee;
    char_set*         cur_set;
    d_ptr             d_ptr_tmp;
    gc_hash_elem*     hash_elem;

    if (NULL != hash_tbl.find(this))   //already saved
        return;
    hash_elem = hash_tbl.insert(this);
    d_ptr_tmp = (void*) (  LAST_DEFAULT_HASH +
                         ++heaps->header.num_unions +
                           heaps->header.num_maps +
                           heaps->header.num_kmp_nodes +
                           heaps->num_char_sets);
    hash_elem->set_new((void*) d_ptr_tmp.as_whole_ptr());

    savee = (d_union_char_set*) heaps->obj_alloc(sizeof(d_union_char_set));
    savee->num_sets = num_sets;
    savee->ref_count = ref_count;

    // save character sets associated with this union_char_set

    for (cur_set = char_sets; cur_set; cur_set = cur_set->get_next()) {
        if (NULL != hash_tbl.find(cur_set))
            break;      // remainder of vector already saved
        cur_set->save(heaps, hash_tbl);
        }

    if (NULL == char_sets) {
        d_ptr_tmp.get_d_ptr(NULL, NULL);
        savee->sets = (void*) d_ptr_tmp.as_whole_ptr();
        }
    else
        savee->sets = hash_tbl.find(char_sets);

    if (NULL == next) {
        d_ptr_tmp.get_d_ptr(NULL, NULL);
        savee->next = (void*) d_ptr_tmp.as_whole_ptr();
        }

    // next not previously saved - will be saved by calling function

    else if (NULL == (savee->next = hash_tbl.find(next)))
        savee->next = (void*) (LAST_DEFAULT_HASH +
                               1 + heaps->header.num_unions +
                                   heaps->header.num_maps +
                                   heaps->header.num_kmp_nodes +
                                   heaps->num_char_sets);
}

void union_char_set::restore(cs_heap* heaps, gc_hash_tbl& hash_tbl)
{
    d_union_char_set* restoree;
    char_set*         cur_char_set;
    unsigned long     i;
    gc_hash_elem*     hash_elem;
    d_char_set*       char_set_restoree;
    cs_type           scrambled_cset_type;
    cs_type           cset_type;
    unsigned long     scrambled_char_size;
    unsigned long     char_size;
    char_set*         last_char_set;
    d_ptr             d_ptr_tmp;

    d_ptr_tmp = (void*) (  LAST_DEFAULT_HASH +
                         ++heaps->header.num_unions +
                           heaps->header.num_maps +
                           heaps->header.num_kmp_nodes +
                           heaps->num_char_sets);
    hash_elem = hash_tbl.insert((void*) d_ptr_tmp.as_whole_ptr());
    hash_elem->set_new((void*) this);
    restoree = (d_union_char_set*) heaps->obj_realloc(sizeof(d_union_char_set));

    num_sets  = restoree->num_sets;
    ref_count = restoree->ref_count;

    last_char_set = NULL;
    for (i = 0; i < num_sets; i++) {
        char_set_restoree = (d_char_set*) heaps->obj_noalloc();
        scrambled_cset_type = (cs_type) char_set_restoree->get_char_set_type();
        cset_type = (cs_type) heaps->their_endian->my_long((unsigned long) scrambled_cset_type);
        cset_type = (cs_type) char_set_get_type_from_index((unsigned long) cset_type, heaps->types);
        scrambled_char_size = char_set_restoree->get_char_size();
        char_size = heaps->their_endian->my_long(scrambled_char_size);
        cur_char_set = ((char_set*) cset_type->instance)->make_another(char_size);
        heaps->set_cur_char_set(cur_char_set);
        cur_char_set->restore(heaps, hash_tbl);

        // "next" of last_char_set has now been restored
        if (last_char_set)
            last_char_set->next = cur_char_set;

        // last char_set restored
        if (NULL == cur_char_set->next)
            break;

        d_ptr_tmp = cur_char_set->next;

        // "next" and remainder of vector already restored
        if (NULL != (cur_char_set->next =
                    (char_set*) hash_tbl.find((void*) d_ptr_tmp.as_whole_ptr())))
            break;

        // "next" is not yet restored
        last_char_set = cur_char_set;
        }
    d_ptr_tmp = restoree->sets;
    char_sets = (char_set*) hash_tbl.find((void*) d_ptr_tmp.as_whole_ptr());

    d_ptr_tmp = restoree->next;
    if (d_ptr_tmp.is_null())
        next = NULL;
    else                    // will be restored later by read_ucs_binary_file()
        next = (union_char_set*) d_ptr_tmp.as_whole_ptr();
}

long union_char_set::write_ucs_config_file(void* file_handle)
{
    return 0;
}

union_char_set* read_ucs_config_file(void* file_handle)
{
    return NULL;
}

/* returns nonzero if only character set is default iso char set */
long union_char_set::default_only()
{
    if (NULL == next) {
        if (1 == num_sets) {
            if (tml_alpha_char_set == char_sets->char_set_type) {
                if (NULL == char_sets->collate_map)
                    return(0 == strcmp((char*)(char_sets->char_set_name), "iso_char_set"));
                }
            }
        }
    return 0;
}

/* derive binary form of a vector of objects of class union_char_set */
/*   for write to disk. */
/* may be called repeatedly with same heaps and hash_tbl arguments */
/*   to consolidate all union_char_set vectors -- first call may use */
/*   NULL hash table and NULL heaps if not integrating with previously */
/*   created hash table */
/* if types is NULL, types will be constructed here */
/* pointers to objects created if any are returned in the reference arguments */
/* pointers to union_char_set objects are hashed to disk identifiers as follows: */
/*   NULL pointer             -> d_ptr with value representing NULL on disk */
/*   csm_dflt_union_char_set  -> d_ptr with value of ((void*) 0)) */
/*   char set of default union-> d_ptr with value of ((void*) 1)) */
/*     (see constant definition LAST_DEFAULT_HASH above) */
/*   other character set      -> d_ptr with incremented value of */
/*     objects                   number of character-set-related objects */
/*                               saved when union_char_set was saved, */
/*                               similarly cast */
void union_char_set::prepare_ucs_binary(gc_hash_tbl*& hash_tbl, cs_heap*& heaps, tml_ext_type*& types)
{
    tml_type*         cur_type;
    union_char_set*   cur_union;
    unsigned long     i;
    char_set_heap*    cur_char_heap;
    gc_hash_elem*     hash_elem;
    d_ptr             d_ptr_tmp;

    if (NULL == hash_tbl)
        hash_tbl = new gc_hash_tbl;
    if (NULL == heaps)
        heaps = new cs_heap();
    if (NULL == types) {
        heaps->load_types();
        types = heaps->types;
        }
    else {
        cur_type = csm_type_q_get_head();
        while (cur_type) {
            ++heaps->header.num_types;
            cur_type = cur_type->next;
            }
        heaps->types = types;
        }

    // hash default union_char_set and char_set if they exist
    //   and are not hashed already
    if (NULL != csm_dflt_union_char_set) {
        d_ptr_tmp = (void*) 0;
        if (NULL == hash_tbl->find(csm_dflt_union_char_set)) {
            hash_elem = hash_tbl->insert(csm_dflt_union_char_set);
            hash_elem->set_new((void*) d_ptr_tmp.as_whole_ptr());
            }
        if (NULL != csm_dflt_union_char_set->char_sets) {
            d_ptr_tmp = (void*) 1;
            if (NULL == hash_tbl->find(csm_dflt_union_char_set->char_sets)) {
                hash_elem = hash_tbl->insert(csm_dflt_union_char_set->get_sets());
                hash_elem->set_new((void*) d_ptr_tmp.as_whole_ptr());
                }
            }
        }

    // save this union_char_set set vector
    if (NULL == hash_tbl->find(this)) {
        for (cur_union = this; cur_union; cur_union = cur_union->next) {
            if (NULL != hash_tbl->find(cur_union)) // remainder of vector already saved
                break;
            cur_union->save(heaps, *hash_tbl);
            }
        }

    // previous call prepared binary with this cs_heap
    if (NULL != heaps->bytes_used)
        delete heaps->bytes_used;
    heaps->bytes_used = new unsigned long[heaps->header.num_heaps];
    for (i = 0, cur_char_heap = heaps->head_heap; i < heaps->header.num_heaps;
      i++, cur_char_heap = cur_char_heap->get_next())
        *(heaps->bytes_used + i) = cur_char_heap->get_bytes_used();
}

/* write character sets to disk */
/* if only one character set vector in application: */
/*   if the application establishes a default union_char_set, call this function */
/*     on the "next" member of each instance of the default */
/*   else call this function on head of union_char_set vector */
/* if more than one union_char_set vector in application: */
/*   call union_char_set::prepare_ucs_binary() for each union_char_set vector, */
/*   then call this function once */
/* may use NULL hash_tbl and heaps if union_char_sets and their member objects */
/*   do not reference user extensions that refer to objects previously hashed */
/* if types is NULL, types will be constructed and written here */
/* pointers to objects created if any are returned in the reference arguments */
long union_char_set::write_ucs_binary_file(void* file_handle, gc_hash_tbl*& hash_tbl, cs_heap*& heaps, tml_ext_type*& types)
{
    long             had_types = 0;
    unsigned long    i;
    long             bytes_written;
    long             total_bytes_written = 0;
    char_set_heap*   cur_char_heap;
    
    if (NULL != types)
        had_types = -1;

    // no previous call to prepare_ucs_binary()

    if (NULL == heaps)
        prepare_ucs_binary(hash_tbl, heaps, types);

    // another module is tracking types -- don't write them here

    if (had_types) {
        heaps->types = NULL;
        heaps->header.num_types = 0;
        }

    // write the header for binary union_char_sets

    if (-1 == (bytes_written = tml_os_write(file_handle, (char*) &heaps->header,
                                            (unsigned long) sizeof(d_union_header))))
        return -1;
    else
        total_bytes_written += bytes_written;

    // write the char_set types if they are not to be written elsewhere

    if (NULL != heaps->types) {
        if (-1 == (bytes_written = tml_os_write(file_handle, (char*) types,
                                                ((unsigned long) sizeof(tml_ext_type)) * (heaps->header.num_types))))
            return -1;
        else
            total_bytes_written += bytes_written;
        }

    // write array of bytes used per heap

    if (-1 == (bytes_written = tml_os_write(file_handle, (char*) heaps->bytes_used,
                                            ((unsigned long) sizeof(unsigned long)) * (heaps->header.num_heaps))))
        return -1;
    else
        total_bytes_written += bytes_written;

    // write the heaps

    for (i = 0, cur_char_heap = heaps->head_heap; i < heaps->header.num_heaps;
      i++, cur_char_heap = cur_char_heap->get_next()) {
        if (-1 == (bytes_written =
                   tml_os_write(file_handle, (char*)(cur_char_heap->get_start()), *(heaps->bytes_used + i))))
            return -1;
        else
            total_bytes_written += bytes_written;
        }
    return total_bytes_written;
}


/* read character sets from disk */
/* if the application establishes a default union_char_set, call this function */
/*   on a preallocated "next" member of an instance of the default */
/* else call this function on a preallocated head of union_char_set vector */
/* may use NULL hash_tbl and heaps if union_char_sets and their member objects */
/*   do not reference user extensions that refer to objects previously hashed */
/* if types is NULL, types will be constructed and read here */
/* pointers to objects created if any are returned in the reference arguments */
union_char_set* read_ucs_binary_file(void* file_handle, gc_hash_tbl*& hash_tbl, tml_ext_type*& types, endian*& their_endian)
{
    long             i;
    long             bytes_read;
    long             total_bytes_read = 0;
    d_union_header*  header;
    unsigned long    num_heaps;
    unsigned long    num_unions;
    cs_heap*         heaps;
    char_set_heap*   cur_char_heap;
    unsigned long*   bytes_used;
    union_char_set*  cur_union;
    union_char_set*  last_union;
    unsigned long    made_endian = 0;
    unsigned long    made_types = 0;
    union_char_set*  first_union = NULL;  // return value of this function
    gc_hash_elem*    hash_elem;
    d_ptr            d_ptr_tmp;

    header = new d_union_header;

    if (-1 == (bytes_read = tml_os_read(file_handle, (char*) header,
                                        (unsigned long) sizeof(unsigned long)))) {
        delete header;
        return first_union;
        }

    if (NULL == their_endian) {
        their_endian = new endian_3210;
        made_endian = 1;
        }
    num_heaps = their_endian->my_long(header->num_heaps);

    if (0 == num_heaps) {
        delete header;
        if (made_endian) {
            delete their_endian;
            their_endian = NULL;
            }
        return first_union;
        }

    // 1. read remainder of header

    if (-1 == (bytes_read = tml_os_read(file_handle, (char*) &header->num_unions,
                                        (unsigned long) (sizeof(d_union_header) - sizeof(unsigned long)))))
        return first_union;
    total_bytes_read += bytes_read;

    header->unscramble(their_endian);

    // 2. prepare to restore heaps

    heaps = new cs_heap(their_endian);
    if (NULL == hash_tbl)
        hash_tbl = new gc_hash_tbl;
    delete heaps->head_heap;
    heaps->num_char_sets = 0;
    heaps->head_heap = NULL;
    heaps->header.num_heaps = header->num_heaps;
    heaps->header.num_unions = 0;
    num_unions = header->num_unions;
    heaps->header.num_maps = 0;
    heaps->header.num_kmp_nodes = 0;
    heaps->header.num_types = header->num_types;
    heaps->header.version_number = header->version_number;

    // 3. read types if not supplied

    if (NULL == types) {
        heaps->types = new tml_ext_type[header->num_types];
        made_types = 1;
        if (-1 == (bytes_read = tml_os_read(file_handle, (char*) heaps->types,
                                            header->num_types * (unsigned long) sizeof(tml_ext_type)))) {
            delete header;
            if (made_endian) {
                delete their_endian;
                their_endian = NULL;
                }
            if (made_types) {
                delete heaps->types;
                heaps->types = NULL;
                }
            delete heaps;
            return first_union;
            }
        types = heaps->types;
        total_bytes_read += bytes_read;
        }
    else
        heaps->types = types;

    // 4. read the array of bytes used per heap

    bytes_used = new unsigned long[num_heaps];
    if (-1 == (bytes_read = tml_os_read(file_handle, (char*) bytes_used,
                                        num_heaps * (unsigned long) sizeof(unsigned long)))) {
        delete header;
        if (made_endian)
            delete their_endian;
        if (made_types) {
            delete heaps->types;
            heaps->types = NULL;
            types = NULL;
            }
        delete heaps;
        delete bytes_used;
        return first_union;
        }
    their_endian->long_arr_shuf(bytes_used, num_heaps);
    heaps->bytes_used = bytes_used;
    total_bytes_read += bytes_read;

    // 5. restore heaps

    for (i = 0; i < num_heaps; i++) {
        cur_char_heap = new char_set_heap(*(bytes_used + i));
        if (-1 == (bytes_read = tml_os_read(file_handle,
                                            (char*) cur_char_heap->get_start(),
                                            *(bytes_used + i)))) {
            delete header;
            if (made_endian) {
                delete their_endian;
                their_endian = NULL;
                }
            if (made_types) {
                delete heaps->types;
                heaps->types = NULL;
                types = NULL;
                }
            delete bytes_used;
            heaps->bytes_used = NULL;
            delete heaps;
            return first_union;
            }
        cur_char_heap->set_bytes_used(*(bytes_used + i));
        total_bytes_read += bytes_read;
        heaps->queue_heap(cur_char_heap);
        }

    // 6. hash default union_char_set and char_set if they exist
    //      and are not hashed already

    if (NULL != csm_dflt_union_char_set) {
        d_ptr_tmp = (void*) 0;
        if (NULL == hash_tbl->find((void*) d_ptr_tmp.as_whole_ptr())) {
            hash_elem = hash_tbl->insert((void*) d_ptr_tmp.as_whole_ptr());
            hash_elem->set_new(csm_dflt_union_char_set);
            }
        if (NULL != csm_dflt_union_char_set->char_sets) {
            d_ptr_tmp = (void*) 1;
            if (NULL == hash_tbl->find((void*) d_ptr_tmp.as_whole_ptr())) {
                hash_elem = hash_tbl->insert((void*) d_ptr_tmp.as_whole_ptr());
                hash_elem->set_new(csm_dflt_union_char_set->char_sets);
                }
            }
        }

    // 7. restore union_char_sets

    last_union = NULL;
    first_union = new union_char_set;
    cur_union = first_union;
    num_unions = header->num_unions;
    for (i = 0; i < num_unions; i++) {
        cur_union->restore(heaps, *hash_tbl);

        // "next" of last_union has now been restored
        if (last_union)
            last_union->next = cur_union;

        // last union_char_set restored
        if (NULL == cur_union->next)
            break;

        d_ptr_tmp = cur_union->next;

        // "next" and remainder of vector have already been restored
        if (NULL != (cur_union->next =
                     (union_char_set*) hash_tbl->find((void*) d_ptr_tmp.as_whole_ptr())))
            break;

        // "next" is not yet restored
        last_union = cur_union;

        cur_union = new union_char_set;
        }

    delete header;
    delete bytes_used;
    heaps->bytes_used = NULL;
    delete heaps;

    return first_union;
}

/**************************************************************************************
    member functions for class char_set_heap
***************************************************************************************/

char_set_heap::char_set_heap(long new_size )
{
    next_byte = 0;
    bytes_used = 0;
    size = (new_size) ? new_size : char_set_heap_incr;
    heap_start = (unsigned char*)(new char[size]);
    next = NULL;
}

char_set_heap::~char_set_heap()
{
    delete heap_start;
    heap_start = NULL;
    delete next;
    next = NULL;
}

unsigned char* char_set_heap::obj_alloc(long num_bytes)
{
    unsigned char* retval;
    long           plus_count;

    if (num_bytes > size - next_byte)
        return NULL;
    retval = heap_start + next_byte;
    plus_count = num_bytes
               += ((num_bytes % CHAR_SET_ALIGN_BLOCK)
               ? (CHAR_SET_ALIGN_BLOCK - (num_bytes % CHAR_SET_ALIGN_BLOCK))
               : 0);
    next_byte += plus_count;
    bytes_used += plus_count;
    return retval;
}

unsigned char* char_set_heap::obj_realloc(long num_bytes)
{
    unsigned char* retval;
    long           plus_count;

    if (num_bytes > bytes_used - next_byte)
        return NULL;
    retval = heap_start + next_byte;
    plus_count =  num_bytes
               += ((num_bytes % CHAR_SET_ALIGN_BLOCK)
               ? CHAR_SET_ALIGN_BLOCK - (num_bytes % CHAR_SET_ALIGN_BLOCK)
               : 0);
    next_byte += plus_count;
    return retval;
}

/**************************************************************************************
    member functions for class cs_heap
***************************************************************************************/

cs_heap::cs_heap(endian* new_endian )
{
    head_heap = new char_set_heap;
    bytes_used = NULL;
    num_char_sets = 0;
    cur_char_set = NULL;
    their_endian = new_endian;
    header.num_heaps = 1;
    types = NULL;
}

cs_heap::~cs_heap()
{
    delete head_heap;
    head_heap = NULL;
    delete bytes_used;
    bytes_used = NULL;
    num_char_sets = 0;
    cur_char_set = NULL;
    their_endian = NULL;
    types = NULL;
}

unsigned char* cs_heap::obj_alloc(long num_bytes)
{
    char_set_heap* new_heap;
    unsigned char* retval;

    if (num_bytes > head_heap->size) {
        printf("cs_heap::obj_alloc: requested size exceeds maximum of %d\n", head_heap->size);
        return NULL;
        }
    if (NULL == (retval = head_heap->obj_alloc(num_bytes))) {
        new_heap = new char_set_heap;
        new_heap->next = head_heap;
        head_heap = new_heap;
        ++header.num_heaps;
        retval = head_heap->obj_alloc(num_bytes);
        }
    return retval;
}

unsigned char* cs_heap::obj_realloc(long num_bytes)
{
    unsigned char* retval;

    if (NULL == (retval = head_heap->obj_realloc(num_bytes))) {
        head_heap = head_heap->next;
        ++header.num_heaps;
        retval = head_heap->obj_realloc(num_bytes);
        }
    return retval;
}

unsigned char* cs_heap::obj_noalloc()
{
    if (head_heap->bytes_used > head_heap->next_byte)
        return (head_heap->heap_start + head_heap->next_byte);
    else
        return head_heap->next->heap_start;
}

void cs_heap::queue_heap(char_set_heap* new_heap)
{
    new_heap->next = head_heap;
    head_heap = new_heap;
}

void cs_heap::load_types()
{
    long i;
    tml_type* cur_type;

    cur_type = csm_type_q_get_head();
    while (cur_type) {
        ++header.num_types;
        cur_type = cur_type->next;
        }
    types = new tml_ext_type[header.num_types];
    cur_type = csm_type_q_get_head();
    for (i = 0; i < header.num_types; i++) {
        *(types + i) = *cur_type->type;     // bit copy
        cur_type = cur_type->next;
        }
}

/**************************************************************************************
    member functions for disk-format character set classes
***************************************************************************************/

void d_csm_character::unscramble(endian* their_endian)
{
    num_chars = their_endian->my_long(num_chars);
    collation = their_endian->my_long(collation);
}

void d_char_range_set::unscramble(endian* their_endian)
{
    num_ranges = their_endian->my_long(num_ranges);
}

void d_kmp_node::unscramble(endian* their_endian)
{
    ((d_ptr*) &right)->unscramble(their_endian);
    ((d_ptr*) &down)->unscramble(their_endian);
    ((d_ptr*) &failure)->unscramble(their_endian);
}

void d_kmp_node::set_down(d_ptr* d_ptr_tmp)
{
    WHOLE_PTR  disk_p;
    
    disk_p = d_ptr_tmp->as_whole_ptr();
    down = (d_kmp_node*) disk_p;
}

void d_kmp_node::set_right(d_ptr* d_ptr_tmp)
{
    WHOLE_PTR  disk_p;
    
    disk_p = d_ptr_tmp->as_whole_ptr();
    right = (d_kmp_node*) disk_p;
}

void d_kmp_node::set_failure(d_ptr* d_ptr_tmp)
{
    WHOLE_PTR  disk_p;
    
    disk_p = d_ptr_tmp->as_whole_ptr();
    failure = (d_kmp_node*) disk_p;
}

void d_collation_map::unscramble(endian* their_endian)
{
    ((d_ptr*) &many_to_one)->unscramble(their_endian);
}

void d_char_set::unscramble(endian* their_endian)
{
    char_set_type = (cs_type) their_endian->my_long((unsigned long) char_set_type);
    name_chars = their_endian->my_long(name_chars);
    char_size = their_endian->my_long(char_size);
    num_chars = their_endian->my_long(num_chars);
    ((d_ptr*) &collate_map)->unscramble(their_endian);
    predom_direc = (char_set_dir) their_endian->my_short(predom_direc);
    char_direc = (lrud_dir) their_endian->my_short(char_direc);
    ((d_ptr*) &next)->unscramble(their_endian);
    flags = their_endian->my_long(flags);
}

void d_syll_frag_char_set::unscramble(endian* their_endian)
{
    ((d_ptr*) &syllable_map)->unscramble(their_endian);
    ((d_ptr*) &output_char_set)->unscramble(their_endian);
}

void d_union_char_set::unscramble(endian* their_endian)
{
    num_sets      = their_endian->my_long(num_sets);
    ref_count     = their_endian->my_long(ref_count);
    ((d_ptr*) &sets)->unscramble(their_endian);
    ((d_ptr*) &next)->unscramble(their_endian);
}

d_union_header::d_union_header()
{
    num_heaps = 0;
    num_unions = 0;
    num_maps = 0;
    num_kmp_nodes = 0;
    num_types = 0;
    version_number = csm_version_number;
}

void d_union_header::unscramble(endian* their_endian)
{
    num_heaps      = their_endian->my_long(num_heaps);
    num_unions     = their_endian->my_long(num_unions);
    num_maps       = their_endian->my_long(num_maps);
    num_kmp_nodes  = their_endian->my_long(num_kmp_nodes);
    num_types      = their_endian->my_long(num_types);
    version_number = their_endian->my_long(version_number);
}

/**************************************************************************************
    save and restore for class kmp_node
***************************************************************************************/

// Written as general, not member functions, since kmp_node is saved/restored
//   only in context of character set save/restore

void kmp_tree_save(kmp_node* root, cs_heap* heaps, gc_hash_tbl& hash_tbl)
{
    unsigned long num_nodes;
    unsigned long base_num;
    kmp_node*     next_node;
    d_kmp_node*   savee_vec;
    d_kmp_node*   savee;
    gc_hash_elem* hash_elem;
    d_ptr         d_ptr_tmp;
    kmp_dir       dir;

    if (NULL != hash_tbl.find(root))  // tree already saved
        return;

    // count nodes in tree and hash them

    base_num = LAST_DEFAULT_HASH +
               heaps->header.num_unions +
               heaps->header.num_maps +
               heaps->num_char_sets;

    // unique count for this tree; heaps->header counts nodes in all trees

    num_nodes = 0;
    next_node = root;
    dir = kmp_down;
    do {
        ++num_nodes;
        hash_elem = hash_tbl.insert(next_node);
        d_ptr_tmp = (void*) (  base_num +
                             ++heaps->header.num_kmp_nodes);
        hash_elem->set_new((void*) d_ptr_tmp.as_whole_ptr());
        do {
            next_node = next_node->depth_first(dir);
            if (next_node == root)
                break;
            } while (kmp_up == dir);
        } while ( ! ((next_node == root) && (kmp_up == dir)));


    // allocate d_kmp_node objects for entire tree

    savee_vec = (d_kmp_node*) heaps->obj_alloc((long) (sizeof(d_kmp_node) * num_nodes));

    // initialize right, down, and failure nodes in savee_vec
    // initialize this_character members in savee_vec

    savee = savee_vec;
    next_node = root;
    dir = kmp_down;
    do {
        if (NULL == next_node->get_right())
            d_ptr_tmp.get_d_ptr(NULL, NULL);
        else
            d_ptr_tmp = hash_tbl.find(next_node->get_right());
        savee->set_right(&d_ptr_tmp);
        if (NULL == next_node->get_down())
            d_ptr_tmp.get_d_ptr(NULL, NULL);
        else
            d_ptr_tmp = hash_tbl.find(next_node->get_down());
        savee->set_down(&d_ptr_tmp);
        if (NULL == next_node->get_failure())
            d_ptr_tmp.get_d_ptr(NULL, NULL);
        else
            d_ptr_tmp = hash_tbl.find(next_node->get_failure());
        savee->set_failure(&d_ptr_tmp);
        next_node->get_character()->save(heaps, hash_tbl);
        ++savee;

        do {
            next_node = next_node->depth_first(dir);
            if (next_node == root)
                break;
            } while (kmp_up == dir);
        } while ( ! ((next_node == root) && (kmp_up == dir)));
}

void kmp_tree_restore(kmp_node*& root, cs_heap* heaps, gc_hash_tbl& hash_tbl)
{
    register d_kmp_node*   restoree;
    register unsigned long i;
    register unsigned long nodes_seen = 0;
    register unsigned long nodes_sought = 1;
    register kmp_node*     next_node;

    unsigned long    base_num;
    d_kmp_node*      restoree_vec;
    csm_character*   next_character;
    gc_hash_elem*    hash_elem;
    d_ptr            d_ptr_tmp;
    kmp_dir          dir;

    // count the nodes

    restoree = (d_kmp_node*) heaps->obj_noalloc();
    while (nodes_seen < nodes_sought) {
        d_ptr_tmp = (void*) restoree->get_down();
        if ( ! d_ptr_tmp.is_null())
            ++nodes_sought;
        d_ptr_tmp = (void*) restoree->get_right();
        if ( ! d_ptr_tmp.is_null())
            ++nodes_sought;
        ++nodes_seen;
        ++restoree;
        }

    // reallocate saved d_kmp_node objects - d_csm_character members were
    //   allocated piecemeal by calls to csm_character::save() and must
    //   be reallocated in parallel

    restoree_vec = (d_kmp_node*) heaps->obj_realloc((long) (sizeof(d_kmp_node) * nodes_seen));

    // allocate new kmp_node and csm_character objects

    root = new kmp_node[nodes_seen];
    next_character = new csm_character[nodes_seen];

    // hash new kmp_node objects against IDs of objects to be restored

    restoree = restoree_vec;
    next_node = root;
    base_num = LAST_DEFAULT_HASH +
               heaps->header.num_unions +
               heaps->header.num_maps +
               heaps->num_char_sets;

    for (i = 0; i < nodes_seen; i++) {
        d_ptr_tmp = (void*) (  base_num +
                             ++heaps->header.num_kmp_nodes);
        hash_elem = hash_tbl.insert((void*) d_ptr_tmp.as_whole_ptr());
        hash_elem->set_new(next_node++);
        }

    root->set_up(NULL);
    root->set_right(NULL);
    root->set_character(NULL);
    root->set_failure(NULL);
    root->set_matched(NULL);

    restoree = restoree_vec;
    next_node = root;
    dir = kmp_down;
    do {
        d_ptr_tmp = (void*) restoree->get_right();
        if (d_ptr_tmp.is_null())
            next_node->set_right(NULL);
        else
            next_node->set_right((kmp_node*) hash_tbl.find((void*) d_ptr_tmp.as_whole_ptr()));

        d_ptr_tmp = (void*) restoree->get_down();
        if (d_ptr_tmp.is_null())
            next_node->set_down(NULL);
        else
            next_node->set_down((kmp_node*) hash_tbl.find((void*) d_ptr_tmp.as_whole_ptr()));

        d_ptr_tmp = (void*) restoree->get_failure();
        if (d_ptr_tmp.is_null())
            next_node->set_failure(NULL);
        else
            next_node->set_failure((kmp_node*) hash_tbl.find((void*) d_ptr_tmp.as_whole_ptr()));

        if (next_node->get_down())
            next_node->get_down()->set_up(next_node);

        if (next_node->get_right())
            next_node->get_right()->set_up(next_node->get_up());

        next_node->set_character(next_character++);
        next_node->get_character()->restore(heaps, hash_tbl);

        next_node->set_matched(NULL);

        ++restoree;

        do {
            next_node = next_node->depth_first(dir);
            if (next_node == root)
                break;
            } while (kmp_up == dir);
        } while ( ! ((next_node == root) && (kmp_up == dir)));
}

/**************************************************************************************
    general functions to compare two strings in different character sets
***************************************************************************************/

// NOTE: These functions have meaning only when the collation_range definitions of
//   the two character sets are coordinated properly. The two csm_characters should
//   be initialized with value and character set.

// compare two strings in different character sets - maximum character size 4 bytes
long char_arr_cmp(csm_character* cc1, csm_character* cc2)
{
    char_set*      cs1 = cc1->get_char_set();
    char_set*      cs2 = cc2->get_char_set();
    collation_map* coll_map1 = cs1->get_collate_map();
    collation_map* coll_map2 = cs2->get_collate_map();
    unsigned char* s1 = cc1->get_value();
    unsigned char* s2 = cc2->get_value();
    unsigned long  coll_1;
    unsigned long  coll_2;

    while ( ! (cs1->is_null(s1) && cs2->is_null(s2))) {
        coll_1 = coll_map1->lookup(s1);
        coll_2 = coll_map2->lookup(s2);
        if (coll_1 > coll_2)
            return 1;
        else if (coll_2 > coll_1)
            return -1;
        }
    return 0;
}

// compare up to n characters of two strings in different character sets
long char_arr_ncmp(csm_character* cc1, csm_character* cc2, unsigned long n)
{
    char_set*      cs1 = cc1->get_char_set();
    char_set*      cs2 = cc2->get_char_set();
    collation_map* coll_map1 = cs1->get_collate_map();
    collation_map* coll_map2 = cs2->get_collate_map();
    unsigned char* s1 = cc1->get_value();
    unsigned char* s2 = cc2->get_value();
    unsigned long  coll_1;
    unsigned long  coll_2;

    while (n--) {
        if ((cs1->is_null(s1)) && (cs2->is_null(s2)))
            return 0;
        coll_1 = coll_map1->lookup(s1);
        coll_2 = coll_map2->lookup(s2);
        if (coll_1 > coll_2)
            return 1;
        else if (coll_2 > coll_1)
            return -1;
        }
    return 0;
}

/**************************************************************************************
    global for csm_cmp_for_qsort() comparison function for UNIX qsort()
***************************************************************************************/

// To make this global per-process, must be declared with attribute specifier
//   for Apollo builds to allow use in global and inlib'd libraries
#ifdef apollo
char_set* csm_char_set_for_qsort __attribute((section(csm_char_set_for_qsort))) = NULL;
#else
char_set* csm_char_set_for_qsort = NULL;
#endif

void set_csm_char_set_for_qsort(char_set* new_char_set)
{
    csm_char_set_for_qsort = new_char_set;
}

/**************************************************************************************
    comparison function to use with UNIX qsort
***************************************************************************************/

int csm_cmp_for_qsort(char** sp1, char** sp2)
{
    if (NULL != csm_char_set_for_qsort) {
        if (NULL != csm_char_set_for_qsort->get_collate_map()) // collation info exists for this char_set
            return csm_char_set_for_qsort->get_collate_map()->compare((unsigned char*)(*sp1), (unsigned char*)(*sp2));
        // this character set self-collates
        return csm_char_set_for_qsort->cs_strcmp((unsigned char*)(*sp1), (unsigned char*)(*sp2));
        }
    return strcmp(*sp1, *sp2);
}

/**************************************************************************************
    functions to preserve architecture neutrality
***************************************************************************************/

endian* get_their_endian(unsigned long their_magic_number)
{
    switch (their_magic_number) {
        case TML_MAGIC_0123:
            return new endian_0123;
        case TML_MAGIC_1032:
            return new endian_1032;
        case TML_MAGIC_2301:
            return new endian_2301;
        case TML_MAGIC_3210:
            return new endian_3210;
        default:
            printf("get_their_endian: unknown magic number\n");
            return NULL;
        }
}

endian* get_my_endian()
{
    return get_their_endian(TML_MAGIC_3210);
}


