/**************************************************************************************
    cursor_traits.cxx
    code for the basic v_cursor object in the rendering component

    (c) Copyright Apollo Computer Inc., 1987 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

#include <tml/tml_objects.hxx>
#include <tml/tml_functions.hxx>
#include <tml/tml_dp_graph.hxx>

/**************************************************************************************
    Constructor for the cursor object

**************************************************************************************/
/*cursor constructor*/
v_cursor::v_cursor()
{
    in_viewport = 0;
    raw_loc.at_x = -1;
    raw_loc.at_y = -1;
    caged_loc = raw_loc;
    in_view_pos = 0;
    in_cline = 0;
    after_char.in_node = 0;
    after_char.char_index = -1;
    is_visible = is_enabled = 0;
    height = width = 10;
    draw_func = tml_dp_get_dflt_vpftbl()->ptr_2_tml_dp_dflt_cursor_draw;
}


/**************************************************************************************
    traits to manipulate the cursor data

**************************************************************************************/
                                                     
/*detach a cursor from a viewport*/
void v_cursor::detach()
{
    if (!in_viewport)
        return;

    if (is_visible)
        is_visible = (*draw_func)(in_viewport->get_handle(), get_origin_x(), get_origin_y(), get_height(), get_width(), 1);
                        
    is_enabled = 0;
    in_viewport = 0;
}

/*enable--make it possible for the cursor to be visible*/
long v_cursor::enable(long force)
{
    is_enabled++;

    if (force && (is_enabled < 1))
        is_enabled = 1;
                                                                                                                          
    if (!is_enabled)
        return 0;

    caged_loc = in_viewport->mpos_2_xy(after_char);
    raw_loc = caged_loc;
    if ((!is_visible) && (caged_loc.at_y > 0) && (draw_func != 0))
        is_visible = (*draw_func)(in_viewport->get_handle(), get_origin_x(), get_origin_y(), get_height(), get_width(), 0);
    else
        is_visible = 0;

    return is_visible;
}

/*disable the cursor--if it is currently drawn, get rid of it*/
long v_cursor::disable(long force)
{
    is_enabled--;

    if (force && (is_enabled > 0))
        is_enabled = 0;

    if (is_enabled) return 1;
                   
    if ((caged_loc.at_y > 0) && is_visible)
        is_visible = (*draw_func)(in_viewport->get_handle(), get_origin_x(), get_origin_y(), get_height(), get_width(), 1);
    else
        is_visible = 0;

    return is_visible;
}

/**************************************************************************************
    traits to draw and move the cursor

**************************************************************************************/
/*draw the cursor at the current position, if it is not visible already and is enabled*/
long v_cursor::draw()
{
    if (is_visible || !is_enabled)
        return 0;

    if (!in_viewport || (draw_func == 0))
        return 0;

    is_visible = (*draw_func)(in_viewport->get_handle(), get_origin_x(), get_origin_y(), get_height(), get_width(), 0);
    return is_visible;
}
              
/*erase the cursor*/
long v_cursor::erase()
{
    if (!is_visible || !is_enabled)
        return 0;
    if (!in_viewport || (draw_func == 0))
        return 0;

    is_visible = (*draw_func)(in_viewport->get_handle(), get_origin_x(), get_origin_y(), get_height(), get_width(), 1);
    return (is_visible ? 0 : 1);
}
                                                     
/*set the drawing functions*/
void v_cursor::set_draw_func(cursor_draw new_func)
{
    /*if it isn't in a viewport, or isn't enabled, just set the function and return*/
    if (!in_viewport || !is_enabled){
        draw_func = new_func;
        return;
        }

    /*if it is visible, erase the old cursor first*/
    if (is_visible)
        is_visible = (*draw_func)(in_viewport->get_handle(), get_origin_x(), get_origin_y(), get_height(), get_width(), 1);
    
    draw_func = new_func;
                                                               
    /*if it is visible, redraw using the new draw function*/
    if (is_visible)
        is_visible = (*draw_func)(in_viewport->get_handle(), get_origin_x(), get_origin_y(), get_height(), get_width(), 0);
}


/*put the cursor at screen position screen_xy.*/
long v_cursor::set(screen_xy position)
{
    model_pos new_after_char;

    if (!in_viewport)
        return 0;

    if (in_viewport)
        (*in_viewport->get_ftbl()->ptr_2_tml_dp_damage_repair)(in_viewport->get_handle());

    new_after_char = in_viewport->xy_2_mpos(position);
    if (new_after_char.in_node)
        position = in_viewport->mpos_2_xy(new_after_char);
    in_view_pos = in_viewport->xy_2_vpos(position);
    if (!in_view_pos) return 0;

    if (is_enabled && is_visible && (draw_func != 0))
        is_visible = (*draw_func)(in_viewport->get_handle(), get_origin_x(), get_origin_y(), get_height(), get_width(), 1);

    after_char = new_after_char;
    in_cline = in_viewport->xy_2_cline(position);
    raw_loc = position;

    if (after_char.in_node && in_view_pos->contains(after_char)){
        caged_loc = in_viewport->mpos_2_xy(after_char);
        raw_loc.at_y = caged_loc.at_y;
        }
    else {
        caged_loc = raw_loc;
        after_char.in_node = 0;
        after_char.char_index = 0;
        }

    if (is_enabled && (!is_visible) && (draw_func != 0))
        is_visible = (*draw_func)(in_viewport->get_handle(), get_origin_x(), get_origin_y(), get_height(), get_width(), 0);

    return 1;
}

/*put the cursor before model position position*/
long v_cursor::set(model_pos position)
{
    model_pos hold_pos;
                                                         
    if (!position.in_node)
        return(set(raw_loc));

    if (in_viewport)
        (*in_viewport->get_ftbl()->ptr_2_tml_dp_damage_repair)(in_viewport->get_handle());

    if (!in_viewport || !in_viewport->get_format() ||
        !(in_viewport->get_format()->get_model() == position.in_node->get_model()))
        return 0;

    if (!position.can_format()){
        hold_pos = position.get_next();
        if (!hold_pos.in_node)
            hold_pos = position.get_prior();
        if (!hold_pos.in_node)
            return 0;
        position = hold_pos;
        }
    
    if (is_enabled && is_visible && (draw_func != 0))
        is_visible = (*draw_func)(in_viewport->get_handle(), get_origin_x(), get_origin_y(), get_height(), get_width(), 1);

    in_view_pos = in_viewport->mpos_2_vpos(position);
    if (!in_view_pos){
        in_cline = 0;
        raw_loc.at_x = -1;
        raw_loc.at_y = -1;
        caged_loc = raw_loc;
        after_char = position;
        return 1;
        }

    raw_loc = in_viewport->mpos_2_xy(position, in_view_pos);
    caged_loc = raw_loc;
    after_char = position;
    if (raw_loc.at_y >= 0){
        in_cline = in_viewport->xy_2_cline(raw_loc);
        if (is_enabled && (!is_visible))
            is_visible = (*draw_func)(in_viewport->get_handle(), get_origin_x(), get_origin_y(), get_height(), get_width(), 0);
        }
    return 1;
}
            
/*move the cursor as from it's present position in the way indicated--return the screen coordinates
where the cursor is drawn or would be drawn*/
screen_xy v_cursor::move(vp_dir move_dir)
{
    screen_xy  return_xy,
                    hold_xy;
    model_pos next_pos;

    if (in_viewport)
        (*in_viewport->get_ftbl()->ptr_2_tml_dp_damage_repair)(in_viewport->get_handle());

    return_xy.at_x = raw_loc.at_x;
    return_xy.at_y = raw_loc.at_y;

    switch (move_dir) {
      case vp_right:
        next_pos = after_char.get_next();
        if (!next_pos.in_node)
            return raw_loc;
        if (!set(next_pos))
            return raw_loc;
        break;

      case vp_left:
        next_pos = after_char.get_prior();
        if (!next_pos.in_node)
            return raw_loc;
        if (!set(next_pos))
            return raw_loc;
        break;

      case vp_up:
        if (!in_cline) break;
        hold_xy = caged_loc;
        return_xy.at_y = in_view_pos->get_y() - 1;
        if (return_xy.at_y < 0)
            return raw_loc;
        if (!set(return_xy))
            return raw_loc;
        if (caged_loc.at_y == hold_xy.at_y){
            if ((in_view_pos->get_up()) && (in_view_pos->get_up()->get_up())){
                return_xy.at_y = in_view_pos->get_up()->get_up()->get_y() + 1;
                if (!set(return_xy))
                    return raw_loc;
                }
            }
        break;

      case vp_down:
        if (!in_cline) break;
        return_xy.at_y = in_view_pos->get_y() + in_cline->get_hi() + 1;
        if (return_xy.at_y > in_viewport->get_hi())
            return raw_loc;
        if (!set(return_xy))
            return raw_loc;
        break;

      case vp_begin_line:
        if (!in_cline) break;
        return_xy.at_x = in_cline->get_ulx() + in_viewport->get_hoffset();
        if (!set(return_xy))
            return raw_loc;
        break;

      case vp_end_line:
        if (!in_cline) break;
        next_pos = in_cline->get_end();
        if (!next_pos.in_node){
            return_xy.at_x = in_cline->get_lrx() + in_viewport->get_hoffset();
            if (!set(return_xy))
                return raw_loc;
            }
        else {
            if (!set(next_pos))
                return raw_loc;
            }
        break;

      case vp_top:
        return_xy.at_x = 0;
        return_xy.at_y = 0;
        if (!set(return_xy))
            return raw_loc;
        break;

      case vp_bottom:
        return_xy.at_x = in_viewport->get_width();
        return_xy.at_y = in_viewport->get_hi();
        if (!set(return_xy))
            return raw_loc;
        break;

      default:
        break;
        }

    return raw_loc;
}

