/**************************************************************************************
    prune.cxx

    (c) Copyright Apollo Computer Inc., 1987 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

// Function gc_linearize()
//    Brings global order to goo by copying goo in traversal sequence to
//    If the argument model or its root is NULL, returns that model
// Function prune()
//    Prunes a model (gargage collection)
//    Takes a pointer to a model to be pruned as its argument
//    Returns a pointer to a pruned copy of the model or NULL on fatal error
//    The prefix "gc" on a function name indicates special relevance to garbage collection
//    If the argument model or its root is NULL, returns that model

#include <stdio.h>
#include <tml/tml_objects.hxx>

model*  prune(model *old_model)
{
    void   gc_linearize(model *);     // linearize goo
    long   gc_prune_data(node *);     // prune needless data-dominating terminals
                                      //   and childless branches
    void   gc_shorten_path(node *);   // shorten external paths - when parent has
                                      //   no properties, or parent and child have
                                      //   same properties, move child to left of
                                      //   parent; if parent becomes childless delete it
    model *gc_copy_model(model*);     // compact object storage

    model              *new_model;    // return value of this function

    history *past;
    long hist_size;
    node *old_root;

    if (NULL == old_model)
        return old_model;

    past = old_model->get_past();
    hist_size = past->get_hist_size();

    gc_linearize(old_model);            // wipes history, then linearizes goo
                                        // then restores history size

    past->set_hist_size(0);             // wipe history so reference counters won't be spoiled
                                        // (necessary redundancy because gc_linearize() may
                                        // be called on its own, and must restore history size
                                        // in that case)
    old_root = old_model->get_root();
    if (NULL == old_root)
        return old_model;
    if (-1 == gc_prune_data(old_root))
        return NULL;                   // fatal error from node::combine_data()
    gc_shorten_path(old_root);
    new_model = gc_copy_model(old_model);
    old_model->move_gheaps(new_model);
    delete old_model;

    // restore history to its former size

    past = new_model->get_past();
    hist_size = past->set_hist_size(hist_size);
    return new_model;                  // return final model
}


void gc_linearize(model* old_model)
{
    history    *past;
    long         hist_size;
    gheap_list *gheap_list_head,
               *new_gheap_list;
    goo_heap   *new_goo_heap;
    node       *next_node,
               *root;
    gnode      *as_goo_node;
    char       *byte_p;
    long        count,
               goo_dex;
    node_dir   dir;
    char      *junk;

    extern char *tml_os_byte_copy(char*, char*, long);

    if (NULL == old_model)
        return;

    // wipe history before compressing goo because
    //   goo pointers will change

    past = old_model->get_past();
    hist_size = past->get_hist_size();
    past->set_hist_size(0);

    gheap_list_head = new gheap_list;
    new_gheap_list = gheap_list_head;
    new_goo_heap = new goo_heap(old_model->get_gheap_size());
    gheap_list_head->set_heap(new_goo_heap);
    byte_p = new_goo_heap->get_start();
    goo_dex = 0;

    dir = n_down;
    if (NULL == (root = old_model->get_root()))
        return;
    if (NULL == (next_node = root->get_node(n_down)))
        return;
    do {
        if (next_node->has_goo()) {
            as_goo_node = (gnode *) next_node;
            count = as_goo_node->get_char_count();
            if (goo_dex + count >= old_model->get_gheap_size()) {
                new_goo_heap->set_byte(goo_dex);
                gheap_list_head = new gheap_list;
                gheap_list_head->set_next(new_gheap_list);
                new_gheap_list = gheap_list_head;
                new_goo_heap = new goo_heap(old_model->get_gheap_size());
                gheap_list_head->set_heap(new_goo_heap);
                byte_p = new_goo_heap->get_start();
                goo_dex = 0;
                }
            junk = tml_os_byte_copy((char*) as_goo_node->get_contents(), byte_p, count);
            as_goo_node->goo_ptr = byte_p;
            as_goo_node->count = count;
            byte_p += count;
            goo_dex += count;
            }
        do {
            next_node = next_node->get_next(dir);
            if (root == next_node)
                break;
            } while (n_up == dir);
        } while ( ! ((root == next_node) && (n_up == dir)));

    new_goo_heap->set_byte(goo_dex);
    while (NULL != old_model->current_gheap)
        delete old_model->current_gheap;
    old_model->first_gheap = gheap_list_head;
    old_model->set_current_gheap(new_goo_heap);

    // restore history to its former size

    hist_size = past->set_hist_size(hist_size);
}


/* prune needless gnodes and childless branches */
long gc_prune_data(node *root)
{
    node_dir dir;
    node *next_node, *was_node, *right_node;
    prop_vec *next_props, *right_props;
    long status;

    // degenerate case of childless root
    if (NULL == (next_node = root->get_node(n_down)))
        return 0;

    // prune childless branches
    dir = n_down;
    do {
        // branch
        if ( ! next_node->is_terminal()) {

            // childless branch
            if (NULL == next_node->get_node(n_down)) {

                // not root
                if (root != (was_node = next_node)) {

                    // prepare to get next node from left sibling
                    if (next_node->get_node(n_left)) {
                        next_node = next_node->get_node(n_left);
                        delete was_node;   // edits tree appropriately
                        }

                    // leftmost sibling; start again with parent
                    else {
                        next_node = next_node->get_node(n_up);
                        delete was_node;   // edits tree appropriately
                        continue;          // see if parent is now childless
                        }
                    }

                // childless root
                else
                    break;
                }
            }
        do {
            next_node = next_node->get_next(dir);
            if (root == next_node)
                break;
            } while (n_up == dir);
        } while ( ! ((root == next_node) && (n_up == dir)));

    // combine data from two nodes under one where possible

    dir = n_down;
    next_node = root->get_node(n_down);
    do {
        // terminal node
        if (next_node->is_terminal()) {

            // with right sibling
            if (NULL != (right_node = next_node->get_node(n_right))) {

                // that is also terminal
                if (right_node->is_terminal()) {
                    next_props = next_node->get_props();
                    right_props = right_node->get_props();

                    // and both have the same props or no props
                    if (((NULL == next_node->get_props())
                      && (NULL == right_node->get_props()))
                      || (next_props == right_props)) {

                        // gnode::combine_data edits tree appropriately
                        //   when data can be combined
                        status = next_node->combine_data(right_node);
                        if (-1 == status) // fatal error
                            return -1;
                        }
                    }
                }
            }
        do {
            next_node = next_node->get_next(dir);
            if (root == next_node)
                break;
            } while (n_up == dir);
        } while ( ! ((root == next_node) && (n_up == dir)));
    return 0;
}


/* shorten external paths - when parent has no properties, */
/*   or parent and child have the same properties, move child */
/*   to left of parent; if parent becomes childless, delete it */
void gc_shorten_path(node *root)
{
    node_dir dir;
    node *parent, *child;
    prop_vec *parent_props;

    // degenerate case of childless root
    if (NULL == (parent = root->get_node(n_down)))
        return;

    // starting with leftmost child of root
    dir = n_down;
    do {
        if (NULL != (child = parent->get_node(n_down))) {
            parent_props = parent->get_props();
            if ((NULL == parent_props)
              || (parent_props == child->get_props())) {
                child->move(parent, n_left);
                if (NULL == parent->get_node(n_down))
                    delete parent;
                parent = child;
                continue;       // continue with former child
                }
            }
        do {
            parent = parent->get_next(dir);
            if (root == parent)
                break;
            } while (n_up == dir);
        } while ( ! ((root == parent) && (n_up == dir)));
}


model* gc_copy_model(model *old_model)
{
    model*          new_model;
    node_dir        dir;
    node*           root, *old_prop_pos, *old_node, *new_node, *prev_old, *prev_new;
    gc_hash_tbl     hash_tbl;

    new_model = new model(old_model->get_oheap_size(), old_model->get_gheap_size(), old_model->get_sheap_size());
    root = old_model->get_root();

    // degenerate case of childless root
    if (NULL == root->get_node(n_down)) {
        new_node = root->clone(new_model, hash_tbl);
        new_model->set_root(new_node);
        new_model->set_prop_pos(new_node);
        return new_model;
        }

    old_prop_pos = old_model->get_prop_pos();
    prev_old = NULL;
    prev_new = NULL;
    dir = n_down;
    old_node = root;
    do {
        // clone this node
        new_node = old_node->clone(new_model, hash_tbl);

        // set new model parameters as appropriate
        if (old_node == root)
            new_model->set_root(new_node);
        if (old_node == old_prop_pos)
            new_model->set_prop_pos(new_node);

        // attach new node to new model appropriately
        if (prev_old) {
            if (old_node == prev_old->get_node(n_down))
                prev_new->set_node(new_node, n_down);
            else if (old_node == prev_old->get_node(n_right))
                prev_new->set_node(new_node, n_right);
            }

        // track last connection
        prev_old = old_node;
        prev_new = new_node;

        // get next node to clone
        do {
            old_node = old_node->get_next(dir);
            if (old_node == root)
                break;
            } while (n_up == dir);

        } while (! ((old_node == root) && (n_up == dir)));

    /* create new references to old_model's character sets */
    /*   in the new model */
    new_model->attach(old_model->get_char_sets());
    return new_model;
}
