/**************************************************************************************
    pvec_traits.cxx
    functions for the property vector class in the portable text management
      library

    (c) Copyright Apollo Computer Inc., 1987 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

#include <tml/tml_objects.hxx>

#define NULL 0

const long tml_pvec_size     =       (long) sizeof(prop_vec);

/*constructor for a property vector element*/
prop_vec::prop_vec(model *to_model)
{
    register prop_vec   *new_pvec;

    if (!this){
        if (to_model){
            new_pvec = to_model->get_current_oheap()->get_tf_pvec();
            if (!new_pvec)
                new_pvec = (prop_vec *)(to_model->get_current_oheap()->alloc(tml_pvec_size));
            else
                to_model->get_current_oheap()->tf_pvec = new_pvec->next;
            if (!new_pvec)
                return;
            this = new_pvec;
            }
        else {
            new_pvec = (prop_vec *)(new char [(long) sizeof(prop_vec)]);
            if (!new_pvec)
                return;
            else
                this = new_pvec;
            }
        }
    else
        this = this;
    in_model = to_model;
    ref_count = 1;
    next = 0;
    prop = 0;
}

/*destructor for a property vector element*/
prop_vec::~prop_vec()
{
    if (prop)
        detach();

    if (in_model) {
        in_model->current_oheap->free(this);
        this = 0;
        }
}
         
/* these two functions included to get pointers to overloaded detach() */
prop_vec* fp_d_pv(prop_vec* subject, prop_vec *from_vec)
{
    return(subject->detach(from_vec));
}

void fp_d(prop_vec* subject)
{
    subject->detach();
}
/* Call this function when a call to prop_vec::detach() returns 0 from */
/*   a request to detach a prop_vec from a node object's properties vector. */
/* Argument "on_node" is the node from whose properties member contained */
/*   only the one prop_vec "removed". */
void first_pvec_hist(node* on_node, prop_vec* removed)
{
    history  *past;
    atom     *this_atom;
    atom_arg *arg1, *arg2;

    if (NULL != on_node->get_model()) {
        past = on_node->get_model()->get_past();
        if (past->has_hist() && past->is_active()) {

            // if history is active, record this call.
            // NOTE: the prop_vec object's reference counter will
            //   have been handled properly by
            //   the earlier call to prop_vec::detach()

            arg1 = new atom_this_arg(past, on_node);
            arg2 = new atom_ptr_arg(past, removed);
            this_atom = new atom(past, (WHOLE_PTR) &node::prop_attach, (WHOLE_PTR) &node::prop_detach,
                                 arg1, arg2, NULL, arg1, NULL);
            past->add_atom(this_atom);
            }
        }
}

/*detach a single property vector element from the property vector beginning with from_vec*/
/*  WARNING: If the prop_vec being detached is the only property vector element in from_vec, */
/*  history is not written, but the return value is 0 in this case and this case only. */
/*  THEREFORE, if you get a 0 return from this function, you are obliged to add the */
/*  appropriate atom to history. A call to first_pvec_hist() (see above) will do this */
/*  for you. EXCEPTION: 0 returns from this function on undo of prop_vec::append are */
/*  not significant because history is not active during undo. */
prop_vec*  prop_vec::detach(prop_vec *from_vec)
{
    register prop_vec   *hold_vec,
                        *return_vec;

    history *past;
    long has_hist = 0;
    long hist_active = 0;
    atom* this_atom;
    atom_arg *arg1 = 0, *arg2 = 0, *arg3 = 0, *arg4 = 0, *arg5;
                                                                            
    /*first, check for trash*/
    if (!this || !from_vec)
        return 0;

    /*next, protect the model property state*/
    if (in_model)
        in_model->move_prop(0);

    // to keep ref_count straight, must have 'this' as both atom_this_arg and atom_ptr_arg

    if (NULL != get_model()) {
        past = get_model()->get_past();
        if (hist_active = ((has_hist = past->has_hist()) && past->is_active())) {

            // if history is active, prepare to record this call

            arg1 = new atom_this_arg(past, NULL);
            arg2 = new atom_this_arg(past, this);       // 'this' as object in redo call
            arg3 = new atom_ptr_arg(past, this);        // 'this' as argument for undo and redo calls
            arg4 = new atom_val_arg(past, from_vec);
            }
        }

    if (this == from_vec){
        return_vec = next;

        // no way to record history of prop_vec::detach() if only one prop_vec in vector
        //   return value of 0 signals this condition

        if (next && hist_active) {

            // complete recording this call of history

            arg5 = new atom_this_arg(past, next);
            // prevent repetitions on undo
            next = 0;
            this_atom = new atom(past, (WHOLE_PTR) &prop_vec::prepend, (WHOLE_PTR) &fp_d_pv,
                                 arg5, arg3, 0, arg1, arg2, arg4, 0);
            past->add_atom(this_atom);
            }
        else if (hist_active) {
            delete arg1;
            delete arg2;
            delete arg3;
            delete arg4;
            }
        }

    else {
        return_vec = from_vec;
        for (hold_vec = from_vec; hold_vec; hold_vec = hold_vec->next)
            if (hold_vec->next == this) break;
        if (!hold_vec) return(from_vec);            // no history because nothing done
                                                                                           
        hold_vec->next = next;
                                                                                           
        if (hist_active) {

            // complete recording this call of history

            arg5 = new atom_this_arg(past, next);
            // prevent repetitions on undo
            next = 0;
            this_atom = new atom(past, (WHOLE_PTR) &prop_vec::prepend, (WHOLE_PTR) &fp_d_pv,
                                 arg5, arg3, 0, arg1, arg2, arg4, 0);
            past->add_atom(this_atom);
            }
        else {
            delete arg1;
            delete arg2;
            delete arg3;
            delete arg4;
            }
    }

    if (0 == has_hist)
        ref_count--;
    if (!ref_count) delete(this);
    return(return_vec);
}


/*copy a property vector from one model to another*/
prop_vec* prop_vec::copy(model* to_model)
{
    register prop_vec* return_pvec;
                                                                           
    if (!to_model)
        to_model = in_model;
    return_pvec = to_model ? to_model->make_pvec() : new prop_vec;
    if (!return_pvec) return 0;

    return_pvec->in_model = to_model;
    if(prop)
        return_pvec->prop = prop->copy(to_model);
    if (!return_pvec->prop){
        delete return_pvec;
        return_pvec = 0;
        }
    return(return_pvec);
}

/*copy the entire property vector beginning with this to the indicated model*/
prop_vec* prop_vec::copy_all(model *to_model)
{
    register prop_vec *return_pvec,
                              *append_vec,
                              *hold_vec;

    if (!to_model)
        to_model = in_model;
    return_pvec = copy(to_model);
    if (!return_pvec) return 0;

    for(hold_vec = next;hold_vec; hold_vec = hold_vec->next){
        append_vec = hold_vec->copy(to_model);
        return_pvec->append(append_vec);
        }
    return(return_pvec);
}

/*make a new prop_vec element, sharing the property with the given prop_vec element*/
prop_vec* prop_vec::dup()
{
    register prop_vec   *return_pvec;
    return_pvec = in_model ? in_model->make_pvec() : new prop_vec;
    if (!return_pvec) return 0;

    return_pvec->in_model = in_model;
    return_pvec->ref_count = 1;
    return_pvec->next = 0;
    return_pvec->prop = prop;
    return return_pvec;
}

/*make a new property vector, with all properties shared*/
prop_vec* prop_vec::dup_all()
{
    register prop_vec   *return_pvec,
                              *append_vec,
                              *hold_vec;

    return_pvec = in_model? in_model->make_pvec() : new prop_vec;
    if (!return_pvec) return 0;
    return_pvec->ref_count = 1;
    return_pvec->next = 0;
    return_pvec->prop = prop;

    for(hold_vec = next;hold_vec; hold_vec = hold_vec->next){
        append_vec = hold_vec->dup();
        return_pvec->append(append_vec);
        }
    return(return_pvec);
}


/*prepend the property vector with head add_element to the property vector with head this*/
prop_vec* prop_vec::prepend(prop_vec* add_element)
{
    register prop_vec *return_vec,
                      *end_element;
    register long i;

    history *past;
    long has_hist = 0;
    long is_active = 0;
    prop_vec *cur_vec;
    prop_vec *next_vec;
    atom *this_atom;
    atom_arg *arg1, *arg2, *arg3, *arg4;

    // If add_element not in same model as "this", copy it to that model
    // NOTE: This block must precede block setting history

    if (in_model != add_element->in_model)
        return_vec = add_element->copy_all(in_model);
    else
        return_vec = add_element;

    /*protect the property state of the model*/
    if (in_model)
        in_model->move_prop(0);
    
    /*find the end element of the new head of the list, and attach the old list to it*/
    for(end_element = return_vec; end_element->next != 0; end_element = end_element->next);
    end_element->next = this;

    // NOTE: This block must follow block copying add_element to same model
    //   as "this" so history will be set properly

    if (NULL != get_model()) {
        past = get_model()->get_past();
        if ((has_hist = past->has_hist()) && (is_active = past->is_active())) {

            // If history is active, record this call

            for (cur_vec = return_vec, next_vec = return_vec->next;
              cur_vec != this;
              cur_vec = cur_vec->next, next_vec = next_vec->next) {
                arg1 = new atom_this_arg(past, NULL);
                arg2 = new atom_this_arg(past, cur_vec);
                arg3 = new atom_ptr_arg(past, return_vec);
                arg4 = new atom_this_arg(past, next_vec);
  
                /*this works because atoms are undone/redone in reverse order of creation*/

                this_atom = new atom(past, (WHOLE_PTR) &fp_d_pv, (WHOLE_PTR) &prop_vec::prepend,
                                     arg1, arg2, arg3, 0, arg4, arg3, 0);
                past->add_atom(this_atom);
                }
            }
        }

    /*update the ref counts on this to reflect references to the new head*/
    if ((is_active) || (0 == has_hist)) {
        i = end_element->ref_count - 1;
        if (i) {
            for(end_element = this; end_element != 0; end_element = end_element->next)
                end_element->ref_count += i;
            }
        }
    return(return_vec);
}
                                       
/*attach the property vector headed by add_vec to the property vector headed by this*/
void prop_vec::append(prop_vec *add_vec)
{
    register prop_vec *end_vec,
                      *new_end;
    register long i;

    history *past;
    long has_hist = 0;
    long is_active = 0;
    atom *this_atom;
    atom_arg *arg1, *arg2, *arg3;

    // If add_vec not in same model as "this", copy it to this model
    // NOTE: This block must precede setting of history

    if (in_model != add_vec->in_model)
        new_end = add_vec->copy_all(in_model);
    else
        new_end = add_vec;
                                             
    /*protect the property state of the model*/
    if (in_model)
        in_model->move_prop(0);

    /*now, find the end of the vector we are adding to*/
    for(end_vec = this; end_vec->next != 0; end_vec = end_vec->next);
    end_vec->next = new_end;

    // NOTE: This block must follow copy of add_vec so history will be set
    //   properly

    if (NULL != get_model()) {
        past = get_model()->get_past();
        if ((has_hist = past->has_hist()) && (is_active = past->is_active())) {

            // If history is active, record this call

            arg1 = new atom_this_arg(past, this);
            arg2 = new atom_ptr_arg(past, 0);
            arg3 = new atom_ptr_arg(past, new_end);
            this_atom = new atom(past, (WHOLE_PTR) &prop_vec::set_m_ptr, (WHOLE_PTR) &prop_vec::set_m_ptr,
                                 arg1, arg2, 0, arg1, arg3, 0);
            past->add_atom(this_atom);
            }
        }

    /*now, update the reference counts on the new end */
    if (is_active || (0 == has_hist)) {
        i = end_vec->ref_count - 1;
        if (i){
            for(end_vec = new_end; end_vec != 0; end_vec = end_vec->next)
                end_vec->ref_count += i;
            }
        }
}

  
/*attach a property to a property vector element*/
void prop_vec::attach(property *add_prop)
{
    register property *new_prop;

    history *past;
    long has_hist;
    long is_active;
    atom *this_atom;
    atom_arg *arg1, *arg2, *arg3;

    // If add_prop not in same model as "this" copy to this model
    // NOTE: This block must precede block setting history

    if (in_model != add_prop->locator.in_model)
        new_prop = add_prop->copy(in_model);
    else
        new_prop = add_prop;

    /*protect the property state of the model*/
    if (in_model)
        in_model->move_prop(0);

    // NOTE: This block must follow block copying add_prop to same
    //   model as "this" so history will be set properly

    if (NULL != get_model()) {
        past = get_model()->get_past();
        if ((has_hist = past->has_hist()) && (is_active = past->is_active())) {

            // If history is active, record this call

            arg1 = new atom_this_arg(past, NULL);
            arg2 = new atom_this_arg(past, this);
            arg3 = new atom_ptr_arg(past, new_prop);
            this_atom = new atom(past, (WHOLE_PTR) &fp_d, (WHOLE_PTR) &prop_vec::attach,
                                 arg1, arg2, 0, arg2, arg3, 0);
            past->add_atom(this_atom);
            }
        }

    prop = new_prop;
    if (is_active || (0 == has_hist))
        new_prop->ref_count++;
}

/*detach a property from a property vector element*/
void prop_vec::detach()
{
    history *past;
    long has_hist = 0;
    long is_active = 0;
    atom *this_atom;
    atom_arg *arg1, *arg2, *arg3;

    /*protect the property state of the model*/
    if (in_model)
        in_model->move_prop(0);

    if (NULL != get_model()) {
        past = get_model()->get_past();
        has_hist = past->has_hist();
        is_active = past->is_active();
        }
    if (0 == has_hist) {
        prop->ref_count--;
        if (!prop->ref_count)
            delete prop;
        }
    else if (is_active) {

        // If history is is active, record this call

        arg1 = new atom_this_arg(past, NULL);
        arg2 = new atom_this_arg(past, this);
        arg3 = new atom_ptr_arg(past, prop);
        this_atom = new atom(past, (WHOLE_PTR) &prop_vec::attach, (WHOLE_PTR) &fp_d,
                             arg2, arg3, 0, arg1, arg2, 0);
        past->add_atom(this_atom);
        }

    prop = 0;
}
                                                          
/*activate all the  properties in a property vector */
void prop_vec::activate()
{
    register prop_vec   *act_ele;

    for (act_ele = this; act_ele; act_ele = act_ele->next)
        act_ele->prop->activate();
}

/*deactivate all of the properties--this is recursive so properties will be deactiveated
    in reverse order from their activation*/
void prop_vec::deactivate()
{
    if (next)
        next->deactivate();
    prop->deactivate();
}

