/**************************************************************************************
     save_restore.cxx
     functions for model save and restore to/from disk

     (c) Copyright Apollo Computer Inc., 1987 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

#include <stdio.h>
#include <string.h>
#ifdef __cplusplus
#include <stdlib.h>
#endif
#include <tml/tml_objects.hxx>
#include <tml/tml_os_depends.hxx>
#include <tml/tml_dp_graph.hxx>

/* function used in garbage collection to move previously linearized goo */
/*   from old model to new */
void model::move_gheaps(model *to_model)
{
    gheap_list *cur_glist;

    to_model->current_gheap = current_gheap;
    to_model->first_gheap = first_gheap;
    for (cur_glist = first_gheap; NULL != cur_glist; cur_glist = cur_glist->get_next())
        (cur_glist->get_heap())->attach(to_model);
    current_gheap = NULL;
    first_gheap = NULL;
}


/* returns, as heap*, a new heap of a given type and size */
/* FOR USE ONLY IN RESTORING OBJECTS FROM DISK */
heap* new_restore_heap(long heap_type, long size)
{
    switch (heap_type) {
        case OBJ_HEAP:
            return(new obj_heap(size));
        case GOO_HEAP:
            return(new goo_heap(size));
        case ST_HEAP:
            return(new st_heap(size));
        case SQUEEZE_HEAP:
            return(new squeeze_heap(size));
        case BIT_HEAP:
            return(new read_bit_heap(size));
        default:
            printf("impossible heap type\n");
            return(NULL);
        }
}


/* RESTORE FUNCTIONS FOR MODEL-RESIDENT HEAPS */

/* definition of virtual function restore in base class heap. SHOULD NEVER BE CALLED. */
/* Returns error indicator -1. */
long heap::restore(void* file_handle, d_model_parts* model_parts, long index)
{
    return(-1);
}


/* restore object heap from disk. Returns -1 on failure, number of bytes read on success */
long obj_heap::restore(void* file_handle, d_model_parts* model_parts, long index)
{
    char* start_addr;
    long bytes_read;

    start_addr = get_start();
    *(model_parts->heaps[OBJ_HEAP] + index) = start_addr;
    set_byte(*(model_parts->bytes_used[OBJ_HEAP] + index));
    if (1 == model_parts->version_number) {
        if (-1 == (bytes_read = tml_os_read(file_handle,
                                start_addr,
                                *(model_parts->bytes_used[OBJ_HEAP] + index)))) {
            printf("obj_heap::restore: read_failed\n");
            return(-1);
            }
        }
    attach(model_parts->for_model);
    model_parts->for_model->attach(this);
    if (index == (long) model_parts->cur_heap[OBJ_HEAP])
        model_parts->for_model->set_current_oheap(this);
    return(bytes_read);
}


/* restore goo heap from disk. Returns -1 on failure, number of bytes read on success */
long goo_heap::restore(void* file_handle, d_model_parts* model_parts, long index)
{
    char *start_addr;
    long bytes_read;

    start_addr = get_start();
    *(model_parts->heaps[GOO_HEAP] + index) = start_addr;
    set_byte(*(model_parts->bytes_used[GOO_HEAP] + index));
    if (-1 == (bytes_read = tml_os_read(file_handle,
                          start_addr,
                          *(model_parts->bytes_used[GOO_HEAP] + index)))) {
        printf("goo_heap::restore: read_failed\n");
        return(-1);
        }
    attach(model_parts->for_model);
    model_parts->for_model->attach(this);
    if (index == (long) model_parts->cur_heap[GOO_HEAP])
        model_parts->for_model->set_current_gheap(this);
    return(bytes_read);
}


/* restore string heap from disk. Returns -1 on failure, number of bytes read on success */
long st_heap::restore(void* file_handle, d_model_parts* model_parts, long index)
{
    char *start_addr;
    long bytes_read;

    start_addr = get_start();
    *(model_parts->heaps[ST_HEAP] + index) = start_addr;
    set_byte(*(model_parts->bytes_used[ST_HEAP] + index));
    if (-1 == (bytes_read = tml_os_read(file_handle,
                                        start_addr,
                                        *(model_parts->bytes_used[ST_HEAP] + index)))) {
        printf("st_heap::restore: read_failed\n");
        return(-1);
        }
    attach(model_parts->for_model);
    model_parts->for_model->attach(this);
    if (index == (long) model_parts->cur_heap[ST_HEAP])
        model_parts->for_model->set_current_sheap(this);
    return(bytes_read);
}


/* RESTORE FUNCTIONS FOR NON-MODEL-RESIDENT HEAPS */

/* restore squeeze heap from disk. Returns -1 on failure, number of bytes read on success */
long squeeze_heap::restore(void* file_handle, d_model_parts* model_parts, long index)
{
    char *start_addr;
    long bytes_read;

    start_addr = get_start();
    *(model_parts->heaps[SQUEEZE_HEAP] + index) = start_addr;
    if (-1 == (bytes_read = tml_os_read(file_handle,
                                        start_addr,
                                        *(model_parts->bytes_used[SQUEEZE_HEAP] + index)))) {
        printf("squeeze_heap::restore: read_failed\n");
        return(-1);
        }

    model_parts->q_object_data(this);
    if (index == (long) model_parts->cur_heap[SQUEEZE_HEAP])
        model_parts->cur_squeeze_heap = this;
    return(bytes_read);
}


/* restore graphics data heap from disk. Returns -1 on failure, number of bytes read on success */
long read_bit_heap::restore(void* file_handle, d_model_parts* model_parts, long index)
{
    char *start_addr;
    long bytes_read;

    start_addr = get_start();
    *(model_parts->heaps[BIT_HEAP] + index) = start_addr;
    set_byte(*(model_parts->bytes_used[BIT_HEAP] + index));
    if (-1 == (bytes_read = tml_os_read(file_handle,
                                        start_addr,
                                        *(model_parts->bytes_used[BIT_HEAP] + index)))) {
        printf("read_graf_heap::restore: read_failed\n");
        return(-1);
        }
    model_parts->q_bit_data(this);
    return(bytes_read);
}


/* for garbage collection: copy node to new model, hashing for previously copied */
/*   components. SHOULD NEVER BE CALLED DIRECTLY, only through implementations for */
/*     derived node classes. */
node* node::clone(model* new_model, gc_hash_tbl& hash_tbl, node* return_node)
{
    node*         clone_node;
    gc_hash_elem* hash_elem;
    prop_vec*     next_prop;
    prop_vec*     new_prop;
    prop_vec*     prev_prop;

    // should be called only from derived class
    if (NULL == return_node) {
        printf("node::clone: return_node argument missing\n");
        return NULL;
        }

    // not previously hashed
    else if (NULL == (clone_node = (node*) hash_tbl.find(this))) {
        hash_elem = hash_tbl.insert(this);
        hash_elem->set_new(return_node);

        // clone's in_model member was set by derived node class constructor
        
        // set left link from clone, right link to clone
        if (left && (clone_node = (node*) hash_tbl.find(left))) {
            // if clone_node returns as NULL means 'this' is a node living in
            //   the model tree that is cited here as value of subtree
            //   argument, but hasn't yet been encountered in the tree;
            //   in that case, left of return_node is set in bnode::clone
            //   when 'this' is encountered as a node in the tree
            return_node->left = clone_node;
            clone_node->right = return_node;
            }
        else
            return_node->left = NULL;
        
        // set up link from clone, down link to clone if any
        if (up && (clone_node = (node*) hash_tbl.find(up))) {
            // if clone_node returns as NULL means 'this' is a node living in
            //   the model tree that is cited here as value of subtree
            //   argument, but hasn't yet been encountered in the tree;
            //   in that case, up of return_node is set in bnode::clone
            //   when 'this' is encountered as a node in the tree
            return_node->up = clone_node;
            if (this == up->get_node(n_down))
                clone_node->set_node(return_node, n_down);
            }
        
        // clone properties
        prev_prop = NULL;
        next_prop = properties;
        while (next_prop) {
            if (NULL == (new_prop = (prop_vec*) hash_tbl.find(next_prop))) {
                new_prop = next_prop->clone(new_model, hash_tbl);
                next_prop = next_prop->get_next();
                }
            else     // non-null means remainder of vector has been cloned
                next_prop = NULL;
            if (prev_prop)
                prev_prop->append(new_prop);
            prev_prop = new_prop;
            }
        if (properties)
            return_node->properties = (prop_vec*) hash_tbl.find(properties);
        return_node->sib_order = sib_order;
        return_node->n_type = n_type;
        }
    return return_node;
}


/* the _vptr member of derived class members is a per-program quantity when */
/*    libraries are bound in to executables. Update the _vptr member of the */
/*    argument node as appropriate. */
void node_set_vptr(node* new_node)
{
    tml_type *this_type; 

    for (this_type = csm_type_q_get_head(); this_type; this_type = this_type->next) {
        if ((this_type == new_node->n_type) && (NULL != this_type->instance)) {
#ifdef __cplusplus
            new_node->__vptr = ((node *) this_type->instance)->__vptr;
#else
            new_node->_vptr = ((node *) this_type->instance)->_vptr;
#endif
            return;
            }
        }
}


/* saves a node to disk. SHOULD NEVER BE CALLED DIRECTLY, only through implementations for */
/*   derived node classes. */
void node::save(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, d_node* savee)
{
    d_ptr         d_ptr_tmp;
    gc_hash_elem* hash_elem;
    prop_vec*     next_prop;

    // should be called only from derived class
    if (0 == savee)
        printf("node::save: savee argument missing\n");

    // not previously hashed
    else if (NULL == (d_node*) hash_tbl.find(this)) {
        d_ptr_tmp.get_d_ptr((char*) this, model_parts.heaps[OBJ_HEAP]);
        hash_elem = hash_tbl.insert(this);
        hash_elem->set_new((void*) d_ptr_tmp.as_whole_ptr());
        if (right)
            d_ptr_tmp.get_d_ptr((char*) right, model_parts.heaps[OBJ_HEAP]);
        else
            d_ptr_tmp.get_d_ptr((char*) NULL, (CHARP *) NULL);
        savee->set_right(&d_ptr_tmp);
        
        if (properties) {
            next_prop = properties;
            while (next_prop) {
                if (NULL == hash_tbl.find(next_prop)) {
                    next_prop->save(model_parts, hash_tbl);
                    next_prop = next_prop->get_next();
                    }
                else
                    break; // non-null means remainder of vector saved previously
                }
            d_ptr_tmp = hash_tbl.find(properties);
            }
        else
            d_ptr_tmp.get_d_ptr((char*) NULL, (CHARP *) NULL);
        savee->set_pv_d_ptr(&d_ptr_tmp);
        savee->set_type_index(this);
        }
}

void node::restore(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, d_node* restoree)
{
    d_ptr d_ptr_tmp;
    gc_hash_elem* hash_elem;
    prop_vec* next_prop;

    if (0 == restoree) {
        printf("node::restore: restoree argument missing\n");
        return;
        }

    restoree->unscramble(model_parts.their_endian);

  
    d_ptr_tmp = restoree->get_right();
    right = (node*) d_ptr_tmp.get_m_ptr(model_parts.heaps[OBJ_HEAP]);
    if (right) {
        right->left = this;
        right->up = up;
        }

    d_ptr_tmp = restoree->get_props();
    properties = (prop_vec*) d_ptr_tmp.get_m_ptr(model_parts.heaps[OBJ_HEAP]);
    next_prop = properties;
    while (next_prop) {
        if (NULL == hash_tbl.find(next_prop)) {
            hash_elem = hash_tbl.insert(next_prop);
            hash_elem->set_new((void*) -1);
            next_prop->init_ref_count();
            next_prop->set_model(get_model());
            next_prop->restore(model_parts, hash_tbl);
            }
        else
            next_prop->incr_ref_count();
        next_prop = next_prop->get_next();
        }

    node_plid_trace = NULL;

    sib_order = 1;
    if (left)
        sib_order += left->sib_order;
}


/* for garbage collection: copy node to new model, consulting/constructing hash */
/*   table of new objects created */
node *bnode::clone(model* new_model, gc_hash_tbl& hash_tbl, node* return_node)
{
    node* clone_node;

    // hashing performed in base class
    // clone's down pointer if any is set by its first child

    // not previously hashed
    if (NULL == (clone_node = (node*) hash_tbl.find(this))) {
        // not called from derived class
        if (NULL == return_node)
            return_node = new bnode(new_model);
        return_node = node::clone(new_model, hash_tbl, return_node);
        return return_node;
        }

    // already hashed and cloned, but if is value of subtree_arg
    //   (root of subtree) must set clone's up and left pointers,
    //   clone's left's right, and clone's up's down if this is
    //   first child of up
    else {
        return_node = clone_node;
        // up (and up's down) may not yet have been set
        if (up && (NULL == return_node->get_up()) &&
          (clone_node = (node*) hash_tbl.find(up))) {
            return_node->set_up(clone_node);
            if (this == up->get_node(n_down))
                clone_node->set_node(return_node, n_down);
            }
        // left (and left's right) may not yet have been set
        if (left && (NULL == return_node->get_left()) &&
          (clone_node = (node*) hash_tbl.find(left))) {
            return_node->set_left(clone_node);
            clone_node->set_right(return_node);
            }
        return return_node;
        }
}


/* for save to disk. */
void bnode::save(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, d_node* savee)
{
    d_bnode* d_bnode_savee;
    d_ptr    d_ptr_tmp;

    // hashing performed in base class

    // not previously hashed
    if (NULL == (d_node*) hash_tbl.find(this)) {
        // not called from derived class
        if (0 == savee)
            savee = d_bnode_savee = (d_bnode*) model_parts.obj_alloc((long) sizeof(d_bnode));
        else
            d_bnode_savee = (d_bnode*) savee;
        if (down_ptr)
            d_ptr_tmp.get_d_ptr((char*) down_ptr, model_parts.heaps[OBJ_HEAP]);
        else
            d_ptr_tmp.get_d_ptr((char*) NULL, (CHARP *) NULL);
        d_bnode_savee->set_down_ptr(&d_ptr_tmp);
        node::save(model_parts, hash_tbl, savee);
        }

    // remaining case:
    // already hashed and saved, but if is value of subtree_arg
    //   must set up's savee's down (if 'this' is the first child of up)
    //   and left's savee's right
    // up and left may not yet have been saved when 'this' was saved
    //   so hash of their values would have returned NULL, meaning
    //   no disk objects yet existed for them, and hence their down
    //   and right disk pointers respectively could not have been set.
    else {
        if (up && (d_bnode_savee = (d_bnode*) hash_tbl.find(up))) {
            d_ptr_tmp = d_bnode_savee->get_down_ptr();
            if (d_ptr_tmp.is_null()) {
                d_ptr_tmp.get_d_ptr((char*) this, model_parts.heaps[OBJ_HEAP]);
                d_bnode_savee->set_down_ptr(&d_ptr_tmp);
                }
            }
        if (left && (d_bnode_savee = (d_bnode*) hash_tbl.find(left))) {
            d_ptr_tmp = d_bnode_savee->get_right();
            if (d_ptr_tmp.is_null()) {
                d_ptr_tmp.get_d_ptr((char*) this, model_parts.heaps[OBJ_HEAP]);
                d_bnode_savee->set_right(&d_ptr_tmp);
                }
            }
        }
}


void bnode::restore(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, d_node* restoree)
{
    d_ptr         d_ptr_tmp;
    d_bnode*      d_bnode_restoree;

    if (0 == restoree)
        restoree = d_bnode_restoree = (d_bnode*) model_parts.obj_realloc((long) sizeof(d_bnode));
    else
        d_bnode_restoree = (d_bnode*) restoree;

    d_bnode_restoree->unscramble(model_parts.their_endian);

    d_ptr_tmp = d_bnode_restoree->get_down_ptr();
    down_ptr = (node*) d_ptr_tmp.get_m_ptr(model_parts.heaps[OBJ_HEAP]);
    if (down_ptr) {
        down_ptr->set_left(NULL);
        down_ptr->set_up(this);
        }

    node::restore(model_parts, hash_tbl, restoree);
}


/* set memory version of leaf node's goo pointer */
/* if with_node is a goo node in the same model and immediately */
/*   to the right of this, and it has goo that immediately follows */
/*   this node's goo within the same heap, increment this gnode's */
/*   char count and delete with_node */
/* returns 0 on success, -2 if data cannot be combined, -1 on fatal error */
long gnode::combine_data(node *with_node)
{
    gheap_list* cur_list;
    goo_heap* this_gheap;
    gnode* with_gnode;

    if (get_model() != with_node->get_model())
        return(-1);
    if (gobj != with_node->get_type())
        return(-2);
    for (cur_list = get_model()->get_first_gheap(); cur_list; cur_list = cur_list->get_next()) {
        this_gheap = cur_list->get_heap();
        if ((get_contents() >= this_gheap->get_start())
          && (get_contents() < this_gheap->get_start() + this_gheap->get_size()))
            break;
        }
    if (NULL == cur_list)
        return(-1);
    with_gnode = (gnode*) with_node;
    if ( ! ((with_gnode->get_contents() >= this_gheap->get_start())
      && (with_gnode->get_contents() < (this_gheap->get_start() + this_gheap->get_size()))))
        return(-2);
    if ((((char*) get_contents()) + get_char_count()) != with_gnode->get_contents())
        return(-2);
    set_char_count(get_char_count() + with_gnode->get_char_count());
    delete with_node;
    return(0);
}


/* for garbage collection: copy node to new model, consulting/constructing hash */
/*   table of new objects created */
node *gnode::clone(model* new_model, gc_hash_tbl& hash_tbl, node* return_node)
{
    // hashing performed in base class

    // not previously hashed
    if (NULL == (node*) hash_tbl.find(this)) {
        // not called from derived class
        if (NULL == return_node)
            return_node = new gnode(new_model);
        return_node = node::clone(new_model, hash_tbl, return_node);
        ((gnode*) return_node)->goo_ptr = goo_ptr;
        ((gnode*) return_node)->count = count;
        }
    return return_node;
}


void gnode::save(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, d_node* savee)
{
    d_ptr    d_ptr_tmp;
    d_gnode* d_gnode_savee;

    // hashing performed in base class

    //not previously hashed
    if (NULL == (d_node*) hash_tbl.find(this)) {
        // not called from derived class
        if (0 == savee)
            savee = d_gnode_savee = (d_gnode*) model_parts.obj_alloc((long) sizeof(d_gnode));
        else
            d_gnode_savee = (d_gnode*) savee;
        if (goo_ptr)
            d_ptr_tmp.get_d_ptr((char*) goo_ptr, model_parts.heaps[GOO_HEAP]);
        else
            d_ptr_tmp.get_d_ptr((char*) NULL, (CHARP *) NULL);
        d_gnode_savee->set_goo_d_ptr(&d_ptr_tmp);
        d_gnode_savee->set_count(count);
        node::save(model_parts, hash_tbl, savee);
        }
}


void gnode::restore(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, d_node* restoree)
{
    d_ptr         d_ptr_tmp;
    d_gnode*      d_gnode_restoree;
    char_set*     cur_char_set;
    unsigned long char_size;

    if (0 == restoree)
        restoree = d_gnode_restoree = (d_gnode*) model_parts.obj_realloc((long) sizeof(d_gnode));
    else
        d_gnode_restoree = (d_gnode*) restoree;

    d_gnode_restoree->unscramble(model_parts.their_endian);

    d_ptr_tmp = d_gnode_restoree->get_goo_ptr();
    goo_ptr = (char*) d_ptr_tmp.get_m_ptr(model_parts.heaps[GOO_HEAP]);

    count = d_gnode_restoree->get_count();

    node::restore(model_parts, hash_tbl, restoree);

    // characters of length 2 or 4 bytes must be unscrambled

    if (cur_char_set = (char_set*) get_model()->inq_state(this, tml_char_set_prop)) {
        char_size = cur_char_set->get_char_size();
        switch (char_size) {
            case 2:
                goo_ptr = (char*) model_parts.their_endian->short_arr_shuf((unsigned short*) goo_ptr, count >> 1);
                break;
            case 4:
                goo_ptr = (char*) model_parts.their_endian->long_arr_shuf((unsigned long*) goo_ptr, count >> 2);
                break;
            default:
                break;
            }
        }
}


/* for garbage collection: copy node to new model, consulting/constructing hash */
/*   table of new objects created */
node *pgnode::clone(model* new_model, gc_hash_tbl& hash_tbl, node* return_node)
{
    // hashing performed in base class

    // not previously hashed
    if (NULL == (node*) hash_tbl.find(this)) {
        // not called from derived class
        if (NULL == return_node)
            return_node = new pgnode(new_model);
        return gnode::clone(new_model, hash_tbl, return_node);
        }
    return return_node;
}


/* for garbage collection: copy prop_vec to new model, consulting/constructing */
/*   hash table of new objects created */
prop_vec* prop_vec::clone(model* new_model, gc_hash_tbl& hash_tbl)
{
    prop_vec*     return_vec;
    gc_hash_elem* hash_elem;

    // not previously hashed
    if (NULL == (prop_vec*) hash_tbl.find(this)) {
        return_vec = new prop_vec(new_model);
        hash_elem = hash_tbl.insert(this);
        hash_elem->set_new(return_vec);
        return_vec->prop = prop->clone(new_model, hash_tbl);
        // next_member set by node::clone or other caller
        // in_model member set by constructor
        return_vec->ref_count = ref_count;
        }
    return return_vec;
}


/* for disk save */
void prop_vec::save(d_model_parts& model_parts, gc_hash_tbl& hash_tbl)
{
    d_ptr         d_ptr_tmp;
    void*         saved;
    gc_hash_elem* hash_elem;
    d_prop_vec*   savee;

    // not previously hashed
    if (NULL == (d_prop_vec*) hash_tbl.find(this)) {
        savee = (d_prop_vec*) model_parts.obj_alloc((long) sizeof(d_prop_vec));
        d_ptr_tmp.get_d_ptr((char*) this, model_parts.heaps[OBJ_HEAP]);
        hash_elem = hash_tbl.insert(this);
        hash_elem->set_new((void*) d_ptr_tmp.as_whole_ptr());
        if (next)
            d_ptr_tmp.get_d_ptr((char*) next, model_parts.heaps[OBJ_HEAP]);
        else
            d_ptr_tmp.get_d_ptr((char*) NULL, (CHARP *) NULL);
        savee->set_d_ptr(&d_ptr_tmp);
        
        if (prop) {
            if (NULL == (saved = hash_tbl.find(prop))) {
                prop->save(model_parts, hash_tbl);
                d_ptr_tmp = hash_tbl.find(prop);
                }
            else
                d_ptr_tmp = saved;
            }
        else
            d_ptr_tmp.get_d_ptr((char*) NULL, (CHARP *) NULL);
        savee->set_prop_d_ptr(&d_ptr_tmp);
        }
}

/* the _vptr member of derived class members is a per-program quantity when */
/*    libraries are bound in to executables. Update the _vptr member of the */
/*    argument prop as appropriate. */
void prop_set_vptr(property* new_prop)
{

    tml_type *this_type;

    for (this_type = csm_type_q_get_head(); this_type; this_type = this_type->next) {
        if ((this_type == new_prop->get_index()) && (NULL != this_type->instance)) {
#ifdef __cplusplus
            new_prop->__vptr = ((property*) this_type->instance)->__vptr;
#else
            new_prop->_vptr = ((property*) this_type->instance)->_vptr;
#endif
            return;
        }
    }
}


void prop_vec::restore(d_model_parts& model_parts, gc_hash_tbl& hash_tbl)
{
    d_ptr          d_ptr_tmp;
    gc_hash_elem*  hash_elem;
    d_prop_vec*    restoree;
    p_index        scrambled_p_index;
    font_property* new_prop;

    restoree = (d_prop_vec*) model_parts.obj_realloc((long) sizeof(d_prop_vec));

    restoree->unscramble(model_parts.their_endian);

    d_ptr_tmp = restoree->get_next();
    next = (prop_vec*) d_ptr_tmp.get_m_ptr(model_parts.heaps[OBJ_HEAP]);

    d_ptr_tmp = restoree->get_prop();
    prop = (property*) d_ptr_tmp.get_m_ptr(model_parts.heaps[OBJ_HEAP]);
    if (prop) {
        if (NULL == hash_tbl.find(prop)) {
            hash_elem = hash_tbl.insert(prop);
            hash_elem->set_new((void*) -1);
            scrambled_p_index = ((d_property*) model_parts.obj_noalloc())->get_type_index();
            prop->set_index((p_index) model_parts.their_endian->my_long((unsigned long) scrambled_p_index));
            prop->set_index(csm_type_q_get_type((long) prop->get_index(), model_parts.types));
            if ((4 > model_parts.version_number) && (tml_font_prop == prop->get_index())) {
                new_prop = (font_property*) in_model->make_prop((long) sizeof(font_property));
                prop_set_vptr(prop);
                prop->init_ref_count();
                prop->set_model(get_model());
                new_prop->restore_v3(model_parts, hash_tbl, prop);
                prop = new_prop;
                }
            else {
                prop_set_vptr(prop);
                prop->init_ref_count();
                prop->set_model(get_model());
                prop->restore(model_parts, hash_tbl);
                }
            }
        else
            prop->incr_ref_count();
        }
}


/* for garbage collection: copy property to new model, consulting/constructing */
/*   hash table of new objects created */
property* property::clone(model* new_model, gc_hash_tbl& hash_tbl, property* return_prop)
{
    gc_hash_elem* hash_elem;
    argument*     arg_vec_heads[2];
    argument*     next_arg;
    argument*     new_arg;
    argument*     prev_arg;
    long          i;

    // not previously hashed
    if (NULL == (property*) hash_tbl.find(this)) {
        // not called from derived class
        if (NULL == return_prop)
            return_prop = new property(new_model);
        hash_elem = hash_tbl.insert(this);
        hash_elem->set_new(return_prop);
        // locator.in_model member set by constructor
        return_prop->index = index;
        return_prop->ref_count = ref_count;
        arg_vec_heads[0] = entry_arg;
        arg_vec_heads[1] = exit_arg;
        for (i = 0; i < 2; i++) {
            prev_arg = NULL;
            next_arg = arg_vec_heads[i];
            while (next_arg) {
                if (NULL == (new_arg = (argument*) hash_tbl.find(next_arg))) {
                    new_arg = next_arg->clone(new_model, hash_tbl);
                    next_arg = next_arg->get_next();
                    }
                else     // non-null means remainder of vector has been cloned
                    next_arg = NULL;
                if (prev_arg)
                    prev_arg->append(new_arg);
                prev_arg = new_arg;
                }
            }
        if (entry_arg)
            return_prop->entry_arg = (argument*) hash_tbl.find(entry_arg);
        if (exit_arg)
            return_prop->exit_arg = (argument*) hash_tbl.find(exit_arg);
        }
    return return_prop;
}


/* for disk save */
void property::save(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, d_property* savee)
{
    d_ptr         d_ptr_tmp;
    gc_hash_elem* hash_elem;
    argument*     arg_vec_heads[2];
    argument*     next_arg;
    long          i;

    // not previously hashed
    if (NULL == (d_property*) hash_tbl.find(this)) {
        // not called from derived class
        if (0 == savee)
            savee = (d_property*) model_parts.obj_alloc((long) sizeof(d_property));
        d_ptr_tmp.get_d_ptr((char*) this, model_parts.heaps[OBJ_HEAP]);
        hash_elem = hash_tbl.insert(this);
        hash_elem->set_new((void*) d_ptr_tmp.as_whole_ptr());

        arg_vec_heads[0] = entry_arg;
        arg_vec_heads[1] = exit_arg;
        
        for (i = 0; i < 2; i++) {
            next_arg = arg_vec_heads[i];
            while (next_arg) {
                if (NULL == hash_tbl.find(next_arg)) {
                    next_arg->save(model_parts, hash_tbl);
                    next_arg = next_arg->get_next();
                    }
                else
                    break; // non-null means remainder of vector saved previously
                }
            }
        
        d_ptr_tmp = hash_tbl.find(entry_arg);
        savee->set_entry_d_ptr(&d_ptr_tmp);
        
        if (exit_arg)
            d_ptr_tmp = hash_tbl.find(exit_arg);
        else
            d_ptr_tmp.get_d_ptr((char*) NULL, (CHARP *) NULL);
        savee->set_exit_d_ptr(&d_ptr_tmp);
        
        savee->set_type_index(this);
        }
}

/* the _vptr member of derived class members is a per-program quantity when */
/*    libraries are bound in to executables. Update the _vptr member of the */
/*    argument as appropriate. */
void arg_set_vptr(argument* arg)
{
    tml_type *this_type;

    for (this_type = csm_type_q_get_head(); this_type; this_type = this_type->next) {
        if ((this_type == arg->get_vtype()) && (NULL != this_type->instance)) {
#ifdef __cplusplus
            arg->__vptr = ((argument*) this_type->instance)->__vptr;
#else
            arg->_vptr = ((argument*) this_type->instance)->_vptr;
#endif
            return;
            }
        }
}


void property::restore(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, d_property* restoree)
{
    d_ptr         d_ptr_tmp;
    gc_hash_elem* hash_elem;
    argument*     arg_vec_heads[2];
    argument*     next_arg;
    long          i;
    arg_type      scrambled_a_type;

    if (0 == restoree)
        restoree = (d_property*) model_parts.obj_realloc((long) sizeof(d_property));

    restoree->unscramble(model_parts.their_endian);

    d_ptr_tmp = restoree->get_entry_arg();
    arg_vec_heads[0] = entry_arg = (argument*) d_ptr_tmp.get_m_ptr(model_parts.heaps[OBJ_HEAP]);
    d_ptr_tmp = restoree->get_exit_arg();
    arg_vec_heads[1] = exit_arg = (argument*) d_ptr_tmp.get_m_ptr(model_parts.heaps[OBJ_HEAP]);

    for (i = 0; i < 2; i++) {
        next_arg = arg_vec_heads[i];
        while (next_arg) {
            if (NULL == hash_tbl.find(next_arg)) {
                hash_elem = hash_tbl.insert(next_arg);
                hash_elem->set_new((void*) -1);
                scrambled_a_type = ((d_argument*) model_parts.obj_noalloc())->get_type_index();
                next_arg->set_vtype((arg_type) model_parts.their_endian->my_long((unsigned long) scrambled_a_type));
                next_arg->set_vtype(csm_type_q_get_type((long) next_arg->get_vtype(), model_parts.types));
                arg_set_vptr(next_arg);
                next_arg->init_ref_count();
                next_arg->set_model(get_model());
                next_arg->restore(model_parts, hash_tbl);
                }
            else
                next_arg->incr_ref_count();
            next_arg = next_arg->get_next();
            }
        }

    act_value = NULL;
}


property* font_property::clone(model* new_model, gc_hash_tbl& hash_tbl, property* return_prop)
{
    return_prop = new_model->make_prop((long) sizeof(font_property));
    property::clone(new_model, hash_tbl, return_prop);
    ((font_property*) return_prop)->set_ftbl(ftbl);
    return return_prop;
}


void font_property::save(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, d_property* savee)
{
    argument*   next_arg;
    long        grafsys_index;

    // not previously hashed
    if (NULL == (d_property*) hash_tbl.find(this)) {
        // not called from derived class
        if (0 == savee)
            savee = (d_property*) model_parts.obj_alloc((long) sizeof(d_bnode));
        if (ftbl) {
            next_arg = entry_arg;
            next_arg = next_arg->get_next()->get_next();
            if (next_arg)
                /* add one to preserve zero for the default */
                grafsys_index = 1 + csm_type_q_get_index((grafsys) ((int_arg*) next_arg)->get_val());
            else
                grafsys_index = 0;
            }
        else
            grafsys_index = 0;
        ((int_arg*) next_arg)->set_val(grafsys_index);
        property::save(model_parts, hash_tbl, savee);
        }
}


void font_property::restore(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, d_property* restoree)
{
    argument* next_arg;
    long      grafsys_index;

    // restoree argument should be zero unless called from derived class
    //   which uses disk-format class larger than d_property
    property::restore(model_parts, hash_tbl, restoree);
    next_arg = entry_arg;
    next_arg = entry_arg->get_next()->get_next();
    if (next_arg) {
        if (0 != (grafsys_index = (long) next_arg->get_val()))
            /* subtract one to compensate for addition on save */
            /*   that preserves zero to stand for the default value */
            ftbl = (vpftbl*) (csm_type_q_get_type(--grafsys_index, model_parts.types))->instance;
        }
    else
        ftbl = 0;
}


/* for TMLib Version 4 and beyond, restore font properties as class font_property */
void font_property::restore_v3(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, property* prop_v3)
{
    set_model(prop_v3->get_model());
    ref_count = prop_v3->get_ref_count();
    // p_index set by constructor
    attach_entry(prop_v3->get_entry_av());
    attach_exit(prop_v3->get_exit_av());
    // act_value set by constructor
    ftbl = 0;
#ifdef __cplusplus
    property::restore(model_parts, hash_tbl);
#else
    // Sun C++ 1.2 cannot generate default arguments
    //   unless they are declared in both header file 
    //   declaration and source code definition
    property::restore(model_parts, hash_tbl, (d_property*) NULL);
#endif
}

/* for garbage collection: copy argument to new model, consulting/constructing */
/*   hash table of new objects created */
argument* argument::clone(model* new_model, gc_hash_tbl& hash_tbl, argument* return_arg)
{
    gc_hash_elem* hash_elem;

    // should be called only from derived class
    if (NULL == return_arg) {
        printf("argument::clone: return_arg argument missing\n");
        return NULL;
        }
    // not previously hashed
    if (NULL == (argument*) hash_tbl.find(this)) {
        hash_elem = hash_tbl.insert(this);
        hash_elem->set_new(return_arg);
        }
    // next member set by property::clone or other caller
    // in_model and a_type members set by constructor
    return_arg->ref_count = ref_count;
    return return_arg;
}


/* for disk_save */
void argument::save(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, d_argument* savee)
{
    gc_hash_elem* hash_elem;
    d_ptr         d_ptr_tmp;

    if (0 == savee) {
        printf("argument::save: savee argument missing\n");
        return;
        }
    // not previously hashed
    else if (NULL == (d_argument*) hash_tbl.find(this)) {
        d_ptr_tmp.get_d_ptr((char*) this, model_parts.heaps[OBJ_HEAP]);
        hash_elem = hash_tbl.insert(this);
        hash_elem->set_new((void*) d_ptr_tmp.as_whole_ptr());
        if (next)
            d_ptr_tmp.get_d_ptr((char*) next, model_parts.heaps[OBJ_HEAP]);
        else
            d_ptr_tmp.get_d_ptr((char*) NULL, (CHARP *) NULL);
        savee->set_d_ptr(&d_ptr_tmp);
        savee->set_type_index(this);
        }
}


void argument::restore(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, d_argument* restoree)
{
    d_ptr d_ptr_tmp;

    if (0 == restoree) {
        printf("argument::restore: restoree argument missing\n");
        return;
        }

    /* restoree is unscrambled by derived class before this call */

    d_ptr_tmp = restoree->get_next();
    next = (argument*) d_ptr_tmp.get_m_ptr(model_parts.heaps[OBJ_HEAP]);

    return;
}


/* for garbage collection: copy argument to new model, consulting/constructing */
/*   hash table of new objects created */
argument* int_arg::clone(model* new_model, gc_hash_tbl& hash_tbl, argument* return_arg)
{
    // not previously hashed
    if (NULL == (return_arg = (argument*) hash_tbl.find(this))) {
        // not called from derived class
        if (NULL == return_arg)
            return_arg = new_model->make_arg(*((long*) arg_val));
        argument::clone(new_model, hash_tbl, return_arg);
        }
    return return_arg;
}


void int_arg::save(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, d_argument* savee)
{
    d_ptr d_ptr_tmp;
    char *dst_int;
    char *src_int;

    //not previously hashed
    if (NULL == (d_argument*) hash_tbl.find(this)) {
        //not called from derived class
        if (0 == savee)
            savee = (d_argument*) model_parts.obj_alloc((long) sizeof(d_argument));
        if (arg_val) {
            d_ptr_tmp.get_d_ptr((char*) arg_val, model_parts.heaps[OBJ_HEAP]);
            dst_int = (char*) model_parts.obj_alloc((long) sizeof(long));
            src_int = (char*) arg_val;
            if (dst_int != tml_os_byte_copy(src_int, dst_int, (long) sizeof(long)))
                printf("int_arg::save: tml_os_byte_copy failed\n");
            }
        else
            d_ptr_tmp.get_d_ptr((char*) NULL, (CHARP *) NULL);
        savee->set_val_d_ptr(&d_ptr_tmp);
        
        argument::save(model_parts, hash_tbl, savee);
        }
}


void int_arg::restore(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, d_argument* restoree)
{
    d_ptr d_ptr_tmp;
    char*  dst_int;
    char*  src_int;

    if (0 == restoree)
        restoree = (d_argument*) model_parts.obj_realloc((long) sizeof(d_argument));

    restoree->unscramble(model_parts.their_endian);

    d_ptr_tmp = restoree->get_val();
    arg_val = (long*) d_ptr_tmp.get_m_ptr(model_parts.heaps[OBJ_HEAP]);
    if (arg_val) {
        dst_int = (char*) arg_val;
        src_int = (char*) model_parts.obj_realloc((long) sizeof(long));
        if (dst_int != tml_os_byte_copy(src_int, dst_int, (long) sizeof(long)))
            printf("int_arg::restore: tml_os_byte_copy failed\n");
        *((long*) dst_int) = (long) model_parts.their_endian->my_long(*((unsigned long*) dst_int));
        }
    argument::restore(model_parts, hash_tbl, restoree);
}


/* for garbage collection: copy argument to new model, consulting/constructing */
/*   hash table of new objects created */
argument* float_arg::clone(model* new_model, gc_hash_tbl& hash_tbl, argument* return_arg)
{
    // not previously hashed
    if (NULL == (argument*) hash_tbl.find(this)) {
        // not called from derived class
        if (NULL == return_arg)
            return_arg = new_model->make_arg(*((float*) arg_val));
        argument::clone(new_model, hash_tbl, return_arg);
        }
    return return_arg;
}


void float_arg::save(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, d_argument* savee)
{
    d_ptr  d_ptr_tmp;
    char  *dst_float;
    float *src_float;

    // not previously hashed
    if (NULL == (d_argument*) hash_tbl.find(this)) {
        //not called from derived class
        if (0 == savee)
            savee = (d_argument*) model_parts.obj_alloc((long) sizeof(d_argument));
        if (arg_val) {
            d_ptr_tmp.get_d_ptr((char*) arg_val, model_parts.heaps[OBJ_HEAP]);
        
            // allocate space for ASCII representation of float value and
            //   byte for null terminator
        
            dst_float = (char*) model_parts.obj_alloc((long) (MAX_FLOAT_BUF_V2 + 1));
            src_float = (float*) arg_val;
        
            // acquire ASCII representation of float value and put it in dst_float
            // precision is specified as maximum possible after allowing for
            //   ordinal sign if any, integer part digit, decimal point,
            //   mantissa sign, letter "e" and two-digit exponent
        
            sprintf(dst_float, "%*.*e", MAX_FLOAT_BUF_V2, MAX_FLOAT_BUF_V2 - 7, *src_float);
            }
        else
            d_ptr_tmp.get_d_ptr((char*) NULL, (CHARP *) NULL);
        savee->set_val_d_ptr(&d_ptr_tmp);
        
        argument::save(model_parts, hash_tbl, savee);
        }
}


void float_arg::restore(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, d_argument* restoree)
{
    d_ptr  d_ptr_tmp;
    float* dst_float;
    char*  src_float;

    if (0 == restoree)
        restoree = (d_argument*) model_parts.obj_realloc((long) sizeof(d_argument));

    restoree->unscramble(model_parts.their_endian);

    d_ptr_tmp = restoree->get_val();
    arg_val = (float*) d_ptr_tmp.get_m_ptr(model_parts.heaps[OBJ_HEAP]);
    if (arg_val) {
        dst_float = (float*) arg_val;
        src_float = (char*) model_parts.obj_realloc((long) sizeof(MAX_FLOAT_BUF_V2 + 1));
        *(dst_float) = (float) atof(src_float);
        }

    argument::restore(model_parts, hash_tbl, restoree);
}


/* for garbage collection: copy argument to new model, consulting/constructing */
/*   hash table of new objects created */
argument* string_arg::clone(model* new_model, gc_hash_tbl& hash_tbl, argument* return_arg)
{
    // not previously hashed
    if (NULL == (return_arg = (argument*) hash_tbl.find(this))) {
        // not called from derived class
        if (NULL == return_arg)
            return_arg = new_model->make_arg((char*) arg_val, count);
        else
            ((string_arg*) return_arg)->set_val((char*) arg_val, count);
        argument::clone(new_model, hash_tbl, return_arg);
        }
    return return_arg;
}


void string_arg::save(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, d_argument* savee)
{
    d_ptr         d_ptr_tmp;
    d_string_arg* d_string_arg_savee;

    // not previously hashed
    if (NULL == (d_argument*) hash_tbl.find(this)) {
        //not called from derived class
        if (0 == savee)
            savee = d_string_arg_savee = (d_string_arg*) model_parts.obj_alloc((long) sizeof(d_string_arg));
        else
            d_string_arg_savee = (d_string_arg*) savee;
        if (arg_val)
           d_ptr_tmp.get_d_ptr((char*) arg_val, model_parts.heaps[ST_HEAP]);
        else
           d_ptr_tmp.get_d_ptr((char*) NULL, (CHARP *) NULL);
        savee->set_val_d_ptr(&d_ptr_tmp);
       
        d_string_arg_savee->set_count(count);
        
        argument::save(model_parts, hash_tbl, savee);
        }
}


void string_arg::restore(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, d_argument* restoree)
{
    d_ptr         d_ptr_tmp;
    d_string_arg* d_string_arg_restoree;

    if (0 == restoree)
        restoree = d_string_arg_restoree = (d_string_arg*) model_parts.obj_realloc((long) sizeof(d_string_arg));
    else
        d_string_arg_restoree = (d_string_arg*) restoree;

    d_string_arg_restoree->unscramble(model_parts.their_endian);

    d_ptr_tmp = restoree->get_val();
    arg_val = (char*) d_ptr_tmp.get_m_ptr(model_parts.heaps[ST_HEAP]);

    count = d_string_arg_restoree->get_count();

    argument::restore(model_parts, hash_tbl, restoree);
}


argument* subtree_arg::clone(model* new_model, gc_hash_tbl& hash_tbl, argument* return_arg)
{
    node*     old_node, *new_node, *prev_old, *prev_new;
    node_dir  dir;

    // not previously hashed
    if (NULL == (argument*) hash_tbl.find(this)) {
        if (arg_val) {
            // call clone on subtree root even if already hashed to ensure
            //   that any left and/or up node points to it
            if (NULL != hash_tbl.find((node*) arg_val)) {
                new_node = ((node*) arg_val)->clone(new_model, hash_tbl);
                return return_arg;
                }
            // clone the argument value subtree
            prev_old = NULL;
            prev_new = NULL;
            dir = n_down;
            old_node = (bnode*) arg_val;
            do {
                // nonnull means remainder of tree has been cloned
                if (NULL != hash_tbl.find(old_node))
                    break;
                new_node = old_node->clone(new_model, hash_tbl);
                if (prev_old) {
                    if (old_node == prev_old->get_node(n_down))
                        prev_new->set_node(new_node, n_down);
                    else if (old_node == prev_old->get_node(n_right))
                        prev_new->set_node(new_node, n_right);
                    }
                prev_old = old_node;
                prev_new = new_node;
                do {
                    old_node = old_node->get_next(dir);
                    if (old_node == (bnode*) arg_val)
                        break;
                    } while (n_up == dir);
                } while (! ((old_node == (bnode*) arg_val) && (n_up == dir)));
            
            // not called from derived class
            if (NULL == return_arg)
                return_arg = new_model->make_arg((node*) hash_tbl.find(arg_val));
            else
                ((subtree_arg*) return_arg)->set_val((node*) hash_tbl.find(arg_val));
            }
        else {
            // not called from derived class
            if (NULL == return_arg)
                return_arg = new_model->make_arg((node*) NULL);
            else
                ((subtree_arg*) return_arg)->set_val(NULL);
            }
        argument::clone(new_model, hash_tbl, return_arg);
        }
    return return_arg;
}


void subtree_arg::save(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, d_argument* savee)
{
    void    *new_arg_val;
    d_ptr    d_ptr_tmp;
    void    *saved;
    node_dir dir;
    node    *old_node;

    // not previously hashed
    if (NULL == (d_argument*) hash_tbl.find(this)) {
        //not called from derived class
        if (0 == savee)
            savee = (d_argument*) model_parts.obj_alloc((long) sizeof(d_argument));
        if (arg_val) {
            if (model_parts.for_model != ((bnode*) arg_val)->get_model())
                new_arg_val = ((bnode*) arg_val)->copy_all(model_parts.for_model);
            else
                new_arg_val = arg_val;
            if (NULL == (saved = hash_tbl.find(new_arg_val))) {
                dir = n_down;
                old_node = (bnode*) new_arg_val;
                do {
                    // old_node becomes NULL when subtree is single node
                    // nonnull return from gc_hash_tbl::find() means 
                    //   remainder of tree has been saved
                    if ((NULL == old_node) || (NULL != hash_tbl.find(old_node)))
                        break;
                    old_node->save(model_parts, hash_tbl);
                    do {
                        old_node = old_node->get_next(dir);
                        if ((NULL == old_node) || (old_node == (bnode*) new_arg_val))
                            break;
                        } while (n_up == dir);
                    } while (! ((old_node == (bnode*) new_arg_val) && (n_up == dir)));
                d_ptr_tmp.get_d_ptr((char*) new_arg_val, model_parts.heaps[OBJ_HEAP]);
                }
            else
                d_ptr_tmp = saved;
            }
        else
            d_ptr_tmp.get_d_ptr((char*) NULL, (CHARP *) NULL);
        savee->set_val_d_ptr(&d_ptr_tmp);
        argument::save(model_parts, hash_tbl, savee);
        }
}


void subtree_arg::restore(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, d_argument* restoree)
{
    d_ptr         d_ptr_tmp;
    gc_hash_elem* hash_elem;
    node*         next_node;
    node_dir      dir;
    node_type     scrambled_ntype;

    if (0 == restoree)
        restoree = (d_argument*) model_parts.obj_realloc((long) sizeof(d_argument));

    restoree->unscramble(model_parts.their_endian);

    d_ptr_tmp = restoree->get_val();
    arg_val = (node*) d_ptr_tmp.get_m_ptr(model_parts.heaps[OBJ_HEAP]);

    if (arg_val) {
        if (NULL == hash_tbl.find(arg_val)) {

            next_node = (node*) arg_val;
            next_node->up = NULL;
            next_node->left = NULL;

            dir = n_down;
            do {
                if (NULL == next_node)
                    break;
                else if (NULL == hash_tbl.find(next_node)) {
                    hash_elem = hash_tbl.insert(arg_val);
                    hash_elem->set_new((void*) -1);
                    scrambled_ntype = ((d_node*) model_parts.obj_noalloc())->get_type_index();
                    next_node->set_type((node_type) model_parts.their_endian->my_long((unsigned long) scrambled_ntype));
                    next_node->set_type(csm_type_q_get_type((long) next_node->get_type(), model_parts.types));
                    node_set_vptr(next_node);
                    next_node->set_model(get_model());
                    next_node->restore(model_parts, hash_tbl);
                    }
                do {
                    next_node = next_node->get_next(dir);
                    if ((NULL == next_node) || (next_node == (node*) arg_val))
                        break;
                    } while (n_up == dir);
                } while (! ((next_node == ((node*) arg_val)) && (n_up == dir)));
            }
        }

    argument::restore(model_parts, hash_tbl, restoree);
}

argument* char_set_arg::clone(model* new_model, gc_hash_tbl& hash_tbl, argument* return_arg)
{
    // not previously hashed
    if (NULL == (argument*) hash_tbl.find(this)) {
        // not called from derived class
        if (NULL == return_arg)
            // make_arg increments reference count on char_set
            //   through call to char_set_arg::set_val
            return_arg = new_model->make_arg((char_set*) arg_val);
        argument::clone(new_model, hash_tbl, return_arg);
        }
    return return_arg;
}

void char_set_arg::save(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, d_argument* savee)
{
    d_ptr d_ptr_tmp;

    // hashing done in base class

    // not previously hashed
    if (NULL == (d_argument*) hash_tbl.find(this)) {
        // not called from derived class
        if (0 == savee)
            savee = (d_argument*) model_parts.obj_alloc((long) sizeof(d_argument));

        // If arg_val is nonnull, assume previous call to prepare_cs_binary()
        //   saved the char_set. Assume also that char_set is properly
        //   registered in model

        if (arg_val)
            d_ptr_tmp = hash_tbl.find(arg_val);
        else
            d_ptr_tmp.get_d_ptr((char*) NULL, (CHARP *) NULL);
        savee->set_val_d_ptr(&d_ptr_tmp);
        argument::save(model_parts, hash_tbl, savee);
        }
}


void char_set_arg::restore(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, d_argument* restoree)
{
    d_ptr         d_ptr_tmp;

    if (0 == restoree)
        restoree = (d_argument*) model_parts.obj_realloc((long) sizeof(d_argument));

    restoree->unscramble(model_parts.their_endian);

    // If arg_val is nonnull, assume previous call to read_ucs_binary_file()
    //   restored the union_char_set. Assume also that union_char_set is
    //   properly registered in model.

    d_ptr_tmp = restoree->get_val();
    if (d_ptr_tmp.is_null())
        arg_val = NULL;
    else
        arg_val = hash_tbl.find((void*) d_ptr_tmp.as_whole_ptr());

    argument::restore(model_parts, hash_tbl, restoree);
}


argument* union_char_set_arg::clone(model* new_model, gc_hash_tbl& hash_tbl, argument* return_arg)
{
    // not previously hashed
    if (NULL == (argument*) hash_tbl.find(this)) {
        // not called from derived class
        if (NULL == return_arg)
            // make_arg increments reference count on union_char_set
            //   through call to union_char_set_arg::set_val
            return_arg = new_model->make_arg((union_char_set*) arg_val);
        argument::clone(new_model, hash_tbl, return_arg);
        }
    return return_arg;
}

void union_char_set_arg::save(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, d_argument* savee)
{
    d_ptr d_ptr_tmp;

    // hashing done in base class

    // not previously hashed
    if (NULL == (d_argument*) hash_tbl.find(this)) {
        //not called from derived class
        if (0 == savee)
            savee = (d_argument*) model_parts.obj_alloc((long) sizeof(d_argument));

        // If arg_val is nonnull, assume previous call to prepare_cs_binary()
        //   saved the union_char_set. Assume also that char_set is properly
        //   registered in model

        if (arg_val)
            d_ptr_tmp = hash_tbl.find(arg_val);
        else
            d_ptr_tmp.get_d_ptr((char*) NULL, (CHARP *) NULL);
        savee->set_val_d_ptr(&d_ptr_tmp);
        argument::save(model_parts, hash_tbl, savee);
        }
}


void union_char_set_arg::restore(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, d_argument* restoree)
{
    d_ptr         d_ptr_tmp;

    if (0 == restoree)
        restoree = (d_argument*) model_parts.obj_realloc((long) sizeof(d_argument));

    restoree->unscramble(model_parts.their_endian);

    // If arg_val is nonnull, assume previous call to read_ucs_binary_file()
    //   restored the union_char_set. Assume also that union_char_set is
    //   properly registered in model.

    d_ptr_tmp = restoree->get_val();
    if (d_ptr_tmp.is_null())
        arg_val = NULL;
    else
        arg_val = hash_tbl.find((void*) d_ptr_tmp.as_whole_ptr());

    argument::restore(model_parts, hash_tbl, restoree);
}

/*******************************************************************/
/* unscrambling functions for TML file headers d_model_info,       */
/*   d_model_info_v4, and d_model_info_v1                          */
/*******************************************************************/

void d_model_info::unscramble(endian* their_endian)
{
    unsigned long* ptr;
    long           heap_type;

    version_number = (long) their_endian->my_long((unsigned long) version_number);
    root.unscramble(their_endian);
    num_types = (long) their_endian->my_long((unsigned long) num_types);
    ptr = their_endian->long_arr_shuf((unsigned long*) num_heaps, NUM_HEAP_TYPES);
    ptr = their_endian->long_arr_shuf((unsigned long*) heap_quanta, NUM_HEAP_TYPES);
    for (heap_type = 0; heap_type < NUM_HEAP_TYPES; heap_type++)
        cur_heap[heap_type] = their_endian->my_half_ptr(cur_heap[heap_type]);
}


void d_model_info_v4::unscramble(endian* their_endian)
{
    unsigned long* ptr;
    long           heap_type;

    version_number = (long) their_endian->my_long((unsigned long) version_number);
    root.unscramble(their_endian);
    num_types = (long) their_endian->my_long((unsigned long) num_types);
    ptr = their_endian->long_arr_shuf((unsigned long*) num_heaps, SQUEEZE_HEAP + 1);
    ptr = their_endian->long_arr_shuf((unsigned long*) heap_quanta, SQUEEZE_HEAP + 1);
    for (heap_type = 0; heap_type < SQUEEZE_HEAP + 1; heap_type++)
        cur_heap[heap_type] = their_endian->my_half_ptr(cur_heap[heap_type]);
}


void d_model_info_v1::unscramble(endian* their_endian)
{
    version_number = (long) their_endian->my_long((unsigned long) version_number);
    root.unscramble(their_endian);
    num_types = (long) their_endian->my_long((unsigned long) num_types);
    num_oheaps = (long) their_endian->my_long((unsigned long) num_oheaps);
    num_gheaps = (long) their_endian->my_long((unsigned long) num_gheaps);
    num_sheaps = (long) their_endian->my_long((unsigned long) num_sheaps);
    last_o = (long) their_endian->my_long((unsigned long) last_o);
    last_g = (long) their_endian->my_long((unsigned long) last_g);
    last_s = (long) their_endian->my_long((unsigned long) last_s);
    cur_oheap = their_endian->my_half_ptr(cur_oheap);
    cur_gheap = their_endian->my_half_ptr(cur_gheap);
    cur_sheap = their_endian->my_half_ptr(cur_sheap);
    first_oheap = their_endian->my_half_ptr(first_oheap);
    first_gheap = their_endian->my_half_ptr(first_gheap);
    first_sheap = their_endian->my_half_ptr(first_sheap);
}


/*******************************************************************/
/* member functions for class d_model_parts tracking structure for */
/*   save and restore                                              */
/*******************************************************************/


/* constructor for class d_model_parts used in saving tml models to disk */
d_model_parts::d_model_parts(model* write_model)
{
    long         i;
    long         heap_type;
    long*        size_arr;
    long*        bytes_used_arr;
    CHARP*       heap_arr;
    heap_key*    key_arr;
    heap_key*    last_key;
    char*        cur_heap_char_p;
    d_ptr        d_ptr_tmp;

    heap_info = new d_heap_info(write_model);

    for_write = 0xFFFFFFFF;
    for_model = write_model;
    version_number = tml_version_number;
    num_types = 0;
    tml_types = NULL;
    load_types();
    num_plids = 0;
    d_plid_col = NULL;

    for (heap_type = 0; heap_type < NUM_MODEL_HEAP_TYPES; heap_type++) {

        // count heaps of this type in the model

        num_heaps[heap_type] = count_heaps(heap_info->first_list[heap_type]);

        // complete allocation of heap_info members based on this figure

        heap_info->complete_for_heap_type(heap_type, num_heaps[heap_type]);

        // create heap tracking arrays and initialize them
        // heap_arr has two more elements than there are heaps of the given
        //   type in the model. In the next to last slot put the sum of the
        //   numerically highest heap origin and the size of that heap
        //   (sorting is performed by call to d_heap_info::complete_for_heap_type
        //   above). The last slot in the array gets a NULL pointer. This sorted
        //   array will be used by member function get_d_ptr of class d_ptr to
        //   derive "disk pointers" from memory addresses of objects in the model.

        size_arr
            = heap_sizes[heap_type]
            = new long[num_heaps[heap_type]];
        bytes_used_arr
            = bytes_used[heap_type]
            = new long[num_heaps[heap_type]];
        heap_arr
            = heaps[heap_type]
            = new CHARP[num_heaps[heap_type] + 2];
        key_arr = heap_info->sort_array[heap_type];

        for (i = 0; i < num_heaps[heap_type]; i++) {
            *size_arr++ = key_arr->heap_size;
            *bytes_used_arr++ = key_arr->bytes_used;
            *heap_arr++ = key_arr->start_addr;
            ++key_arr;
            }

        last_key = heap_info->sort_array[heap_type] + num_heaps[heap_type] - 1;

        // The increment must be size of heap, not number of bytes
        //   used, because a heap might be allocated without having
        //   any bytes in it used (as is the case at model construction).
        //   If the increment were bytes used, attempts in this case to
        //   derive the index of the current heap would fail with a "pointer
        //   out of range" error in d_ptr::get_d_ptr().

        *heap_arr++ = last_key->start_addr + last_key->heap_size;

        *heap_arr = NULL;

        // derive the index of the last heap allocated of this type and record it

        cur_heap_char_p = heap_info->cur_heap[heap_type]->get_start();
        d_ptr_tmp.get_d_ptr(cur_heap_char_p, heaps[heap_type]);
        cur_heap[heap_type] = d_ptr_tmp.get_heap();
        }

                                               
    their_magic = TML_MAGIC_3210;
    their_endian = new endian_3210;

    heap_quanta[OBJ_HEAP] = write_model->get_oheap_size();
    heap_quanta[GOO_HEAP] = write_model->get_gheap_size();
    heap_quanta[ST_HEAP] = write_model->get_sheap_size();
    heap_quanta[SQUEEZE_HEAP] = write_model->get_oheap_size();
    heap_quanta[BIT_HEAP] = 0;

    for (heap_type = NUM_MODEL_HEAP_TYPES; heap_type < NUM_HEAP_TYPES; heap_type++) {
        num_heaps[heap_type] = 0;
        heap_quanta[heap_type] = 0;
        cur_heap[heap_type] = 0;
        heap_sizes[heap_type] = NULL;
        bytes_used[heap_type] = NULL;
        heaps[heap_type] = NULL;
        }

    cur_squeeze_heap = new squeeze_heap(write_model->get_oheap_size());
    squeeze_data = NULL;
    q_object_data(cur_squeeze_heap);

    bit_data = NULL;
}


/* constructor for class d_model_parts used in restoring tml models from disk */
d_model_parts::d_model_parts()
{
    long         i;

    heap_info = NULL;

    for_model = NULL;
    for_write = 0;
    version_number = tml_version_number;
    root.get_d_ptr((char*) NULL, (CHARP*) NULL);
    num_types = 0;
    types = NULL;
    tml_types = NULL;
    num_plids = 0;
    d_plid_col = NULL;
    for (i = 0; i < NUM_HEAP_TYPES; i++) {
        num_heaps[i] = 0;
        heap_quanta[i] = 0;
        cur_heap[i] = 0;
        heap_sizes[i] = NULL;
        bytes_used[i] = NULL;
        heaps[i] = NULL;
        }
    cur_squeeze_heap = NULL;
    squeeze_data = NULL;
    bit_data = NULL;

    their_magic = 0;
    their_endian = NULL;
}


/* destructor for class d_model_parts */
d_model_parts::~d_model_parts()
{
    long heap_type;
    long i;

    if (for_write) {
        heap_info->prepare_for_destruction(this);
        delete heap_info;
        }
    heap_info = NULL;

    delete types;
    types = NULL;
    delete tml_types;
    tml_types = NULL;
    delete d_plid_col;
    d_plid_col = NULL;
    for_model = NULL;
    for (heap_type = NUM_MODEL_HEAP_TYPES; heap_type <= LAST_EPHEMERAL_HEAP_TYPE; heap_type++) {
        for (i = 0; i < num_heaps[heap_type]; i++)
            delete *(heaps[heap_type] + i);
        }
    for (heap_type = 0; heap_type < NUM_MODEL_HEAP_TYPES; heap_type++) {
        for (i = 0; i < num_heaps[heap_type] + ((for_write) ? 2 : 0); i++)
            *(heaps[heap_type] + i) = NULL;
        }
    for (heap_type = NUM_MODEL_HEAP_TYPES; heap_type < NUM_HEAP_TYPES; heap_type++) {
        for (i = 0; i < num_heaps[heap_type]; i++)
            *(heaps[heap_type] + i) = NULL;
        }
    for (heap_type = 0; heap_type < NUM_HEAP_TYPES; heap_type++) {
        delete bytes_used[heap_type];
        bytes_used[heap_type] = NULL;
        delete heap_sizes[heap_type];
        heap_sizes[heap_type] = NULL;
        delete heaps[heap_type];
        heaps[heap_type] = NULL;
        }
    cur_squeeze_heap = NULL;
    squeeze_data = NULL;
    bit_data = NULL;

    their_magic = 0;
    delete their_endian;
    their_endian = NULL;
}


/* function to set types member for model save */
void d_model_parts::load_types()
{
    long      i;
    tml_type *cur_type;

    cur_type = csm_type_q_get_head();
    while(cur_type) {
        ++num_types;
        cur_type = cur_type->next;
        }
    types = new tml_ext_type[num_types];
    cur_type = csm_type_q_get_head();
    for (i = 0; i < num_types; i++) {
        *(types + i) = *cur_type->type;      // bit copy
        cur_type = cur_type->next;
        }
}


/* function to count heaps of each type and set appropriate value */
/*   in d_model_parts for model write */
long d_model_parts::count_heaps(heap_list *cur_list)
{
    long retval;

    retval = 0;
    while (NULL != cur_list) {
        ++retval;
        cur_list = cur_list->get_next();
        }
    return(retval);
}


/* function to save the objects in a model by converting their in-memory */
/*   pointers to on-disk equivalents while transcribing each object into a */
/*   corresponding and smaller on-disk form */
void d_model_parts::save_pointers(cs_heap*& csm_heaps, gc_hash_tbl*& hash_tbl)
{
    node            *next_node,
                    *root;
    node_dir        dir;

    /* save character sets, hashing character and union character set ids */
    /*   for later use in saving char_set_arg and union_set_arg */
    if (for_model->get_char_sets())
        for_model->get_char_sets()->prepare_ucs_binary(hash_tbl, csm_heaps, types);

    next_node = root = for_model->get_root();
    dir = n_down;
    do {
        next_node->save(*this, *hash_tbl);
        do {
            next_node = next_node->get_next(dir);
            if ((NULL == next_node) || (root == next_node))
                break;
            } while (n_up == dir);
        } while ( ! ((next_node == root) && (n_up == dir)));
}


/* function to complete initialization of d_model_parts object based on */
/*   information gained during save of objects in its associated model */
/* also, complete heap_info structure for types of heaps created during */
/*   pointer conversion */
void d_model_parts::complete_for_save()
{
    long       i;
    long       heap_type;
    long*      size_arr;
    long*      bytes_used_arr;
    CHARP*     heap_arr;
    heap_list* cur_list;
    heap*      this_heap;

    num_heaps[SQUEEZE_HEAP] = count_heaps(squeeze_data);
    num_heaps[BIT_HEAP] = count_heaps(bit_data);

    heap_info->first_list[SQUEEZE_HEAP] = squeeze_data;
    heap_info->first_list[BIT_HEAP] = bit_data;

    for (heap_type = NUM_MODEL_HEAP_TYPES; heap_type < NUM_HEAP_TYPES; heap_type++) {
        size_arr
            = heap_sizes[heap_type]
            = new long[num_heaps[heap_type]];
        bytes_used_arr
            = bytes_used[heap_type]
            = new long[num_heaps[heap_type]];
        heap_arr
            = heaps[heap_type]
            = new CHARP[num_heaps[heap_type]];
        cur_list = heap_info->first_list[heap_type];
        for (i = num_heaps[heap_type] - 1; i >= 0; i--) {
            this_heap = cur_list->get_heap();
            *(size_arr + i) = this_heap->get_size();
            *(bytes_used_arr + i) = this_heap->get_byte();
            *(heap_arr + i) = this_heap->get_start();
            cur_list = cur_list->get_next();
            }

        // non-model-resident heaps are created and queued as needed, with the
        // most recently created at the head of the queue. They will be needed
        // on restore in order of creation. They are written to disk in reverse
        // order of creation and queued on restore with most recently restored
        // at the head of the queue, thus in the proper order. The index of the
        // current heap for non-model-resident heaps on restore is therefore
        // always zero.

        cur_heap[heap_type] = 0;
        }

    root.get_d_ptr((char *) for_model->get_root(), heaps[OBJ_HEAP]);
}


/* function to write model to disk */
long d_model_parts::save(void* file_handle, cs_heap*& csm_heaps, gc_hash_tbl*& hash_tbl)
{
    long           bytes_written = 0;
    long           total_bytes_written = 0;
    char*          cur_heap_char_p;
    long           num_bytes;
    long           heap_type;
    long           i, j;
    char**         bytes_arr;
    long*          num_bytes_arr;
    csm_lock**     locks;
    bit_heap_list* cur_bit_heap_list;

    // 1. Write the TML magic number

    if (-1 == (bytes_written = tml_os_write(file_handle, (char*) &their_magic, (long) sizeof(long))))
        return(-1);
    else
        total_bytes_written = bytes_written;

    // 2. Write general model information

    if (-1 == (bytes_written = tml_os_write(file_handle, (char *) ((d_model_info *) this), (long) sizeof(d_model_info))))
        return(-1);
    else
        total_bytes_written = bytes_written;

    // 3. Write the types

    if (-1 == (bytes_written = tml_os_write(file_handle, (char *) types,
                                            num_types * (long) sizeof(tml_ext_type))))
        return(-1);
    else
        total_bytes_written += bytes_written;

    // 4. Write the character sets

    if (for_model->get_char_sets()) {
        if (-1 == (bytes_written =
          for_model->get_char_sets()->write_ucs_binary_file(file_handle, hash_tbl, csm_heaps, types)))
            return(-1);
        else
            total_bytes_written += bytes_written;
        }
    else {
        i = 0;
        if (-1 == (bytes_written = tml_os_write(file_handle, (char*) &i,
                                                sizeof(long))))
            return(-1);
        else
            total_bytes_written += bytes_written;
        }

    // 5. Write the arrays showing size of each heap
                                        
    for (heap_type = 0; heap_type < NUM_HEAP_TYPES; heap_type++) {
        if (-1 == (bytes_written = tml_os_write(file_handle, (char*) heap_sizes[heap_type],
                                               (long) sizeof(long) * num_heaps[heap_type])))
            return(-1);
        else
            total_bytes_written += bytes_written;
        }

    // 6. Write the arrays showing number of bytes used in each heap

    for (heap_type = 0; heap_type < NUM_HEAP_TYPES; heap_type++) {
        if (-1 == (bytes_written = tml_os_write(file_handle, (char*) bytes_used[heap_type],
                                               (long) sizeof(long) * num_heaps[heap_type])))
            return(-1);
        else
            total_bytes_written += bytes_written;
        }

    // 7. Write the model-resident heaps, except type OBJ_HEAP, which have been
    //    compressed in heaps of type SQUEEZE_HEAP.

    for (heap_type = FIRST_WRITTEN_HEAP; heap_type < NUM_MODEL_HEAP_TYPES; heap_type++) {
        for (i = 0; i < num_heaps[heap_type]; i++) {
            cur_heap_char_p = *(heaps[heap_type] + i);
            num_bytes = *(bytes_used[heap_type] + i);

            if (-1 == (bytes_written = tml_os_write(file_handle, cur_heap_char_p, num_bytes)))
                return(-1);
            else
                total_bytes_written += bytes_written;
            }
        }

    // 8. Unlock the model in multiprocessing applications

    if (for_model->get_model_lock())
        for_model->unlock();

    // 9. Write the squeeze heaps

    for (i = num_heaps[SQUEEZE_HEAP] - 1; i >= 0; i--) {
        cur_heap_char_p = *(heaps[SQUEEZE_HEAP] + i);
        num_bytes = *(bytes_used[SQUEEZE_HEAP] + i);
        if (-1 == (bytes_written = tml_os_write(file_handle, cur_heap_char_p, num_bytes)))
            return(-1);
        else
            total_bytes_written += bytes_written;
        }

    // 10. Write the bit heaps asynchronously

    if (num_heaps[BIT_HEAP]) {
        bytes_arr = new char*[num_heaps[BIT_HEAP]];
        num_bytes_arr = new long[num_heaps[BIT_HEAP]];
        locks = new csm_lock*[num_heaps[BIT_HEAP]];
        for (i = num_heaps[BIT_HEAP] - 1, j = 0, cur_bit_heap_list = (bit_heap_list*) heap_info->first_list[BIT_HEAP];
          i >= 0; 
          i--, j++, cur_bit_heap_list->get_next()) {
            cur_heap_char_p = *(heaps[BIT_HEAP] + i);
            *(bytes_arr + j) = cur_heap_char_p;
            num_bytes = *(bytes_used[BIT_HEAP] + i);
            *(num_bytes_arr + j) = num_bytes;
            *(locks + j) = cur_bit_heap_list->get_heap()->get_lock();
            }
        tml_os_asynch_write(file_handle, num_heaps[BIT_HEAP], bytes_arr, num_bytes_arr, locks);
        for (i = 0; i < num_heaps[BIT_HEAP]; i++)
            total_bytes_written += *(bytes_used[BIT_HEAP] + i);
        }

    // 11. Unlock the non-model-resident heaps in multiprocessing applications

    if (for_model->get_heaps_lock())
        for_model->unlock_heaps();

    // total_bytes_written does not reflect true size in multiprocessing applications
    return(total_bytes_written);
}


/* function to set types member for model restore */
long d_model_parts::read_types(void* file_handle)
{
    long bytes_read;
    long i;

    types = new tml_ext_type[num_types];
    bytes_read = num_types * (long) sizeof(tml_ext_type);
    if (-1 == (bytes_read = tml_os_read(file_handle,
                                       (char *) types,
                                       bytes_read)))
        return(-1);
    for (i = 0; i < num_types; i++)
        (types + i)->unscramble(their_endian);
    return(bytes_read);
}


/* function to allocate heap tracking structures for model restore */
/*   based on header information read from disk file */
/* returns 0 if successful, -1 otherwise */
long d_model_parts::prepare_for_restore(void* file_handle, long max_heap_types)
{
    long           i;
    long           heap_type;
    long           size;
    unsigned long* ptr;

    static long sizes[NUM_HEAP_TYPES] = { obj_heap_incr,
                                          goo_heap_incr,
                                          st_heap_incr,
                                          obj_heap_incr,
                                          goo_heap_incr};
                                              // default sizes for each heap type

    for (heap_type = 0; heap_type < max_heap_types; heap_type++) {
        heap_sizes[heap_type] = new long[num_heaps[heap_type]];
        bytes_used[heap_type] = new long[num_heaps[heap_type]];
        heaps[heap_type] = new CHARP[num_heaps[heap_type]];
        }

    // heap sizes

    for (heap_type = 0; heap_type < max_heap_types; heap_type++) {
        if (1 == version_number) {
            size = sizes[heap_type];
            for (i = 0; i < num_heaps[heap_type]; i++)
                *(heap_sizes[heap_type] + i) = size;
            }
        else {
            if (-1 == tml_os_read(file_handle,
                                  (char *) heap_sizes[heap_type],
                                  ((long) sizeof(long)) * num_heaps[heap_type]))
                return(-1);
            ptr = their_endian->long_arr_shuf((unsigned long*) heap_sizes[heap_type], num_heaps[heap_type]);
            }
        }

    // bytes used per heap (== bytes written to disk per heap)

    for (heap_type = 0; heap_type < max_heap_types; heap_type++) {
        if (-1 == tml_os_read(file_handle,
                              (char *) bytes_used[heap_type],
                              ((long) sizeof(long)) * num_heaps[heap_type]))
            return(-1);
        ptr = their_endian->long_arr_shuf((unsigned long*) bytes_used[heap_type], num_heaps[heap_type]);
        }

    return(0);
}


/* function to restore heap arrays - returns 0 on success, -1 otherwise */
long d_model_parts::restore_heap_array(void* file_handle, long heap_type)
{
    long  i;
    long  this_size;
    heap* new_heap;

    // model-resident heaps

    if (heap_type < NUM_MODEL_HEAP_TYPES) {
        for (i = 0; i < num_heaps[heap_type]; i++) {
            this_size = *(heap_sizes[heap_type] + i);
            new_heap = new_restore_heap(heap_type, this_size);
            if (-1 == new_heap->restore(file_handle, this, i))
                return(-1);
            }
        }

    // non-model-resident heaps

    else {
        for (i = num_heaps[heap_type] - 1; i >= 0; i--) {
            this_size = *(heap_sizes[heap_type] + i);
            new_heap = new_restore_heap(heap_type, this_size);
            if (-1 == new_heap->restore(file_handle, this, i))
                return(-1);
            }
        }

    return(0);
}


/* function to restore the objects in a model to in-memory format from */
/*   their on-disk equivalents */
/* IMPORTANT: Success of this restore strategy depends on fact that */
/*   node::set_type() is not virtual. */
/* NOTE: Type of root node is set twice, once in loop as for all other */
/*   nodes and once before loop to allow use of node::set_node() on root. */
/*   Second setting in loop does no harm and it obviates a test for every */
/*   node restored. Calls to node::set_node() before loop are necessary to */
/*   allow proper setting of up and left pointers of other nodes. */
/*   Type setting is performed only once, in the loop, for all other nodes. */
void d_model_parts::restore_pointers(gc_hash_tbl& hash_tbl)
{
    node       *new_root,                     // root of new model
               *next_node;                    // next node to restore
    node_dir    dir;                          // sentinel for tree traversal
    node_type   scrambled_ntype;

    new_root = (node *) root.get_m_ptr(heaps[OBJ_HEAP]);
    new_root->up = NULL;
    new_root->left = NULL;

    next_node = new_root;
    dir = n_down;
    do {
        scrambled_ntype = ((d_node*) obj_noalloc())->get_type_index();
        next_node->set_type((node_type) their_endian->my_long((unsigned long) scrambled_ntype));
        next_node->set_type(csm_type_q_get_type((long) next_node->get_type(), types));
        node_set_vptr(next_node);
        next_node->set_model(for_model);

        next_node->restore(*this, hash_tbl);
        do {
            next_node = next_node->get_next(dir);
            if ((NULL == next_node) || (new_root == next_node))
                break;
            } while (n_up == dir);
        } while ( ! ((next_node == new_root) && (n_up == dir)));

    for_model->set_root(new_root);
}


/* function to queue heaps of reduced-size objects for later write/read to/from disk */
/*  during saving/restoring of tml objects */
void d_model_parts::q_object_data(squeeze_heap* data)
{
    sqheap_list* new_sqheap;

    new_sqheap = new sqheap_list;
    new_sqheap->set_heap(data);
    new_sqheap->set_next(squeeze_data);
    squeeze_data = new_sqheap;
}


/* function to queue heaps of graphic data for later write to disk */
/*  during saving of tml objects */
void d_model_parts::q_bit_data(void* data, long count)
{
    bit_heap*      new_bit_heap;
    bit_heap_list* new_bit_heap_list;

    new_bit_heap = new write_bit_heap(data, count);
    new_bit_heap_list = new bit_heap_list;
    new_bit_heap_list->set_heap(new_bit_heap);
    new_bit_heap_list->set_next(bit_data);
    bit_data = new_bit_heap_list;
}


/* function to queue heaps of graphic data as they are read from disk */
/*  during restoring of tml objects */
void d_model_parts::q_bit_data(read_bit_heap* new_heap)
{
    bit_heap_list* new_bit_heap_list;

    new_bit_heap_list = new bit_heap_list;
    new_bit_heap_list->set_heap(new_heap);
    new_bit_heap_list->set_next(bit_data);
    bit_data = new_bit_heap_list;
}


/* function to allocate reduced-size objects from heaps created for write to disk */
/*   during saving of tml objects */
char* d_model_parts::obj_alloc(long count)
{
    if (cur_squeeze_heap->get_byte() + count > cur_squeeze_heap->get_size()) {
        if (count > cur_squeeze_heap->get_size())
            cur_squeeze_heap = new squeeze_heap(count);
        else
            cur_squeeze_heap = new squeeze_heap(cur_squeeze_heap->get_size());
        q_object_data(cur_squeeze_heap);
        }
    return(cur_squeeze_heap->alloc(count));
}


/* function to reallocate reduced-size objects from heaps read in from disk */
/*   during restoring of tml objects */
char* d_model_parts::obj_realloc(long count)
{
    if (cur_squeeze_heap->get_byte() + count > cur_squeeze_heap->get_size()) {
        ++cur_heap[SQUEEZE_HEAP];
        cur_squeeze_heap = squeeze_data->get_next()->get_heap();
        squeeze_data = squeeze_data->get_next();
        }
    return(cur_squeeze_heap->alloc(count));
}


/* function to return first byte available in model_parts object heaps, i.e. */
/*   what will be the first byte of the next reduced-size object to be reallocated */
/*   during restoring of tml objects */
/* Used to obtain pointer to reduced size object corresponding to next object to be */
/*   restored so that a type field for next object to be restored can be located. */
/*   Type field of reduced-size object in turn is used to initialize type field of */
/*   object to be restored, allowing restoration of its virtual function pointer and */
/*   call to object's restore function. */
/* Never affects data stored in or pointed to by d_model_parts. */
char* d_model_parts::obj_noalloc()
{
    if (*(bytes_used[SQUEEZE_HEAP] + cur_heap[SQUEEZE_HEAP]) > cur_squeeze_heap->get_byte())
        return(cur_squeeze_heap->get_start() + cur_squeeze_heap->get_byte());
    else
        return(*(heaps[SQUEEZE_HEAP] + cur_heap[SQUEEZE_HEAP] + 1));
}


/*******************************************************************/
/* member functions for class d_heap_info                          */
/*******************************************************************/

/* constructor for class heap_info */
d_heap_info::d_heap_info(model* write_model)
{
    long i;

    cur_heap[OBJ_HEAP] = write_model->get_current_oheap();
    cur_heap[GOO_HEAP] = write_model->get_current_gheap();
    cur_heap[ST_HEAP] =  write_model->get_current_sheap();

    first_list[OBJ_HEAP] = write_model->get_first_oheap();
    first_list[GOO_HEAP] = write_model->get_first_gheap();
    first_list[ST_HEAP] = write_model->get_first_sheap();

    for (i = 0; i < NUM_MODEL_HEAP_TYPES; i++)
        sort_array[i] = NULL;
    for (i = NUM_MODEL_HEAP_TYPES; i < NUM_HEAP_TYPES; i++) {
        cur_heap[i] = NULL;
        first_list[i] = NULL;
        sort_array[i] = NULL;
        }
}


/* comparsion function used as argument for call to tml_os_qsort() */
/*   in d_heap_info::complete_for_heap_type() below */
long heap_key_cmp(heap_key* ptr1, heap_key* ptr2)
{
    return(tml_os_ptr_cmp(&ptr1->start_addr, &ptr2->start_addr));
}


/* completer function for d_heap_info called by the d_model_parts constructor */
/*   used in saving tml objects to disk. Function is called once for each type of */
/*   model_resident heap */
void d_heap_info::complete_for_heap_type(long heap_type, long num_heaps)
{
    heap_key*  this_key;
    heap*      cur_heap;
    heap_list* cur_list;

    // Allocate and initialize an array of sort keys for heaps of
    //   the given type

    this_key
        = sort_array[heap_type]
        = new heap_key[num_heaps];
    for (cur_list = first_list[heap_type];
         cur_list;
         cur_list = cur_list->get_next()) {
        cur_heap = cur_list->get_heap();
        this_key->start_addr = cur_heap->get_start();
        this_key->heap_size = cur_heap->get_size();
        this_key->bytes_used = cur_heap->get_byte();
        ++this_key;
        }

    // Sort the array in increasing numerical order

    tml_os_qsort((char*) sort_array[heap_type], (unsigned long) num_heaps,
                 (unsigned long) sizeof(heap_key), (SORT_FUNC) heap_key_cmp);
}


/* prepare a d_heap_info for deletion */
void d_heap_info::prepare_for_destruction(d_model_parts* model_parts)
{
    long heap_type;
    long i;

    for (heap_type = 0; heap_type < NUM_MODEL_HEAP_TYPES; heap_type++) {
        cur_heap[heap_type] = NULL;
        first_list[heap_type] = NULL;
        for (i = 0; i < model_parts->num_heaps[heap_type]; i++)
            (sort_array[heap_type] + i)->start_addr = NULL;
        delete sort_array[heap_type];
        sort_array[heap_type] = NULL;
        }
}


/************************************************************************/
/* functions to save into and restore from reduced-size objects on disk */
/************************************************************************/


/* set disk version of d_node tree pointer in direction specified */
void d_node::set_right(d_ptr *d_ptr_tmp)
{
    WHOLE_PTR  disk_p;
    
    disk_p = d_ptr_tmp->as_whole_ptr();
    right = (node *) disk_p;
}


/* set disk version of d_node's property vector pointer */
void d_node::set_pv_d_ptr(d_ptr *d_ptr_tmp)
{
    WHOLE_PTR disk_p;

    disk_p = d_ptr_tmp->as_whole_ptr();
    properties = (prop_vec *) disk_p;
}


void d_node::set_type_index(node *save_node)
{
    n_type = (node_type) csm_type_q_get_index(save_node->get_type());
}


void d_node::unscramble(endian* their_endian)
{
    ((d_ptr*) &right)->unscramble(their_endian);
    ((d_ptr*) &properties)->unscramble(their_endian);
    n_type = (node_type) their_endian->my_long((unsigned long) n_type);
}

void d_bnode::set_down_ptr(d_ptr *d_ptr_tmp)
{
    WHOLE_PTR  disk_p;

    disk_p = d_ptr_tmp->as_whole_ptr();
    down_ptr = (node *) disk_p;
}


void d_bnode::unscramble(endian* their_endian)
{
    ((d_ptr*) &down_ptr)->unscramble(their_endian);
}


void d_gnode::set_goo_d_ptr(d_ptr *d_ptr_tmp)
{
    WHOLE_PTR  disk_p;

    disk_p = d_ptr_tmp->as_whole_ptr();
    goo_ptr = (char *) disk_p;
}


void d_gnode::unscramble(endian* their_endian)
{
    ((d_ptr*) &goo_ptr)->unscramble(their_endian);
    count = (long) their_endian->my_long((unsigned long) count);
}


void d_prop_vec::set_d_ptr(d_ptr *d_ptr_tmp)
{
    WHOLE_PTR disk_p;

    disk_p = d_ptr_tmp->as_whole_ptr();
    next = (prop_vec *) disk_p;
}


void d_prop_vec::unscramble(endian* their_endian)
{
    ((d_ptr*) &next)->unscramble(their_endian);
    ((d_ptr*) &prop)->unscramble(their_endian);
}


/* set prop_vec's prop member to pseudo-pointer for disk save */
void d_prop_vec::set_prop_d_ptr(d_ptr *d_ptr_tmp)
{
    WHOLE_PTR disk_p;

    disk_p = d_ptr_tmp->as_whole_ptr();
    prop = (property *) disk_p;
}


void d_property::set_entry_d_ptr(d_ptr *d_ptr_tmp)
{
    WHOLE_PTR disk_p;

    disk_p = d_ptr_tmp->as_whole_ptr();
    entry_arg = (argument *) disk_p;
}


/* set the disk pseudo-pointer version of the property's exit_arg queue for disk save */
void d_property::set_exit_d_ptr(d_ptr *d_ptr_tmp)
{
    WHOLE_PTR disk_p;

    disk_p = d_ptr_tmp->as_whole_ptr();
    exit_arg = (argument *) disk_p;
}


/* utility function searches through global array of types */
/*   to get index for disk save */
void d_property::set_type_index(property* save_prop)
{
    index = (p_index) csm_type_q_get_index(save_prop->get_index());
}


void d_property::unscramble(endian* their_endian)
{
    ((d_ptr*) &entry_arg)->unscramble(their_endian);
    ((d_ptr*) &exit_arg)->unscramble(their_endian);
    index = (p_index) their_endian->my_long((unsigned long) index);
}


void d_argument::set_type_index(argument *save_arg)
{
    vtype = (arg_type) csm_type_q_get_index(save_arg->get_vtype());
}


/* utility function searches through array of types read from disk */
/*   to set index on restore from disk */
void d_argument::set_d_ptr(d_ptr *d_ptr_tmp)
{
    WHOLE_PTR disk_p;

    disk_p = d_ptr_tmp->as_whole_ptr();
    next = (argument *) disk_p;
}


/*set value field of d_argument to pseudo-pointer for disk save */
void d_argument::set_val_d_ptr(d_ptr *d_ptr_tmp)
{
    WHOLE_PTR disk_p;

    disk_p = d_ptr_tmp->as_whole_ptr();
    arg_val = (void*) disk_p;
}


void d_argument::unscramble(endian* their_endian)
{
    vtype = (arg_type) their_endian->my_long((unsigned long) vtype);
    ((d_ptr*) &next)->unscramble(their_endian);
    ((d_ptr*) &arg_val)->unscramble(their_endian);
}


void d_string_arg::unscramble(endian* their_endian)
{
    count = (long) their_endian->my_long((unsigned long) count);
    d_argument::unscramble(their_endian);
}


/************************************************************************/
/* FUNCTIONS USED TO RESTORE TML VERSION 1 FILES ONLY                   */
/************************************************************************/

/* restore nodes written to disk by TML Version 1 programs */
/* NOTE: Casts to node* in restore of right and left pointers is */
/*   appropriate for conversion from TML Version 1 to TML Version 2 */
/*   but changes to profile of class node in later TML versions may */
/*   require cast to an appropriate intermediate version type */
void node::restore_v1(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, node_v1* node_ver_1)
{
    d_ptr         d_ptr_tmp;
    gc_hash_elem* hash_elem;
    prop_vec*     next_prop;
    short*        src;
    long          temp;

    node_ver_1->unscramble(model_parts.their_endian);

    // up and left pointers are restored to appropriate node pointers
    //   in main loop of d_model_parts::restore_pointers_v1() but left
    //   pointer (still a disk-format node_v1* at this point) must also
    //   be restored here to convert disk-format NULL pointers to
    //   memory-format NULL pointers

    d_ptr_tmp = node_ver_1->left;
    left = (node*) d_ptr_tmp.get_m_ptr(model_parts.heaps[OBJ_HEAP]);

    d_ptr_tmp = node_ver_1->right;
    right = (node*) d_ptr_tmp.get_m_ptr(model_parts.heaps[OBJ_HEAP]);

    d_ptr_tmp = node_ver_1->properties;
    properties = (prop_vec*) d_ptr_tmp.get_m_ptr(model_parts.heaps[OBJ_HEAP]);
    if (properties) {
        next_prop = properties;
        while (next_prop) {
            if (NULL == hash_tbl.find((void*) d_ptr_tmp.as_whole_ptr())) {
                hash_elem = hash_tbl.insert((void*) d_ptr_tmp.as_whole_ptr());
                hash_elem->set_new(next_prop);
                next_prop->set_model(get_model());
                next_prop->restore_v1(model_parts, hash_tbl);
                d_ptr_tmp = next_prop->get_next();
                next_prop->set_m_ptr((prop_vec*) d_ptr_tmp.get_m_ptr(model_parts.heaps[OBJ_HEAP]));
                next_prop = next_prop->get_next();
                }
            else
                break; // non-null means remainder of vector restored previously
            }
        }

    // sib_order was short in TML Version 1
    //   must restore to long for TML Version 2

    src = (short*) &(node_ver_1->sib_order);
    temp = *src;
    if (0 == temp)
        temp = 1;
    sib_order = temp;
}


/* restore bnodes written to disk by tml version 1 programs */
void bnode::restore_v1(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, node_v1* node_ver_1)
{
    d_ptr     d_ptr_tmp;
    bnode_v1 *bnode_ver_1;

    bnode_ver_1 = (bnode_v1*) node_ver_1;
    bnode_ver_1->unscramble(model_parts.their_endian);

    d_ptr_tmp = bnode_ver_1->down_ptr;
    down_ptr = (node*) d_ptr_tmp.get_m_ptr(model_parts.heaps[OBJ_HEAP]);

    node::restore_v1(model_parts, hash_tbl, node_ver_1);
}


/* restore gnodes written to disk by tml version 1 programs */
void gnode::restore_v1(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, node_v1* node_ver_1)
{
    d_ptr     d_ptr_tmp;
    short*    src;
    long      temp;
    gnode_v1* gnode_ver_1;

    gnode_ver_1 = (gnode_v1*) node_ver_1;
    gnode_ver_1->unscramble(model_parts.their_endian);

    d_ptr_tmp = gnode_ver_1->goo_ptr;
    goo_ptr = (char*) d_ptr_tmp.get_m_ptr(model_parts.heaps[GOO_HEAP]);

    // count member was short in tml version 1
    //   must restore to long for version 2

    src = (short*) &(gnode_ver_1->count);
    temp = *src;
    count = temp;

    node::restore_v1(model_parts, hash_tbl, node_ver_1);
}


/* restore property vectors written to disk by tml version 1 programs */
/*   note that the property must be reallocated according to the version 2 */
/*   definition of class property */
void prop_vec::restore_v1(d_model_parts& model_parts, gc_hash_tbl& hash_tbl)
{
    d_ptr          d_ptr_tmp;
    gc_hash_elem*  hash_elem;
    property_v1*   prop_v1;
    p_index        unscrambled_p_index;
    long           i;
    type_trans_v1* cur_trans;

    unscramble_v1(model_parts.their_endian);

    // next pointer restored by node::restore_v1

    d_ptr_tmp = prop;
    prop_v1 = (property_v1*) d_ptr_tmp.get_m_ptr(model_parts.heaps[OBJ_HEAP]);
    if (prop_v1) {
        if (NULL == (prop = (property*) hash_tbl.find((void*) d_ptr_tmp.as_whole_ptr()))) {
            hash_elem = hash_tbl.insert((void*) d_ptr_tmp.as_whole_ptr());

            // the TML Version 2 equivalent of the TML Version 1 property
            //   is created here. Users who extended TML Version 1 property
            //   must queue on global type_trans_q a type_trans_v1 appropriately
            //   initialized for each extended type. The type_trans_v1 objects
            //   must be queued at the same queue positions that the
            //   corresponding types were queued on type_q in the user's
            //   TML Version 1 program. Note that user must also supply a
            //   ::restore_v1() member function for the corresponding
            //   TML Version 2 extended class.

            unscrambled_p_index = (p_index) model_parts.their_endian->my_long((unsigned long) prop_v1->index);
            cur_trans = tml_type_trans_q_v1_get_head();

            // cast to long is appropriate here as index is actually a long at this point
            for (i = 0; i < (long) unscrambled_p_index; i++)
                cur_trans  = cur_trans->next;
            prop = (property*) (*(cur_trans->make_func))(in_model);

            // must set type here since make function is generic for all
            //   preloaded property types

            prop->set_index(cur_trans->type);
            prop_set_vptr(prop);
            hash_elem->set_new(prop);
            prop->restore_v1(model_parts, hash_tbl, prop_v1);
            }
        }
}


void prop_vec::unscramble_v1(endian* their_endian)
{
    ((d_ptr*) &next)->unscramble(their_endian);
    ((d_ptr*) &prop)->unscramble(their_endian);
    ref_count = (long) their_endian->my_long((unsigned long) ref_count);
}


/* restore properties and arguments written to disk by tml version 1 programs */
/* note that the arguments must be reallocated due to change in definition */
/*   of class argument */
void property::restore_v1(d_model_parts& model_parts, gc_hash_tbl& hash_tbl, property_v1 *prop_v1)
{
    d_ptr         d_ptr_tmp;
    gc_hash_elem* hash_elem;
    d_ptr         arg_vec_heads_v1[2];
    argument_v1*  next_arg_v1;
    argument*     next_arg;
    argument*     prev_arg;
    argument*     saved;
    short*        src;
    long           i,
                  temp;

    prop_v1->unscramble(model_parts.their_endian);

    arg_vec_heads_v1[0] = prop_v1->entry_arg;
    arg_vec_heads_v1[1] = prop_v1->exit_arg;
    for (i = 0; i < 2; i++) {
        prev_arg = NULL;
        next_arg_v1 = (argument_v1*) arg_vec_heads_v1[i].get_m_ptr(model_parts.heaps[OBJ_HEAP]);
        while (next_arg_v1) {
            next_arg_v1->unscramble(model_parts.their_endian);
            if (NULL == (saved = (argument*) hash_tbl.find(next_arg_v1))) {
                hash_elem = hash_tbl.insert(next_arg_v1);
                switch (next_arg_v1->a_type) {
                    case int_val:
                        next_arg = locator.in_model->make_arg(next_arg_v1->arg_val.ival);
                        break;
                    case float_val:
                        next_arg = locator.in_model->make_arg(next_arg_v1->arg_val.fval);
                        break;
                    case string_val:
                        d_ptr_tmp = next_arg_v1->arg_val.sval.s_start;
                        next_arg = locator.in_model->make_arg((char*) d_ptr_tmp.get_m_ptr(model_parts.heaps[ST_HEAP]),
                                                      next_arg_v1->arg_val.sval.s_count);
                        break;
                    case no_val:
                        printf("property::restore_v1: argument_v1 with a_type no_val - cannot restore\n");
                        break;
                    default:
                        break;
                    }
                hash_elem->set_new(next_arg);

                if (NULL == prev_arg) {
                    if (0 == i)
                        entry_arg = next_arg;
                    else
                        exit_arg = next_arg;
                    }
                else
                    prev_arg->append(next_arg);
                prev_arg = next_arg;

                d_ptr_tmp = next_arg_v1->next;
                next_arg_v1 = (argument_v1*) d_ptr_tmp.get_m_ptr(model_parts.heaps[OBJ_HEAP]);
                }
            else {
                if (NULL == prev_arg) {
                    if (0 == i)
                        entry_arg = saved;
                    else
                        exit_arg = saved;
                    }
                else
                    prev_arg->append(saved);
                break; // non-null means remainder of vector restored previously
                }
            }
        }

    src = (short*) &(prop_v1->ref_count);
    temp = *src;
    ref_count = temp;

    act_value = NULL;
}


/* function to initialize a d_model_parts object from a d_model_info_v4 */
/*   header read from a file written by TML Versions 2 through 4 */
void d_model_parts::from_model_info_v4(d_model_info_v4* model_info_v4)
{
    root = model_info_v4->root;
    num_types = model_info_v4->num_types;
    tml_os_byte_copy((char*) model_info_v4->num_heaps,   (char*) num_heaps,   (long) (sizeof(long)     * (SQUEEZE_HEAP + 1)));
    tml_os_byte_copy((char*) model_info_v4->heap_quanta, (char*) heap_quanta, (long) (sizeof(long)     * (SQUEEZE_HEAP + 1)));
    tml_os_byte_copy((char*) model_info_v4->heap_quanta, (char*) heap_quanta, (long) (sizeof(HALF_PTR) * (SQUEEZE_HEAP + 1)));
}


/* function to initialize a d_model_parts object from a d_model_info_v1 */
/*   header read from a TML Version 1 file */
void d_model_parts::from_model_info_v1(d_model_info_v1* model_info_v1)
{
    root = model_info_v1->root;
    num_types = model_info_v1->num_types;
    num_heaps[OBJ_HEAP] = model_info_v1->num_oheaps;
    num_heaps[GOO_HEAP] = model_info_v1->num_gheaps;
    num_heaps[ST_HEAP] = model_info_v1->num_sheaps;
    heap_quanta[OBJ_HEAP] = obj_heap_incr;
    heap_quanta[GOO_HEAP] = goo_heap_incr;
    heap_quanta[ST_HEAP] = st_heap_incr;
    cur_heap[OBJ_HEAP] = model_info_v1->cur_oheap;
    cur_heap[GOO_HEAP] = model_info_v1->cur_gheap;
    cur_heap[ST_HEAP] = model_info_v1->cur_sheap;
}

                                                                        
/* function to set types member for TML Version 2 restore of */
/*   TML Version 1 models */
long d_model_parts::read_types_v1(void* file_handle)
{
    tml_ext_type_v1* types_v1;
    long             bytes_read;
    long             i;
    type_trans_v1*   cur_trans;

    types_v1 = new tml_ext_type_v1[num_types];
    types = new tml_ext_type[num_types];
    bytes_read = num_types * (long) sizeof(tml_ext_type_v1);
    if (-1 == (bytes_read = tml_os_read(file_handle,
                                       (char *) types_v1,
                                       bytes_read)))
        return(-1);
    for (i = 0; i < num_types; i++) {
        (types_v1 + i)->unscramble(their_endian);
        for (cur_trans = tml_type_trans_q_v1_get_head();
             cur_trans;
             cur_trans = cur_trans->next) {
            if (*(cur_trans->type_v1) == *(types_v1 + i))
                *(types + i) = *(cur_trans->type->type);
            }
        }
    delete types_v1;
    return(bytes_read);
}


/* function to restore the objects in a TML Version 1 model to */
/*   TML Version 2 in-memory format from their on-disk equivalents */
/* NOTE: Casts in navigation loop are valid between TML Version 1 */
/*   and TML Version 2 but may have to be altered to casts to an */
/*   intermediate type if later TML versions alter the profile of */
/*   class node */
void d_model_parts::restore_pointers_v1(gc_hash_tbl& hash_tbl)
{
    node_v1       *old_root,      // root saved by TML Version 1
                  *old_node;      // next TML Version 1 node to restore
    node          *new_root,      // root of new model (TML Version 2)
                  *next_node,     // TML Version 2 node to restore
                  *prev_node;     // previous TML Version 2 node restored
    node_dir       dir;           // sentinel for tree traversal
    long           type_index;    // TML Version 1 type index of
                                  //   next node to be restored
    long           i;
    type_trans_v1* cur_trans;     // translates TML Version 1 type to
                                  //   TML Version 2

    old_root = (node_v1*) root.get_m_ptr(heaps[OBJ_HEAP]);
    old_node = old_root;

    new_root = NULL;
    next_node = NULL;

    dir = n_down;
    prev_node = NULL;
    do {

        // the TML Version 2 equivalent of the TML Version 1 node
        //   is created here. Users who extended TML Version 1 node
        //   must queue on global type_trans_q a type_trans_v1 appropriately
        //   initialized for each extended type. The type_trans_v1 objects
        //   must be queued at the same queue positions that the
        //   corresponding types were queued on type_q in the user's
        //   TML Version 1 program. Note that user must also supply a
        //   ::restore_v1() member function for the corresponding
        //   TML Version 2 extended class.

        // cast to long is appropriate since that is the type of
        //   index into the type array which is saved to disk

        type_index = (long) old_node->get_type();
        type_index = (long) their_endian->my_long((unsigned long) type_index);
        cur_trans = tml_type_trans_q_v1_get_head();
        for (i = 0; i < type_index; i++)
            cur_trans  = cur_trans->next;
        next_node = (node*) (*(cur_trans->make_func))(for_model);
        if (old_node == old_root) {
            new_root = next_node;
            for_model->set_root(new_root);
            }

        // type actually needn't be set for preloaded node types since
        //   make function for nodes is not generic owing to variation
        //   in size among preloaded node types
        next_node->set_type(cur_trans->type);
        node_set_vptr(next_node);
        next_node->restore_v1(*this, hash_tbl, old_node);

        // not root or leftmost sibling

        if (next_node->get_node(n_left)) {
            prev_node->set_node(next_node, n_right);
            next_node->set_node(prev_node, n_left);
            next_node->set_node(prev_node->get_node(n_up), n_up);
            }
        else {
            if (prev_node)
                prev_node->set_node(next_node, n_down);
            next_node->set_node(prev_node, n_up);
            }

        do {
            prev_node = next_node;
            next_node = next_node->get_next(dir);
            } while ( ! (next_node == new_root) && (n_up == dir));

        old_node = (node_v1*) next_node;

        } while ( ! ((next_node == new_root) && (n_up == dir)));
}


/************************************************************************/
/* MEMBER FUNCTIONS FOR CLASSES REPRESENTING OBJECTS                    */
/*   USED BY TML VERSION 1 ONLY                                         */
/************************************************************************/

/* constructor for class node_v1, the TML version 1 node */
/* SHOULD NEVER BE CALLED - this constructor is included solely to satisfy C++ */
node_v1::node_v1(long size, model *to_model)
{
}


void node_v1::unscramble(endian* their_endian)
{
    ((d_ptr*) &up)->unscramble(their_endian);
    ((d_ptr*) &right)->unscramble(their_endian);
    ((d_ptr*) &left)->unscramble(their_endian);
    ((d_ptr*) &properties)->unscramble(their_endian);
    sib_order = their_endian->my_short((unsigned short) sib_order);
    // type field is unscrambled before this function is called
}


/* constructor for class bnode_v1, the TML version 1 bnode */
/* SHOULD NEVER BE CALLED - this constructor is included solely to satisfy C++ */
bnode_v1::bnode_v1(model *to_model, long size) : (size ? size : (long) sizeof(bnode_v1), to_model)
{
}


void bnode_v1::unscramble(endian* their_endian)
{
    ((d_ptr*) &down_ptr)->unscramble(their_endian);
}


/* constructor for class gnode_v1, the TML version 1 gnode */
/* SHOULD NEVER BE CALLED - this constructor is included solely to satisfy C++ */
gnode_v1::gnode_v1(model *to_model, long size) : (size ? size : (long) sizeof(gnode_v1), to_model)
{
}


void gnode_v1::unscramble(endian* their_endian)
{
    ((d_ptr*) &goo_ptr)->unscramble(their_endian);
    count = their_endian->my_short((unsigned short) count);
}


void property_v1::unscramble(endian* their_endian)
{
    ref_count = their_endian->my_short((unsigned short) ref_count);
    index = (p_index) their_endian->my_long((unsigned long) index);
    ((d_ptr*) &entry_arg)->unscramble(their_endian);
    ((d_ptr*) &exit_arg)->unscramble(their_endian);
}


void argument_v1::unscramble(endian* their_endian)
{
    a_type = (arg_type_v1) their_endian->my_short((unsigned short) a_type);
    ref_count = (long) their_endian->my_long((unsigned long) ref_count);
    switch (a_type) {
        case int_val:
            arg_val.ival = (long) their_endian->my_long((unsigned long) arg_val.ival);
            break;
        case float_val:
            arg_val.fval = (float) their_endian->my_long((unsigned long) arg_val.fval);
            break;
        case string_val:
            ((d_ptr*) &(arg_val.sval.s_start))->unscramble(their_endian);
            arg_val.sval.s_count = (long) their_endian->my_long((unsigned long) arg_val.sval.s_count);
            break;
        case no_val:
            printf("argument_v1::unscramble: a_type is no_val - cannot unscramble\n");
            break;
        default:
            break;
        }
}


