/**************************************************************************************
    tml_os_bsd.cxx
    definitions of the operating system dependent routines in tml++ 
      for UNIX bsd4.3
    
    (c) Copyright Apollo Computer Inc., 1987 (unpublished)
     --------------------------------------------------------------------
    |THE FOLLOWING PROGRAMS ARE THE SOLE PROPERTY OF APOLLO COMPUTER INC.|
    |      AND CONTAIN ITS PROPRIETARY AND CONFIDENTIAL INFORMATION.     |
     --------------------------------------------------------------------
**************************************************************************************/

#include <stdio.h>

#ifdef __cplusplus
#include <stdlib.h>
#include <osfcn.h>
#endif

#include <string.h>

#ifndef apollo
#include <strings.h>
#endif

#include <tml/tml_cxx2.0_linkage_decls.hxx>
#include <tml/arch_neutral.hxx>          /* defines WHOLE_PTR */
#include <tml/csm_lock.hxx>
#include <tml/tml_os_depends.hxx>

/* copy count bytes at from address into space beginning at to address */
char* tml_os_byte_copy(char* from, char* to, long count)
{
#ifndef __cplusplus
    extern long bcopy(char*, char*, long);
#endif

    /*check for incoming trash*/
    if (!from || !to)
        return 0;
    bcopy(from, to, (int) count);
    return(to);
}

/* sort the nel objects of size "size" at base according to comparison function cmp */
void tml_os_qsort(char *base, unsigned long nel, unsigned long size, long (*cmp)(void*, void*))
{
    /*check for incoming trash*/
    if (!base || !nel || !size || !cmp)
        return;

#ifdef __cplusplus
    qsort(base, (unsigned int) nel, (unsigned int) size, (int (*)(const void*, const void*)) cmp);
#else
    qsort(base, nel, size, cmp);
#endif
}

/* return NULL on failure or pointer to nelem * elsize bytes of free space. */
/*   Each byte is initialized to zero. */
char* tml_os_alloc(unsigned long nelem, unsigned long elsize)
{
    register char* temp;
    register long  i;
    register long  quantity;

    quantity = nelem * elsize;
    temp = new char[quantity];
    for (i = 0; i < quantity; i++)
        *(temp + i) = 0;
    return temp;
}

/* free the space at ptr, which must have been obtained from a call to */
/*   tml_os_alloc(), or the result is unpredictible. */
void tml_os_free(char* ptr)
{
    delete ptr;
}

/* Pointer comparison function for calls to tml_os_qsort() in d_write_model() that sort */
/*   the disk model information structure's arrays of heap pointers */
long tml_os_ptr_cmp(char* *ptr1, char* *ptr2)
{
    if ((WHOLE_PTR) *ptr1 > (WHOLE_PTR) *ptr2)
        return(1);
    else if ((WHOLE_PTR) *ptr1 < (WHOLE_PTR) *ptr2)
        return(-1);
    else
        return(0);
}

/* Semaphore structure for bsd -- restricted to use by */
/*   parent in forking children since bsd has no shared */
/*   memory or kernel support for semaphores */
struct bsd_tml_sem {
    long   capacity;
    long   used;
    char** lockfiles;
};

/* Initialize a semaphore structure with a given capacity */
void* tml_os_init_sem(long pool_capacity)
{
    bsd_tml_sem* new_sem;
    long         i, j;
    char         tmpname[100];
    long         numstart;
    void*        fdp;

    new_sem = new bsd_tml_sem;
    new_sem->capacity = pool_capacity;
    new_sem->used = 0;
    new_sem->lockfiles = new char*[pool_capacity];
    sprintf(tmpname, "%s%d%d%c", "/usr/tmp/csmlock", tml_os_getpid(), (long) new_sem, '\0');
    numstart = strlen(tmpname);
    for (i = 0; i < pool_capacity; i++) {
        sprintf(tmpname + numstart, "%d%c", i, '\0');
        if (NULL == (fdp = (tml_os_open(tmpname, strlen(tmpname), O_RDWR, 0)))) {
            printf("tml_os_init_sem: can't open %s\n", tmpname);
            for (j = 0; j < i; j++)
                delete *(new_sem->lockfiles + j);
            delete new_sem->lockfiles;
            delete new_sem;
            return NULL;
            }
        if (-1 == tml_os_close(fdp)) {
            printf("tml_os_init_sem: can't close %s\n", tmpname);
            for (j = 0; j < i; j++)
                delete *(new_sem->lockfiles + j);
            delete new_sem->lockfiles;
            delete new_sem;
            return NULL;
            }
        *(new_sem->lockfiles + i) = tmpname;
        }
    return (void*) new_sem;
}

/* Free a semaphore structure */
void tml_os_free_sem(void* old_sem)
{
    long i;

    bsd_tml_sem* real_sem;

    real_sem = (bsd_tml_sem*) old_sem;
    for (i = 0; i < real_sem->capacity; i++)
        delete *(real_sem->lockfiles + i);
    delete real_sem->lockfiles;
    delete real_sem;
}

/* Busy wait on a semaphore */
long tml_os_busy_wait(void* on_sem)
{
    bsd_tml_sem* real_sem;

    real_sem = (bsd_tml_sem*) on_sem;

    if (real_sem->used == real_sem->capacity)
        return 1;
    return 0;
}

/* Lock a semaphore */
void tml_os_p_sem(void* on_sem)
{
    void*        wassig;
    bsd_tml_sem* real_sem;
    long         fd;

#ifndef __cplusplus
    // Without this declaration, Sun C++ 1.2 assumes
    //   that call of flock() below is an attempt to
    //   call a (nonexistent) constructor for struct
    //   flock (defined in /usr/include/sys/file.h)
    extern int flock(int, int);
#endif

    // record current interrupt handling status
    //   and then ignore interrupts
    wassig = tml_os_ignore_interrupts();

    real_sem = (bsd_tml_sem*) on_sem;
    if (real_sem->used < real_sem->capacity)
        // resource is available
        ++real_sem->used;
    if (-1 == (fd = open(*(real_sem->lockfiles + real_sem->used - 1), O_RDWR, 0755))) {
        printf("tml_os_p_sem: can't open %s\n", *(real_sem->lockfiles + real_sem->used - 1));
        return;
        }
    /* prior lock will cause process to sleep here */
    else if (-1 == flock((int) fd, LOCK_EX)) {
        printf("tml_os_p_sem: can't lock %s\n", *(real_sem->lockfiles + real_sem->used - 1));
        return;
        }
    else
        close((int) fd);

    // restore interrupt handling on wakeup
    //   or immediately if semaphore was unlocked
    wassig = tml_os_restore_interrupts(wassig);
}

/* Continue process waiting on head of queue for semaphore */
void* tml_os_v_sem(void* proc_id, void* on_sem)
{
    void*        wassig;
    bsd_tml_sem* real_sem;
    void*        retval;
    long         fd;

#ifndef __cplusplus
    // Without this declaration, Sun C++ 1.2 assumes
    //   that call of flock() below is an attempt to
    //   call a (nonexistent) constructor for struct
    //   flock (defined in /usr/include/sys/file.h)
    extern int flock(int, int);
#endif

    real_sem = (bsd_tml_sem*) on_sem;

    // record current interrupt handling status
    //   and then ignore interrupts
    wassig = tml_os_ignore_interrupts();

    if (real_sem) {
        --real_sem->used;
        
        if (-1 == (fd = open(*(real_sem->lockfiles + real_sem->used), O_RDWR, 0755))) {
            printf("tml_os_v_sem: can't open %s\n", *(real_sem->lockfiles + real_sem->used));
            retval = (void*) 0;
            }
        else if (-1 == flock((int) fd, LOCK_UN)) {
            printf("tml_os_v_sem: can't unlock %s\n", *(real_sem->lockfiles + real_sem->used));
            retval = (void*) 0;
            }
        if (proc_id)
            retval = (void*) kill((int) proc_id, SIGCONT);
        else
            retval = (void*) 0;
        }
    else if (proc_id)
        retval = (void*) kill((int) proc_id, SIGCONT);
    else
        retval = (void*) 0;

    // restore interrupt handling
    wassig = tml_os_restore_interrupts(wassig);
    
    return retval;
}

/* Check availability of a resource via its semaphore */
long tml_os_resource_available(void* on_sem)
{
    bsd_tml_sem* real_sem;

    real_sem = (bsd_tml_sem*) on_sem;

    if (real_sem->used < real_sem->capacity)
        return -1;
    return 0;
}

/* Return process identifier for the current process -- return value is */
/*   a long value cast as a void* */
void* tml_os_getpid()
{
    return (void*) getpid();
}

/* Reduce vulnerability of the current process to interrupts during critical */
/*   code sections -- returns status of interrupts before the call */
void* tml_os_ignore_interrupts()
{
#ifndef __cplusplus
    return (void*) signal(SIGINT, SIG_IGN);
#else
    return (void*) signal(SIGINT, (SIG_PF) SIG_IGN);
#endif
}

/* Restore vulnerability of the current process to interrupts after completing */
/*   critical code sections -- returns status of interrupts after the call */
void* tml_os_restore_interrupts(void* interrupts)
{
#ifndef __cplusplus
    typedef int (*PFIIAV)(...);
    return (void*) signal(SIGINT, (PFIIAV) interrupts);
#else 
    return (void*) signal(SIGINT, (SIG_PF) interrupts);
#endif

}

/* Kill a process dead -- returns a pointer to the appropriate object */
/*   cast as a void* */
void* tml_os_kill_proc(void* proc_id)
{
    if (proc_id)
        return (void*) kill((int) proc_id, SIGKILL);
    return (void*) 0;
}

/* Argument "file_handle" is the target of the write operation, "bufs" is a set of buffers, */
/*   "nbyte_arr" is their lengths in bytes, "sems" is a semaphore for each buffer. */
/* Do not wait for write operations to complete. */
/* The buffers are guaranteed to be written in the order given */
/*   except when a pending call to the same function accesses */
/*   the same semaphores. */
/* The semaphores are the tml_sem members of csm_lock objects corresponding to the buffers. */
long tml_os_asynch_write(void* file_handle, long num_bufs, char** bufs, long* nbyte_arr, csm_lock** locks)
{
    long i;
    long bytes_written;
    long proc_id;

    if (0 == (proc_id = fork())) {
        /* in child */
        for (i = 0; i < num_bufs; i++) {
            (*(locks + i))->p_suspend();
            if (-1 == (tml_os_write(file_handle, *(bufs + i), *(nbyte_arr + i)))) {
                printf("tml_os_asynch_write: can't write\n");
                return -1;
                }
            (*(locks + i))->v();
            }
        }
    for (bytes_written = 0, i = 0; i < num_bufs; i++)
        bytes_written += *(nbyte_arr + i);
    return bytes_written;
}

